# This is Makefile for the ‹hello› program with implicit rules and with
# conventions in mind.

# We will keep the list of variables.
CFLAGS = -std=c99 -pedantic -Wall -Wextra

# Conventionally, the first target is called ‹all› and it usually
# only calls other targets.
all: hello

# We know that GNU Make has an implicit rule for creating binary files
# from ‹.o› files of this form:
#
# 	%: %.o
# 		$(CC) $(LDFLAGS) -o $@ $^ $(LDLIBS) $(LOADLIBES)
#
# which basically says: If I have ‹X.o› and I want to create ‹X›, I just
# link ‹X.o› (and possibly other dependencies) together using a compiler
# in the variable ‹$(CC)›. (See slides for other variables).
#
# So, we don't need this:
#
#	hello: hello.o greeting.o
#		gcc -o hello hello.o greeting.o
#
# If we say ‹make›, it will attempt to create ‹all›, which will attempt
# to create ‹hello›, ‹hello.o› and so on by using this implicit rule.
#
# However, we still need to tell Make that ‹greeting.o› is also
# a dependency of ‹hello›. How, you may ask? Well, it's ez:
hello: greeting.o

# Now Make will have two rules in it's database:
#
# 	%: %.o
# 		$(CC) $(LDFLAGS) -o $@ $^ $(LDLIBS) $(LOADLIBES)
#
# 	hello: greeting.o
#
# Note that the second rule does not have any commands, it just
# adds dependencies. So, if ‹make hello› gets called, Make will
# infer the following rule:
#
# 	#  % = hello = $*
# 	# $@ = hello
# 	# $^ = hello.o greeting.o
# 	hello: hello.o greeting.o
# 		$(CC) $(LD…) -o hello hello.o greeting.o $(LD…) $(…)
#
# And that's exactly what we wanted, and we didn't have to write a single
# command.

# For object files, we had the following two rules:
#
# 	hello.o: hello.c greeting.h
#		gcc $(CFLAGS) -c -o hello.o hello.c
# 	greeting.o: greeting.c greeting.h
#		gcc $(CFLAGS) -c -o greeting.o greeting.c
#
# Again, GNU Make has an implicit rule for creating ‹.o› files from ‹.c›:
#
# 	%.o: %.c
# 		$(CC) $(CPPFLAGS) $(CFLAGS) -c -o $@ $<
#
# Similarly to the previous example, we only need to specify other
# dependencies for ‹.o› files:
hello.o: greeting.h
greeting.o: greeting.h

# Say that Make attempts to create ‹hello.o›, again, it will put together
# the implicit rule and the added dependency:
#
#	#  % = hello = $*
#	# $@ = hello.o
#	# $^ = hello.c greeting.h
#	# $< = hello.c
# 	hello.o: hello.c greeting.h
# 		$(CC) $(…) $(…) -c -o hello.o hello.c
#
# Also note that the implicit rule uses ‹$<› instead of ‹$^›. This is because
# each ‹.o› file is created from exactly one ‹.c› file, and header files
# are *not* compiled directly.

# To improve on ‹clean›, we just change ‹rm› to ‹$(RM)›, which is
# a platform-independent way to clean files:
clean:
	$(RM) hello.o greeting.o hello

# Make does not usually care about target names. If for some reason
# ‹all› or ‹clean› get created as regular files, ‹make› or ‹make clean›
# will stop working properly.
#
# However, there are special targets, usually of the form ‹.UPPERCASE›,
# that have special meaning.
#
# For example, dependencies of ‹.PHONY› will instruct Make to *always*
# consider them to be “abstract” and to ignore files with the same name,
# if they exist.
#.PHONY: \
#    all \
#    clean
