/* This program deomnstrates use of ‹dlfcn› interface, the real time
 * linker's API.
 *
 * The accompanying ‹Makefile› creates two binaries, 32bit and 64bit.
 * Use one appropriate for your platform (for example, try ‹locate libm.so›).
 *
 * The program requires two arguments: the library and the symbol.
 * It will open the library, read the symbol and call it on a value.
 *
 *  $ ./rtmath64 /usr/lib/libm.so.6 log10
 *  log10> 1000
 *  log10(100.000000) = 2.000000
 *
 *  $ ./rtmath64 /usr/lib/libm.so.6 cos
 *  cos> 3.14159265
 *  cos(3.141593) = -1.000000
 *
 *  $ ./rtmath64 /usr/lib/libm.so.6 covfefe
 *  rtmath64: dlsym: /usr/lib/libm.so.6: undefined symbol: covfefe
 */

#include <stdio.h>
#include <stdlib.h>


/* Remember, using ‹dlfcn.h› requires linking with ‹-ldl›! */
#include <dlfcn.h>
#include <err.h>

/* This typedef declares a type ‹f_dbl› which is a function that takes
 * a value of type ‹double› and returns ‹double›. */
typedef double(*f_dbl)(double);

static
void help(const char *argv0)
{
	printf(
		"usage: %s LIBRARY FUNCTION\n"
		"\n"
		"Opens the given math library and symbol\n"
		"then asks for a number and returns the output.\n"
		"\n"
		"You can locate math libraries on your system by running\n"
		"    locate libm.so\n"
		, argv0);
}

int main(int argc, char *argv[])
{
	if (argc < 2 || argc > 3) {
		help(argv[0]);
		return EXIT_FAILURE;
	}

	/* As we are going to do several operations that may fail, we will
	 * set ‹status› to ‹EXIT_FAILURE› from the start and only change it
	 * to ‹EXIT_SUCCESS› when we are finished. */
	int status = EXIT_FAILURE;

	char *libname = argv[1];
	char *symbol = argv[2];

	/* First, we will open the library. The ‹RTLD_LAZY› will skip binding
	 * symbols in the object's function until they are needed. The opposite
	 * is ‹RTLD_NOW›. */
	void *libhndl = dlopen(libname, RTLD_LAZY);

	/* If the function fails, it returns ‹NULL›. We can check the error by
	 * using ‹dlerror()›. */
	if (!libhndl) {
		warnx("dlopen: %s", dlerror());
		goto main_dlopen_error;
	}

	/* We will get the symbol from the library by calling ‹dlsym()›. Note
	 * that here we typecast ‹void *› to a function pointer. This is something
	 * that standard C99 nor C11 does not specify behaviour for. However,
	 * it is perfectly valid in POSIX for this situation. */
	f_dbl f = (f_dbl) dlsym(libhndl, symbol);

	/* Again, we get ‹NULL› on error. */
	if (!f) {
		warnx("dlsym: %s", dlerror());
		goto main_dlsym_error;
	}

	double input;
	printf("%s> ", symbol);

	if (scanf("%lf", &input) != 1) {
		fprintf(stderr, "invalid input\n");
		goto main_dlsym_error;
	}

	printf("%s(%lf) = %lf\n", symbol, input, f(input));
	status = EXIT_SUCCESS;

	/* Clean up is performed using ‹goto› cleaner blocks. While ‹goto› itself
	 * is usually regarded as bad practice, when working with resources
	 * it is actually very useful to use like this, since we don't need to
	 * have “cleaner functions” or clean resources in ‹if› branches. */
main_dlsym_error:
	dlclose(libhndl);

main_dlopen_error:
	return status;
}
