#define _POSIX_C_SOURCE 200809L
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>

/* ‹getopt()› function is included in ‹unistd.h› header file. Aside from the
 * function itself, it allows us to access four global variables:
 *
 * • ‹optarg› is a pointer to ‹argv› when ‹getopt()› encounters an option
 *      with parameter.
 * • ‹optind› is an index in ‹argv› (not a pointer!) of the next element
 *      that will be processed in the next iteration. After ‹getopt()›
 *      returns ‹-1›, then ‹argv[optind]› is the first positional argument.
 * • ‹opterr› is by default set to a non-zero value, which causes ‹getopt()›
 *      to write error message in pathological situation. Setting this variable
 *      to zero disables error messages.
 * • ‹optopt› contains the invalid option character in case an invalid option
 *      is encountered. */

/* It is often a good practice to store program options in a structure.
 * This will allow you to add more options later without the need to change
 * signatures of functions that use these options. */
struct options {
	bool a;
	bool b;
	char *c;

	/* Sometimes it is also useful to include program parameters by including
	 * a pointer to the relevant part of ‹argv› (without options), especially
	 * if they need to be passed to other functions.
	 * See below how these are set. */
	int argc;
	/* NOTE: We cannot declare ‹argv› as ‹char *argv[]› here. Why? */
	char **argv;
};

/* Typically, programs have two forms of help text: usage and help.
 *  1. Usage is a oneliner explaining briefly how the program should be invoked.
 *     It is often displayed on standard error when the program is invoked
 *     wrongly. */
static
void usage(FILE *target, const char *pname)
{
	/* Synopsis syntax usually follows the following conventions:
	 * • ‹[α]› indicates an optional construct, e.g. ‹[-f]› or ‹[NAME]›.
	 * 	 Even ‹[-f [OPTION]]› things exist sometimes.
	 * • Long named parameters can be displayed as either ‹UNDERSCORE_CASE›
	 *   or using ‹<ANGLE BRACKETS>›.
	 * • ‹α|β› denotes mutually exclusive arguments.
	 * • ‹...› (ellipsis) indicates repetition.
	 *
	 * Regarding the choice of options, POSIX Utility Guidelines states:
	 * • Option characters should belong to set of alphanumeric characters
	 *   only, i.e. A-Z, a-z or 0-9.
	 *   ◦ However, use of ‹-NUMBER› is discouraged nowadays. Avoid it
	 *     if you can.
	 * • Option ‹-W› should be reserved for “vendor options”. */
	fprintf(target, "usage: %s [-h] [-ab] [-c ARG] [PARAMETER]...\n", pname);
}

/*  2. Help is usually a longer text explaining briefly how program behaves
 *     and what its options are. Note that this should «not» be the full
 *     documentation, just a bit more verbose synopsis.
 *
 *     It is often requested by ‹-h›, and thus goes to ‹stdout›. */
static
void help(const char *pname)
{
	/* Sometimes, help includes the synopsis written to ‹stdout›. */
	usage(stdout, pname);

	/* In C, you can write list of strings without a separator (e.g. no ‹,›)
	 * and they will be joined into a single string by the preprocessor. */
	printf(
		"Example of getopt(3p) usage.\n"
		"\n"
		"Brief description of options and arguments:\n"
		"  -h           Show this text and exit.\n"
		"  -a           Set option a.\n"
		"  -b           Set option b.\n"
		"  -c ARG       Set option c to value ARG.\n"
		"\n"
		"  PARAMETER    A positional string\n"
	);
}

static
bool options_parse(int argc, char *argv[], struct options *options)
{
	/* An advantage of local ‹static const› variable is that it is basically
	 * a compile-time constant, but is not accessible outside the scope
	 * of this function. */
	static const char optstring[] = "habc:";

	int option;
	/* POSIX ‹getopt()› scans ‹argv› in each call. When the currently processed
	 * argument is an option found in ‹optstring›, its character is returned,
	 * otherwise ‹?› gets returned. ‹optstring› can have the following quirks:
	 * • If a character is followed by ‹:›, it declares an option with parameter.
	 *   This parameter is accessible via ‹optarg›.
	 * • If the string begins with ‹:›, the ‹:› character is also returned when
	 *   there is an option parameter missing, otherwise ‹?› is returned in
	 *   this case.
	 *
	 * When there are no more options, ‹getopt()› returns ‹-1›. */

	/* GNU ‹getopt()› has a few more features:
	 * • It automatically permutes arguments so that options are always
	 *   processed even if mixed with parameters.
	 *   ◦ This can be disabled by ‹POSIXLY_CORRECT› environment variable or
	 *     adding ‹+› as the first character in ‹optstring›.
	 * • If the first character of ‹optstring› is ‹-›, positional parameters
	 *   are processed as parameters to a special option with value ‹1›.
	 * • Using ‹::› after a character creates an option with «optional» value.
	 *   This, however, only works with ‹-αVALUE› syntax; ‹-α VALUE› will
	 *   be treated as the option and un unrelated program parameter. */
	while ((option = getopt(argc, argv, optstring)) != -1) {
		switch (option) {
		/* Civilized programs often recognize at least ‹-h› (rarely ‹-?›)
		 * to display help. Sometimes also ‹-v› is used for version, but
		 * it is also commonly used to increase verbosity. */
		case 'h':
			help(argv[0]);
			/* Note that help in this case is «requested» by the user, so we
			 * should exit successfully, as we did what the user wanted. */
			exit(EXIT_SUCCESS);

		case 'a':
			options->a = true;
			break;

		case 'b':
			options->b = true;
			break;

		case 'c':
			/* This option requires an argument. It is stored in ‹optarg›. */
			options->c = optarg;
			break;

		/* If there is an unknown argument, POSIX ‹getopt()› returns:
		 * • If unknown option is encountered, returns ‹?›.
		 * • If option is missing an argument, then returns ‹?› by default
		 * 	 or ‹:› if ‹optstring› begins with ‹:›.
		 * This distinction allows us to differentiate these cases.
		 * We don't care about them here. */
		default:
			/* NOTE: ‹getopt()› and ‹getopt_long()› write an error message on
			 * their own. It is not necessary to add our own error message,
			 * we will just show the program synopsis. */
			usage(stderr, argv[0]);
			return false;
		}
	}

	/* If ‹struct options› included ‹argc› and ‹argv›, we can initialize
	 * them here. We will keep the convention that ‹argv[0]› is the program
	 * name and ‹argv[argc]› is ‹NULL› purely for consistency. This means
	 * we need to adjust ‹argv› and ‹argc› a little bit. */
	options->argc = argc - optind + 1;
	argv[optind - 1] = argv[0];
	options->argv = &argv[optind - 1];

	return true;
}

static
void options_dump(const struct options *options)
{
	printf("Program options:\n");
	printf("  [%c] a\n", options->a ? '*' : ' ');
	printf("  [%c] b\n", options->b ? '*' : ' ');
	printf("  [%c] c %s\n", options->c ? '*' : ' ', options->c);

	putchar('\n');

	printf("Program parameters:\n");
	for (int i = 0; i < options->argc; i++)
		printf("  [%02d] \"%s\"\n", i, options->argv[i]);
}

int main(int argc, char *argv[])
{
	/* In C99, initialization ‹{ 0 }› creates a structure filled with zeroes,
	 * no matter what actual types are in it. */
	struct options options = { 0 };

	if (!options_parse(argc, argv, &options))
		return EXIT_FAILURE;

	options_dump(&options);
	return EXIT_SUCCESS;
}
