#define _GNU_SOURCE
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>

#include <getopt.h>
#include <unistd.h>

/* This example is functionally the same as ‹ex01t›, so only the differences
 * will be commented. */

struct options {
	bool a;
	bool b;
	char *c;

	int argc;
	char **argv;
};

static
void usage(FILE *target, const char *pname)
{
	/* If option synopsis gets too long, POSIX Utility Guidelines allow
	 * them to be shortened to ‹OPTIONS›. Specifying the option that invokes
	 * help text is helpful in this case. */
	fprintf(target, "usage: %s [--help] [OPTIONS] [PARAMETER]...\n", pname);
}

static
void help(const char *pname)
{
	usage(stdout, pname);

	printf(
		/* When listing short and long options, parameters are only shown
		 * for long options, but are expected to be mandatory for short ones
		 * as well. Some utilities (mostly GNU) explicitly state this as
		 * “Mandatory arguments to long options are mandatory
		 * for short options too”. */
		"Example of getopt(3p) usage.\n"
		"\n"
		"Brief description of options and arguments:\n"
		"  -h, --help           Show this text and exit.\n"
		"  -a, --alpha          Set option a.\n"
		"  -b, --beta           Set option b.\n"
		"  -c, --gamma=ARG      Set option c to value ARG.\n"
		"\n"
		"  PARAMETER    A positional string\n"
	);
}

static
bool options_parse(int argc, char *argv[], struct options *options)
{
	static const char optstring[] = "habc:";

	/* Long options reside in an array of ‹struct option› records: */
	static const struct option longopts[] = {
		/* Mostly only two attributes are used, so we will use the
		 * ‹.attribute = value› syntax. */
		{ .name = "help", .val = 'h' },
		{ .name = "alpha", .val = 'a' },
		{ .name = "beta", .val = 'b' },

		/* To indicate a required argument, we also set ‹.has_arg› to
		 * one of the values:
		 * • 0 or ‹no_argument›
		 * • 1 or ‹required_argument›
		 * • 2 or ‹optional_argument›
		 * Using explicit names is preferred for clarity. */
		{ .name = "gamma", .val = 'c', .has_arg = required_argument },

		/* NOTE: The array «must» be terminated with a structure filled with
		 * zeros. This is where we can also use C99 initializer: */
		{ 0 },

#if 0
		/* If we were not using C99 attribute initialized syntax, we would
		 * have to rely on the order of attributes in the struct. We can omit
		 * those that are to be zero, but alas, ‹.val› is last, so the above
		 * array would be written as */

		{ "help", no_argument, NULL, 'h' },
		{ "alpha", no_argument, NULL, 'a' },
		{ "beta", no_argument, NULL, 'b' },
		{ "gamma", required_argument, NULL, 'c' },
		{ 0 }, // Never forget ‹{ 0 }›!

		/* As you can see, the above is not so great. */
#endif
	};

	int option;
	/* When ‹getopt_long()› encounters a long option, there are more ways it
	 * can communicate its value to the program:
	 *
	 *  1. Most often long options are aliased to some short option via their
	 *     ‹.val› attribute. In this case, ‹.flag› attribute must be set
	 *     to ‹NULL›, which causes ‹getopt_long()› to return value of ‹.val›.
	 *  2. If ‹.flag› is not ‹NULL›, then ‹getopt_long()› returns ‹0›
	 *     and sets the value of ‹.val› to the memory pointed to by ‹.flag›.
	 *  3. «In combination» of 1 and 2, if the last argument to ‹getopt_long()›
	 *     is not a ‹NULL› pointer, it is set to an index into ‹longopts›
	 *     for the entry. However, it is not reset if a short option is
	 *     encountered, so you should reset it to some invalid value (e.g. ‹-1›)
	 *     in each iteration.
	 *
	 * GNU ‹getopt_long()› has also an extension for ‹optstring›:
	 * • If ‹W;› (yes, semicolon) is included, then option ‹-W OPT[=VAL]› is
	 *   equivalent to ‹--OPT[=VAL]›. This is kind-of relying on the older
	 *   reserved option ‹-W› (see ‹ex01t›).
	 *
	 * The easiest way of using ‹getopt_long()› is the 1st option, which is
	 * displayed in this demo. */
	while ((option = getopt_long(argc, argv, optstring, longopts, NULL)) != -1) {
		switch (option) {
		case 'h':
			help(argv[0]);
			exit(EXIT_SUCCESS);

		case 'a':
			options->a = true;
			break;

		case 'b':
			options->b = true;
			break;

		case 'c':
			options->c = optarg;
			break;

		default:
			usage(stderr, argv[0]);
			return false;
		}
	}

	options->argc = argc - optind + 1;
	argv[optind - 1] = argv[0];
	options->argv = &argv[optind - 1];

	return true;
}

static
void options_dump(const struct options *options)
{
	printf("Program options:\n");
	printf("  [%c] a\n", options->a ? '*' : ' ');
	printf("  [%c] b\n", options->b ? '*' : ' ');
	printf("  [%c] c %s\n", options->c ? '*' : ' ', options->c);

	putchar('\n');

	printf("Program arguments:\n");
	for (int i = 0; i < options->argc; i++)
		printf("  [%02d] \"%s\"\n", i, options->argv[i]);
}

int main(int argc, char *argv[])
{
	struct options options = { 0 };

	if (!options_parse(argc, argv, &options))
		return EXIT_FAILURE;

	options_dump(&options);
	return EXIT_SUCCESS;
}
