#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/* This example demostrates common pitfalls with macro expansion.
 * The single rule of thumb is: There is never enough parenthesis with macros. */

/* The first example simply replaces ‹POW(X)› with ‹X * X›. */
#define POW1(X) \
	X * X

/* As we see from the command output, that doesn't work well. If we add
 * parentheses around expanded ‹X›, we can still find a counter-example. */
#define POW2(X) \
	(X) * (X)

/* The final macro is “the best” solution for standard C preprocesor.
 * It still, however, fails if we use a function in the argument. */
#define POW3(X) \
	((X) * (X))

/* If we allow some GCC extensions, we can use “statement expressions” feature
 * to solve the problem in a completely different way. */
#if __GNUC__ && !__STRICT_ANSI__
#   define HAS_POW4
#   define POW4(X) \
	({ typeof(X) tmp = (X); tmp * tmp; })
#else
#	warning "POW4 disabled"
#endif

/* However, ultimately, this example is just best solved with a function. */
long long llpow(long long n)
{
	return n * n;
}

/* Test expressions. */
#define EXPRESSIONS(EXPAND) \
	EXPAND(POW(5), llpow(5)) \
	EXPAND(POW(2 + 3), llpow(2 + 3)) \
	EXPAND(4 / POW(2), 4 / llpow(2)) \
	EXPAND(POW(rand() % 50), llpow(rand() % 50))

#define HEAD(STR) \
	printf("\x1b[1m" STR "\x1b[0m\n")

#ifdef SIMPLE
#define EXPR(T, E) \
	printf("%s = %d\n", #T, (T));
#else
#define EXPR(T, E) \
	do { \
		time_t seed = time(NULL); \
		srand(seed); \
		long long t = (T); \
		srand(seed); \
		long long e = (E); \
		if (t == e) { \
			printf("\x1b[92m ✔ \x1b[0m%s = %s\n", #T, #E); \
		} else { \
			printf("\x1b[91m ✘ \x1b[0m%s ≠ %s (%lld ≠ %lld)\n", #T, #E, t, e); \
		}\
	} while (0);
#endif

int main(void)
{
	HEAD("POW1()");
	#define POW POW1
	EXPRESSIONS(EXPR);
	putchar('\n');
	#undef POW

	HEAD("POW2()");
	#define POW POW2
	EXPRESSIONS(EXPR);
	putchar('\n');
	#undef POW

	HEAD("POW3()");
	#define POW POW3
	EXPRESSIONS(EXPR);
	putchar('\n');
	#undef POW

#ifdef HAS_POW4
	HEAD("POW4()");
	#define POW POW4
	EXPRESSIONS(EXPR);
	putchar('\n');
	#undef POW
#endif
}
