#include "strvec.h"

#include <assert.h>
#include <stdlib.h>

/* Initial capacity of the list. */
#ifndef STRVEC_INITIAL
#	define STRVEC_INITIAL 0
#endif

/* Increment for the capacity. */
#ifndef STRVEC_INCREMENT
#	define STRVEC_INCREMENT 1
#endif

/* If set to non-zero value, ‹NULL› pointer are allowed for ‹strvec_add()›. */
#ifndef STRVEC_ALLOW_NULL
#	define STRVEC_ALLOW_NULL 0
#endif

/* This macro allows us to modify ‹const› attributes. If used in functions
 * that pass ‹struct strvec› (not ‹const …›), then this operation is safe. */
#define rw(ATTRIBUTE) \
	(*((size_t *) &(ATTRIBUTE)))

static
struct strvec *_strvec_reserve(struct strvec *sv, size_t capacity)
{
	size_t total = sizeof(*sv) + sizeof(sv->data[0]) * capacity;
	/* ‹var = realloc(var, N)› is usually an antipattern, but we are working
	 * with a copy of the variable anyway here, so we leave the cleanup to
	 * the caller. */
	if ((sv = realloc(sv, total)) == NULL)
		return NULL;

	rw(sv->capacity) = capacity;
	return sv;
}

static
bool _strvec_grow(struct strvec **vecptr)
{
	struct strvec *vec = *vecptr;
	if (vec->size < vec->capacity)
		return true;

	if ((vec = _strvec_reserve(vec, vec->capacity + STRVEC_INCREMENT)) == NULL)
		return false;

	*vecptr = vec;
	return true;
}

bool strvec_create(struct strvec **vecptr)
{
	assert(vecptr != NULL);
	if ((*vecptr = _strvec_reserve(NULL, STRVEC_INITIAL)) == NULL)
		return false;

	rw((*vecptr)->size) = 0;
	return true;
}

void strvec_destroy(struct strvec **vecptr)
{
	assert(vecptr != NULL);
	/* ‹free(NULL)› is allowed, but we do not allow to create ‹NULL› strvec,
	 * therefore we also check for this case. */
	assert(*vecptr != NULL);

	free(*vecptr);
	*vecptr = NULL;
}

void strvec_clean(struct strvec **vecptr)
{
	strvec_for_each(*vecptr, (strvec_cb *) &free);
	rw((*vecptr)->size) = 0;
}

bool strvec_push(struct strvec **vecptr, char *str)
{
	assert(vecptr != NULL);
	assert(*vecptr != NULL);
#if !STRVEC_ALLOW_NULL
	assert(str != NULL);
#endif

	if (!_strvec_grow(vecptr))
		return false;

	struct strvec *vec = *vecptr;
	vec->data[rw(vec->size)++] = str;
	return true;
}

char *strvec_pop(struct strvec **vecptr)
{
	assert(vecptr != NULL);
	assert(*vecptr != NULL);

	struct strvec *vec = *vecptr;

	/* Note that if ‹STRVEC_ALLOW_NULL› is set, this case must be differentiated
	 * by the caller. */
	if (vec->size == 0)
		return NULL;

	rw(vec->size)--;
	return vec->data[vec->size];
}

void strvec_for_each(const struct strvec *vec, strvec_cb *cb)
{
	assert(vec != NULL);
	assert(cb != NULL);

	for (size_t i = 0; i < vec->size; i++)
		cb(vec->data[i]);
}
