#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <err.h>
#include <error.h>
#include <fcntl.h>
#include <unistd.h>

/* Some demo macros, which you may safely skip. */
#define DEMOV(EXPR) \
	printf("%2d  %s\n", __LINE__, #EXPR), (EXPR)

#define DEMOP(EXPR) \
	printf("%2d  %s → ", __LINE__, #EXPR), (EXPR)

#define DEMOI(EXPR) \
	({ int n; printf("%2d  %s → %d\n", __LINE__, #EXPR, n = (EXPR)); n; })

int main(void)
{
	/* errno
	 * ----- */

	/* We have already seen ‹errno› a few times, so a quick recap:
	 *
	 *   ‹errno› is a (pseudo)variable¹ that is set to a non-zero value
	 *   when a standard library fails.
	 *
	 * Two points need to be always kept in mind, though:
	 *
	 *  1. It is «not» cleared when a function succeeds.
	 *  2. It only makes sense when a function fails. Specifically,
	 *     successful run of a function «may» change ‹errno›.
	 *
	 * ¹ It may actually be something else, as we will see when we will be
	 *   discussing threads. */

	/* Let's see how it works in action. First, let's make sure it's 0. */
	errno = 0;

	/* Then we call a function that is guaranteed to fail. */
	DEMOI(unlink("/proc/self/exe"));

	/* The function «must» have failed, so let's see if it changed errno: */
	printf("\terrno: %d\n\tstr: %s\n", errno, strerror(errno));

	/* In the code above, ‹strerror()› translated the error code to
	 * a human-readable description of the error.
	 *
	 * See also ‹man 1 errno›. */

	/* Let's try some other error. */
	DEMOI(unlink("/proc/self"));
	printf("\terrno: %d\n\tstr: %s\n", errno, strerror(errno));

	/* Let's demonstrate that «successful» run does not change ‹errno›: */
	int fd = DEMOI(open("/dev/random", O_RDONLY));
	printf("\terrno: %d\n\tstr: %s\n", errno, strerror(errno));
	DEMOI(close(fd));
	printf("\terrno: %d\n\tstr: %s\n", errno, strerror(errno));

	/* Up until this point, ‹errno› should still be non-zero despite all
	 * functions above succeeding. */
	assert(errno != 0);


	/* Error Reporting
	 * --------------- */

	/* With ‹errno› set, let's discuss various error reporting functions
	 * that we have available. */

	/* The most basic one in C standard is ‹perror()›: */
	DEMOV(perror("unlink"));

	/* This function takes the argument, adds ‹": "› and then description
	 * of the error. Effectively it's the same as the following: */
	DEMOV(fprintf(stderr, "%s: %s\n", "unlink", strerror(errno)));
	/* … except much shorter and more convenient. */


	/* POSIX adds four more handy reporting functions:
	 *
	 * ‹err(X, FMT, ...)› prints the formatted error string, followed by
	 *     the error message, and then calls ‹exit(X)›.
	 * ‹warn(FMT, ...)› prints the formatted error string, followed by
	 *     the error message.
	 *
	 * ‹errx()› and ‹warnx()› are the same as their x-less equivalent except
	 *     they do not print the error message.
	 *
	 * These messages may be prefixed with program name. */

	/* So, the above example is roughly equivalent to */
	DEMOV(warn("unlink"));

	/* Note that hte ‹*x› version skips the error code: */
	DEMOV(warnx("unlink"));

	/* However, aside from ‹perror()›, this function allows us to use format
	 * strings like ‹printf()›: */
	DEMOV(warn("Cannot unlink %s", "/proc/self/exe"));

	/* Showing ‹err()› would exit the program, which we do not want yet.
	 * Instead, we will look at one more function. */


	/* GNU error
	 * --------- */

	/* The ‹error()› function combines all four POSIX functions:
	 *
	 *      err(X, FMT...)   → error(X, errno, FMT...)
	 *      errx(X, FMT...)  → error(X, 0, FMT...)
	 *      warn(FMT...)     → error(0, errno, FMT...)
	 *      warnx(FMT...)    → error(0, 0, FMT...)
	 *
	 * Let's demonstrate this on the same example as above first: */
	DEMOV(error(0, errno, "unlink %s", "/proc/self"));

	/* Finally, we will exit the program with the ‹error()› function. */
	DEMOV(error(EXIT_FAILURE, errno, "unlink %s", "/proc/self"));

	/* This code will not execute: */
	abort();

	/* You may be interested in ‹patterns/01-defensive-offensive-programming.adoc›
	 * to learn about two categories of errors, where the above functions
	 * might be helpful, and where you should avoid them. */
}
