#define _GNU_SOURCE
#include <assert.h>
#include <errno.h>
#include <stdbool.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>

#include <err.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>

/* This program showcases an error in process privilege management.
 *
 * There are comments provided for every function in this code, you are advised
 * to read them before attempting to solve the problem. */

/* NOTE:
 * The function ‹_dump()› and the macro ‹DUMP()› will dump current process
 * User IDs. Use it to debug how privileges change. Refer to the graph
 * in slides to see what numbers should change and how. */
static
void _dump(const char *file, int line)
{
	uid_t ruid, euid, suid;

	if (getresuid(&ruid, &euid, &suid) == -1)
		err(EXIT_FAILURE, "Cannot determine process user IDs");

	printf("%s:%d:\tRUID %5d\t| EUID %5d\t| Saved-UID %5d\n", file, line,
			ruid, euid, suid);
}

#define DUMP() \
	_dump(__FILE__, __LINE__)

/* Auxiliary functions
 * ------------------- */

static
void run_external_command(void)
{
	printf("\n--- >8 --- external command ouput --- >8 ---\n");
	system("./resuid");
	printf("--- >8 --- --- >8 ---  --- >8 --- --- >8 ---\n\n");
}

/* Privileges manipulation
 * ----------------------- */

/* This function tries to detect whether we are privileged process or not.
 *
 * The detection relies on the fact that only a privileged process can set
 * value 0 as its EUID, either because it already has EUID = 0 or 0 is stored
 * in RUID or Saved-UID. We could theoretically have ‹CAP_SETUID› capability,
 * but we have not covered capabilities at this point yet. */
static
bool is_privileged(void)
{
	/* This function is a predicate, so we will save the current EUID
	 * to restore it before the function ends. */
	uid_t original = geteuid();

	/* Only a privileged process can do this. */
	if (seteuid(0) == -1)
		return false;

	/* Try to restore EUID (this is a predicate after all). */
	if (seteuid(original) == -1) {
		/* The above «can» fail under some circumstances, but if it does,
		 * there is no way to reasonably continue. */
		err(EXIT_FAILURE, "is_privileged: Failed to restore original privileges");
	}

	return true;
}

/* This function drops process privileges temporarily.
 *
 * For Set-UID program, this means that the process will have RUID ≠ 0
 * and EUID ≠ 0. Only Saved-UID will retain 0, but it's never checked in any
 * privileged operation, so effectively the process will be non-privileged.
 * Also, any ‹exec()› will copy EUID to Saved-UID, so that new process will
 * never be able to become elevated again.
 *
 * However, it's a «temporary» drop, because we can still restore EUID = 0
 * from Saved-UID inside currently running process. */
static inline
bool privilege_push(void)
{
	return seteuid(getuid()) == 0;
}

/* For completion's sake, we include a function that restores the original
 * privileges. This function relies on the fact that Saved-UID still contains
 * value 0 at this point. If it doesn't, we have lost priviliges completely. */
static
bool privilege_pop(void)
{
	return seteuid(0) == 0;
}

/* The function that drops all privileges completely, once and for all. */
static
bool privilege_drop(void)
{
	/* TODO:
	 * This function gets called with temporarily dropped privileges (EUID
	 * is set to non-zero). We should still have 0 in Saved-UID. The goal here
	 * is to drop privileges completely.
	 *
	 * However, the following does not work. What should we do instead? */
	return setuid(getuid()) == 0;
}

/* Main
 * ---- */

static
void require_setuid(void)
{
	if (getuid() != 0 && geteuid() == 0)
		return;

	fprintf(stderr, "The process did not start with correct Set-UID privileges.\n");
	fprintf(stderr, "It should have RUID ≠ 0 (is %d) and EUID = 0 (is %d).\n",
			getuid(), geteuid());
	fprintf(stderr, "See the assignment and set up the program as required.\n");
	exit(EXIT_FAILURE);
}

int main(int argc, char *argv[])
{
	/* The program must start as a privileged process. */
	require_setuid();
	/* If we get here, then we for sure have
	 *
	 *  • RUID ≠ 0 and it contains the ID of the user who started this program.
	 *  • EUID = 0 because the binary is Set-UID,
	 *  • Saved-UID = 0 because it's copied from EUID on ‹exec()›. */

	printf("Privileges when started:\n");
	DUMP();

	/* Now let's say we want to perform some operation on behalf of the user.
	 * If we want to run some external application, we cannot leave UIDs
	 * unchanged, because they would be inherited by ‹fork()› and ‹exec()›.
	 * We need to set EUID ← RUID, so that EUID will be copied to Saved-UID
	 * upon ‹exec()›, but we can still restore EUID from Saved-UID in our
	 * process. */

	if (!privilege_push())
		err(EXIT_FAILURE, "Cannot drop privileges temporarily");

	/* Demonstrate that transit through ‹execve()› will drop privileges
	 * completely for the new process. */
	run_external_command();

	/* At this point, EUID ≠ 0, so system calls do not see our process as
	 * privileged. But we could still set EUID to Saved-UID, which contains 0,
	 * so we are “potentially privileged” still. */
	assert(is_privileged());

	/* Now we attempt to drop privileges completely. */
	if (!privilege_drop())
		errx(EXIT_FAILURE, "Attempt to drop privileges failed somehow");

	/* Did it work? */
	if (is_privileged())
		errx(EXIT_FAILURE, "FAIL: The process did not drop the elevated privileges correctly");

	puts("Success!");
	return EXIT_SUCCESS;
}
