#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <error.h>
#include <syslog.h>
#include <unistd.h>

/* These macros were already explained previously. */
#define _unused __attribute__((unused))
#define fatal(FMT, ...) \
	do { \
		fprintf(stderr, FMT "\n", ## __VA_ARGS__); \
		abort(); \
	} while (0)

/* The program presented here uses ‹pause()› to wait for a signal. It also
 * demonstrates a few problems with standard signals. */

/* The counter variable will need to be accessed from signal handler. Thus,
 * ‹volatile› and ‹sig_atomic_t› are recommended for the type.
 *
 * Note, however, that this does «not guarantee» that the program will
 * behave correctly. It just removes one potential synchronization problem. */
volatile sig_atomic_t sig_received;

/* This function gets called when a signal gets received. We do not really
 * care about the signal in the argument, we just set the ‹sig_received›. */
void handler(_unused int signum)
{
	sig_received++;
}

int main(void)
{
	openlog(NULL, LOG_PERROR|LOG_PID, 0);
	syslog(LOG_INFO, "Started %d", getpid());

	/* When creating an instance of ‹struct sigaction›, always prefer C99
	 * initialization so that unused fields are zeroed out.
	 *
	 * Do «not» do this:
	 *
	 * 		struct sigaction action;
	 * 		action.sa_handler = handler;
	 *
	 * unless you also initialize at least ‹sa_flags› and ‹sa_mask›. */
	struct sigaction action = {
		.sa_handler = handler,
	};

	/* This function should not fail for our arguments. Thus, if it does
	 * fail, we have a bug in the program. */
	if (sigaction(SIGINT, &action, NULL) == -1)
		fatal("Cannot set up signal handler: %s", strerror(errno));

	while (1) {
		/* Here process stops until a signal is delivered. */
		pause();

		if (sig_received > 0)
			syslog(LOG_INFO, "SIGINT received: %d", sig_received);

		/* We clean up the counter and sleep for a second. */
		sig_received = 0;
		sleep(1);
	}

	/* As you have probably noticed, we use ‹SIGINT›, so we cannot possibly
	 * exit the program using ‹^C› key. Instead, you will have to
	 * use ‹kill $PID› or ‹kill -KILL $PID›. */

	/* Hence, we shall never reach this code. Bummer. */

	/* Now, try running the program above. Notice anything weird in the output?
	 * How can you get more than 2 signals reported by the ‹syslog()› line? */

	closelog();
	return EXIT_SUCCESS;
}
