#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <error.h>
#include <syslog.h>
#include <unistd.h>

/* These macros were already explained previously. */
#define _unused __attribute__((unused))
#define fatal(FMT, ...) \
	do { \
		fprintf(stderr, FMT "\n", ## __VA_ARGS__); \
		abort(); \
	} while (0)

/* The program presented here works, technically, almost exactly as
 * ‹ex02t›, except uses ‹sigsuspend()› instead of ‹pause()›. */
volatile sig_atomic_t sig_received;

void handler(_unused int signum)
{
	sig_received++;
}

int main(void)
{
	openlog(NULL, LOG_PERROR|LOG_PID, 0);
	syslog(LOG_INFO, "Started %d", getpid());

	/* We will need to swap between a mask with SIGINT and the original
	 * mask without this signal. */
	sigset_t sig_intset, sig_oldset;

	/* These operations basically create ⟦S ← Ø⟧ and ⟦S ← S ∪ {SIGINT}⟧.
	 * Note that these functions «could» in theory fail if we provide
	 * invalid signal, but to keep the code short, we can reasonably assume
	 * that the pathological case will never happen. */
	sigemptyset(&sig_intset);
	sigaddset(&sig_intset, SIGINT);

	/* From now on, SIGINT will not be delivered, because it's masked
	 * by ‹sig_intset›. */
	if (sigprocmask(SIG_BLOCK, &sig_intset, &sig_oldset) == -1)
		fatal("sigprocmask: %s", strerror(errno));

	struct sigaction action = {
		.sa_handler = handler,
	};

	if (sigaction(SIGINT, &action, NULL) == -1)
		fatal("sigaction: %s", strerror(errno));

	while (1) {
		/* This call temporarily sets the original signal mask. If ‹SIGINT›
		 * gets delivered, the ‹sig_intset› will get restored into the signal
		 * mask and ‹sigsuspend()› returns with error (Interrupted).
		 *
		 * We don't need to check for that; this function cannot exit
		 * successfully. */
		sigsuspend(&sig_oldset);

		/* At this point, SIGINT is masked again. */

		if (sig_received > 0)
			syslog(LOG_INFO, "SIGINT received: %d", sig_received);

		sig_received = 0;
		sleep(1);
	}

	/* Again, this code is unreachable.
	 *
	 * This time, the code behaves, at the first glance, even more weird.
	 * Why is that the counter never reaches 2? */

	closelog();
	return EXIT_SUCCESS;
}
