#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <err.h>
#include <signal.h>
#include <unistd.h>

/* This example shows how ‹abort()› insists on killing the process that
 * calls it. */

/* This auxiliary function creates a singleton set of the provided signal. */
static inline
sigset_t sigset_singleton(int signal)
{
	sigset_t res;
	sigemptyset(&res);
	int rv = sigaddset(&res, signal);
	/* ‹sigaddset()› can fail for unknown signal. This would indicate a bug
	 * in our program, thus we use ‹assert()›.
	 *
	 * Note that you «should not» call functions inside ‹assert()›:
	 *
	 * 		assert(sigaddset(&res, signal) == 0);
	 *
	 * is «WRONG», because assertions get disabled via ‹-DNDEBUG› flag,
	 * it would also disable ‹sigaddset()› itself. */
	assert(rv == 0);
	return res;
}

/* This signal handler writes a message about what signal is delivered,
 * but ignores it. */
static
void signal_nop(int signal)
{
	/* Using ‹printf()› family of functions in a signal handler is, strictly
	 * speaking, not ideal, as these functions may access global buffers.
	 * You can «reasonably» use them if you make sure the handler should not
	 * run concurrently (e.g. by setting masks).
	 *
	 * This example uses only functions listed in ‹man 7 signal-safety›
	 * (that is, only ‹write()›) to print output to show you how painful
	 * it is to live without formatting functions. */
	const char msg[] = "Signal ignored: ";
	write(STDOUT_FILENO, msg, strlen(msg));

	/* This is a trick to quickly convert some fixed integer value to
	 * a hexadecimal string. It splits the number into groups of four bits
	 * and then adds ‹'0'› or ‹'a'› which are at the same time integer constants
	 * (you read that right, ‹'0'› is of type ‹int›, not ‹char›).
	 *
	 * 		N ≈ aaaabbbb ccccdddd
	 *
	 * 	        → 0xabcd
	 *
	 * These values are then directly placed into an array that can be passed
	 * to ‹write()›.
	 *
	 * Also note that since we access bits with bit operators (not directly
	 * in memory), we do not need to care about endianity. */
#define HEX_DIGIT(N) \
	((N) < 10 ? '0' + (N) : 'a' + (N))

	char sig_str[] = {
		'0', 'x',
		HEX_DIGIT((signal & 0xf000) >> 12),
		HEX_DIGIT((signal & 0x0f00) >>  8),
		HEX_DIGIT((signal & 0x00f0) >>  4),
		HEX_DIGIT(signal & 0x000f),
	};

	write(STDOUT_FILENO, sig_str, sizeof(sig_str));

	char nl = '\n';
	write(STDOUT_FILENO, &nl, sizeof(char));

	/* If you were to implement formatting of decimal numbers here without
	 * ‹printf()›-like function, it would be a bit more fun. */
}

int main(void)
{
	/* We will try two things to protect the process from ‹abort()›:
	 * 1. Mask the signal. */
	sigset_t sig_abrt = sigset_singleton(SIGABRT);
  	if (sigprocmask(SIG_BLOCK, &sig_abrt, NULL) != 0)
  		err(EXIT_FAILURE, "sigprocmask");

	/* 2. Set a signal handler that does nothing. */
	struct sigaction sa_abrt = {
		.sa_handler = &signal_nop,
	};

	if (sigaction(SIGABRT, &sa_abrt, NULL) != 0)
		err(EXIT_FAILURE, "sigaction");

	printf("%d Ready\n", getpid());

	/* First, we will demonstrate that ‹raise(SIGABRT)› really does nothing. */
	printf("1: raise(SIGABRT)\n");
	raise(SIGABRT);

	/* We then call ‹abort()› to show it's magic. */
	printf("2: abort()\n");
	abort();

	/* The call above does the following:
	 *
	 *  1. First, ‹SIGINT› is unblocked. This causes the delivery of the signal
	 *     caused by ‹raise()›.
	 *  2. Now ‹abort()› tries ‹raise(SIGINT)› again. This delivers the signal,
	 *     but our handler prints the second message and ignores the signal.
	 *  3: ‹abort()› continues; at this point it tries even harder by
	 *     setting the default signal disposition, ‹signal(SIGINT, SIG_DFL)›.
	 *     Then it tries ‹raise(SIGINT)› again, which now succeeds in killing
	 *     the process.
	 *
	 * Long story short, you could implement your own ‹abort()› similarly
	 * to this:
	 *
	 * 		sigprocmask(SIG_UNBLOCK, [SIGINT set], NULL);
	 *		raise(SIGABRT);
	 *		signal(SIGINT, SIG_DFL);
	 *		raise(SIGABRT);
	 *
	 * The actual ‹abort()› implementation is much more complicated though:
	 * https://sourceware.org/git/?p=glibc.git;a=blob;f=stdlib/abort.c */

	return EXIT_SUCCESS;
}
