#define _GNU_SOURCE

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <error.h>
#include <fcntl.h>
#include <unistd.h>
#include <seccomp.h>
#include <sys/wait.h>

/* This example demonstrates use of SecComp system call and SecComp library
 * built on top of that call. */

/* The high-level idea of SecComp is that it provides an ability to audit
 * system calls. This is done by creating a set of filters (a “policy”)
 * which, once installed, immediately takes effect and cannot be removed.
 *
 * These filters are Berkeley Packet Filter (BPF) programs which allow
 * very simple expressions to be created, e.g. “intercept a system call
 * ‹open()› if the flags argument contains ‹O_WR›”. Once a system call
 * is intercepted, there are several ways to dispose of this intercept:
 *
 * • The process may be killed without mercy.
 * • The call may be forced to fail with a chosen error code, e.g. ‹ENOSYS›
 *   or ‹EPERM›.
 * • The call may be permitted to continue.
 * • The process may be stopped for the tracer to inspect the situation
 *   (this is called a “trap”). */

/* Let's see how it's done. */

static
void setup_seccomp(void)
{
	int rc;

	/* The call to ‹seccomp_init()› creates a new filter state.
	 * The argument tells the “default policy”. In this case, we allow all
	 * system calls by default, and we will add calls that are forbidden;
	 * i.e. blocklisting. */
	scmp_filter_ctx ctx = seccomp_init(SCMP_ACT_ALLOW);

	if (ctx == NULL)
		error(EXIT_FAILURE, errno, "seccomp_init");

	/* We will add a lot of system calls. In a security-sensitive programs where
	 * we need SecComp we should never ignore system call return values,
	 * otherwise we might botch the setup and let the program run with
	 * some system unintended system calls available.
	 *
	 * However, checking every system call like this can be a bit tedious.
	 * We could use X-macros, but in this case we could possibly want to
	 * add more arguments. This is an excuse to create a macro that is not
	 * exactly clean (uses ‹ctxt› and ‹goto›), but hides these repetitious
	 * details well. We will only use it inside this function, so it should
	 * be fine. */
#define SECCOMP_ADD(...) \
	do { \
		rc = seccomp_rule_add(ctx, ## __VA_ARGS__); \
		if (rc < 0) \
			goto seccomp_error; \
	} while (0)

	/* We block a few file system operations that can be harmful. These
	 * are blocked unconditionally -- the ‹0› in the end tells SecComp
	 * that there are zero argument filters. */
	SECCOMP_ADD(SCMP_ACT_ERRNO(EPERM), SCMP_SYS(unlink), 0);
	SECCOMP_ADD(SCMP_ACT_ERRNO(EPERM), SCMP_SYS(unlinkat), 0);
	SECCOMP_ADD(SCMP_ACT_ERRNO(EPERM), SCMP_SYS(rmdir), 0);
	SECCOMP_ADD(SCMP_ACT_ERRNO(EPERM), SCMP_SYS(truncate), 0);
	SECCOMP_ADD(SCMP_ACT_ERRNO(EPERM), SCMP_SYS(ftruncate), 0);

	/* The following example is a bit more comples. The ‹1› after
	 * ‹SCMP_SYS(open)› says that there is one argument test employed.
	 * This test is a masked equality test:
	 *
	 *		SCMP_CMP(ι, SCMP_CMP_MASKED_EQ, μ, τ)
	 *
	 * The test succeeds if masked (μ) value of ι-th argument is equal
	 * to masked value of τ, i.e. ‹(ARG[ι] & μ) == (τ & μ)›. This means…*/
	SECCOMP_ADD(SCMP_ACT_ERRNO(EPERM), SCMP_SYS(open), 1,
			/* … that this test intercepts calls where the first argument
			 * (flags) has ‹O_WRONLY› bit set. */
			SCMP_CMP(1, SCMP_CMP_MASKED_EQ, O_WRONLY, O_WRONLY));

	/* We will also block ‹openat()› which is sometimes used
	 * instead of ‹open()›. In this case, the flag is the third argument. */
	SECCOMP_ADD(SCMP_ACT_ERRNO(EPERM), SCMP_SYS(openat), 1,
			SCMP_CMP(2, SCMP_CMP_MASKED_EQ, O_WRONLY, O_WRONLY));

	/* We no longer need the ‹SECCOMP_ADD()› macro. */
#undef SECCOMP_ADD

	/* Up until now, nothing above has actually taken effect. We change it
	 * by the following call, which installs and enforces the created policy.
	 * Once this is done, the policy cannot be removed. The process can,
	 * however, add more policies unless ‹prctl()› and ‹seccomp()› are
	 * also blocked. */
	if ((rc = seccomp_load(ctx)) < 0)
		goto seccomp_error;

	/* If the policy is installed, remove it from the memory. */
	seccomp_release(ctx);
	return;

seccomp_error:
	error(0, errno, "failed to set up seccomp");
	seccomp_release(ctx);
	exit(EXIT_FAILURE);
}

/* The rest of the program simply calls ‹setup_seccomp()› and then runs
 * a program specified as the arguments.
 *
 * Try, for example, to compare
 *
 * 		$ bash -c 'echo foo >file'
 *		$ seccomp bash -c 'echo foo >file'
 *
 * In the second case you should see an error, because the Bash's attempt
 * to open the file for writing should fail. */

static
void usage(const char *argv0)
{
	fprintf(stderr, "usage: %s PROGRAM [PARAMETERS]\n", basename(argv0));
}

int main(int argc, char *argv[])
{
	if (argc <= 1) {
		usage(argv[0]);
		return EXIT_FAILURE;
	}

	setup_seccomp();
	execvp(argv[1], argv + 1);
	error(EXIT_FAILURE, errno, "failed to exec %s", argv[1]);
}
