/* This demo shows how to use ‹ptrace()› to inspect system calls made
 * by some program. It is quite a bit complicated, but we will try to
 * decompose it bit by bit. */

#include <errno.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <err.h>
#include <sys/prctl.h>
#include <sys/syscall.h>
#include <sys/user.h>
#include <sys/wait.h>
#include <unistd.h>

/* First of all, some “nicer” features of ‹ptrace()› that allow us to
 * directly inspect system calls are only available since Linux 5.3.
 * Before this version, we have to inspect registers of the process.
 *
 * ‹LINUX_VERSION_CODE› is a number that uniquely identifies Linux Kernel
 * version. ‹KERNEL_VERSION()› on the other hand creates such number
 * from integer literals, so we simply compare the two and enable the
 * feature if we have the newer kernel. */

#include <linux/version.h>
#ifndef CONFIG_USE_PTRACE_SYSCALL_INFO
#	define CONFIG_USE_PTRACE_SYSCALL_INFO \
		(LINUX_VERSION_CODE >= KERNEL_VERSION(5, 3, 0))
#endif

/* The ‹ptrace()› interface lives in ‹sys/ptrace.h›. However, if we want
 * to use the modern features, we also need ‹linux/ptrace.h›. */
#include <sys/ptrace.h>
#if CONFIG_USE_PTRACE_SYSCALL_INFO
#  include <linux/ptrace.h>
#endif


/* Auxiliary functions
 * ------------------- */

static
void usage(const char *argv0)
{
	fprintf(stderr, "usage: %s PROGRAM [ARGS...]\n", basename(argv0));
}


/* Child process (the tracee)
 * -------------------------- */

/* This part is being run exactly after ‹fork()›. The task here is to prepare
 * the process to be traced by the parent. This is, so far, quite simple. */

static
void run_tracee(char *path, char *argv[])
{
	/* Tell the kernel we want to be traced. */
	if (ptrace(PTRACE_TRACEME, 0, NULL, NULL) == -1)
		err(EXIT_FAILURE, "%d: Failed to set up tracing", getpid());

	/* This will deliver ‹SIGSTOP› to the child and immediately stops it.
	 * This is purely to avoid potential race condition, where the child
	 * could ‹exec()› sooner than the parent catches up, potentially starting
	 * to run some untrusted code. */
	if (raise(SIGSTOP) != 0)
		err(EXIT_FAILURE, "%d: Failed to stop", getpid());

	/* Once the child process resumes, it is because the parent restarted it,
	 * now fully paying attention to the child. We execute the program. */
	execvp(path, argv);
	err(EXIT_FAILURE, "%d: exec '%s'", getpid(), argv[1]);
}


/* Parent process (the tracer)
 * --------------------------- */

/* We will use the following status codes for more descriptive code. */
enum process_syscall_status {
	PSS_NOP,
	PSS_ERROR,
	PSS_SYSCALL_ENTER,
	PSS_SYSCALL_LEAVE,
};

/* This function reacts to the system call trap.
 *
 * When we trace a process from the tracer's perspective, there are several
 * modes in which we can tell kernel when to stop the tracee:
 * • By default, tracee always stops when a signal is about to be delivered
 *   to it.
 * • Using ‹PTRACE_SYSCALL› from the tracer, kernel runs the tracee until
 *   it enters or exists a syscall.
 * • Similarly, ‹PTRACE_SINGLESTEP› executes one instruction in tracee.
 *
 * For this demo program, we use ‹PTRACE_SYSCALL› (below) to check system calls. */
enum process_syscall_status process_syscall(pid_t child)
{
#if CONFIG_USE_PTRACE_SYSCALL_INFO
	/* This version uses newer ‹PTRACE_GET_SYSCALL_INFO› command. */
	struct ptrace_syscall_info syscall_info;

	/* The command inspects the child provided by the second argument and fills
	 * the structure ‹syscall_info› with various information about the syscall.
	 * Obviously we expect that the process stopped due to trap at the syscall
	 * boundary, that is, we used ‹PTRACE_SYSCALL›. */
	if (ptrace(PTRACE_GET_SYSCALL_INFO, child, sizeof(syscall_info), &syscall_info) == -1) {
		warn("%d: Cannot get syscall info", getpid());
		return PSS_ERROR;
	}

	/* The ‹op› attribute contains either ‹PTRACE_SYSCALL_INFO_ENTRY›,
	 * ‹…_EXIT› or ‹…_SECCOMP›. The first two are obvious, the last is caused
	 * by ‹seccomp()› when ‹PTRACE_O_TRACESECCOMP› is set by the tracer.
	 * We are not using that here, though, but it allows easy integration of
	 * seccomp and ptrace.
	 *
	 * There is also ‹PTRACE_SYSCALL_INFO_NONE› which is set when there is
	 * no information about the syscall. This can happen for example by
	 * unfortunate interaction of several settings. */
	if (syscall_info.op == PTRACE_SYSCALL_INFO_EXIT) {
		/* The structure contains, for example, the return value of the system
		 * call. If we want, we could even change it, though it is a bit
		 * more complicated. */
		printf("%d: Leave rc=0x%llx\n", getpid(), syscall_info.exit.rval);
		return PSS_SYSCALL_LEAVE;
 	}

	if (syscall_info.op == PTRACE_SYSCALL_INFO_ENTRY) {
		/* On entry, we can inspect what system call is being invoked.
		 * We do not get the symbolic name, only the number. */
		printf("%d: Enter %lld\n", getpid(), syscall_info.entry.nr);

		switch (syscall_info.entry.nr) {
		/* When the child attempts ‹fork()› or a similar syscall, the tracer
		 * should set up tracing of the new process as well. However, in this
		 * simplified demo, we will not do that and instead we will kill the
		 * child if it attempts these calls. */
		case SYS_fork:
		case SYS_vfork:
		case SYS_clone:
			warnx("%d: Tracee created a new process", getpid());
			return PSS_ERROR;

		default:
			return PSS_SYSCALL_ENTER;
		}
	}

	return PSS_NOP;

#else

	/* In kernel before 5.3.0, we cannot use ‹PTRACE_GET_SYSCALL_INFO›.
	 * We have to inspect registers of the process that is entering the system
	 * call on our own.
	 *
	 * This brings its own problems, obviously -- we have to distinguish
	 * 32 and 64 bit programs. For simplicity we are going to assume 64 bit
	 * platform. */

	/* We will create a static flip-flop variable that tells us whether we
	 * are entering or leaving the syscall. Obviously a multi-threaded or
	 * multi-process task would immediately render this solution incorrect.
	 * Thus, we will not allow the tracee to create new processes. */
	static char syscall_enter = 1;

	/* This structure comes from ‹sys/user.h›, where it is fairly undocumented.
	 * The header states that it is intended to be used by GDB only, which
	 * makes sense -- we are essentially in the area of debuggers right now. */
	struct user_regs_struct regs;

	/* ‹PTRACE_GETREGS› reads registers of the tracee and puts them in ‹&regs›. */
	if (ptrace(PTRACE_GETREGS, child, 0, &regs) == -1) {
		warn("%d: Cannot read registers", getpid());
		return PSS_ERROR;
	}

	if (syscall_enter) {
		/* System call number is usually passed in the ‹rax› register,
		 * which we can inspect here. However, under some circumstances,
		 * ‹rax› gets overwritten, for example when the system call returns.
		 * The original ‹rax› value is therefore stored in ‹orig_rax› attribute. */
		switch (regs.orig_rax) {
		case SYS_fork:
		case SYS_vfork:
		case SYS_clone:
			warnx("%d: Tracee created a new proces", getpid());
			return PSS_ERROR;

		default:
			printf("%d: Enter %llu\n", getpid(), regs.orig_rax);
			syscall_enter = (syscall_enter + 1) % 2;
			return PSS_SYSCALL_ENTER;
		}
	}

	printf("%d: Leave (0x%llx)\n", getpid(), regs.rax);
	syscall_enter = (syscall_enter + 1) % 2;
	return PSS_SYSCALL_LEAVE;
#endif
}

/* This wrapper function deals with notifications from ‹ptrace()› about
 * the tracee. */
static
size_t count_syscalls(pid_t child)
{
	size_t counter = 0;

	/* Notifications are delivered via ‹wait*()› interface, even in case
	 * the process isn't the child of the tracer (e.g. ‹PTRACE_ATTACH›).
	 * Linux actually does some magic to pretend the tracee is a child
	 * of tracer for some time. */
	int wstatus;

	while (1) {
		if (waitpid(child, &wstatus, 0) == -1) {
			warn("%d: waitpid %d", getpid(), child);
			return counter;
		}

		/* This case we already know, nothing new here. */
		if (WIFEXITED(wstatus)) {
			printf("Child %d exited with status %d\n", getpid(), WEXITSTATUS(wstatus));
			break;
		}

		/* When tracing is enabled, a new status can be observed: ‹WIFSTOPPED()›.
		 * This case happens when the tracee is stopped due to some event.
		 * For example, we can set up tracing so that the tracee is stopped
		 * every time it enters a system call. This will cause the tracee to be
		 * stopped, and from the tracer's perspective it will appear that the
		 * tracee has been delivered ‹SIGTRAP› signal. In addition, tracee
		 * is always stop if any signal other than ‹SIGKILL› is delivered.
		 *
		 * Remember from comment for ‹PTRACE_O_TRACESYSGOOD› that system call
		 * traps have 8th bit set. */
		if (WIFSTOPPED(wstatus) && WSTOPSIG(wstatus) == (SIGTRAP | 0x80)) {
			switch (process_syscall(child)) {
			case PSS_ERROR:
				return counter;

			case PSS_SYSCALL_ENTER:
				++counter;
				break;

			default:
				/* nop */
				break;
			}
		}

		/* This call will resume the tracee until another signal gets delivered
		 * (which is always the default) or until it reaches a system call
		 * boundary. */
		if (ptrace(PTRACE_SYSCALL, child, 0, NULL) == -1) {
			warnx("%d: Failed to resume child", getpid());
			return counter;
		}
	}

	return counter;
}

/* The ‹run_tracer()› function sets up tracing and then passes control
 * to ‹count_syscalls()› above. */
static
size_t run_tracer(pid_t child)
{
	printf("%d: Tracee PID is %d\n", getpid(), child);

	int wstatus;

	/* At this point, we are not synchronized with the tracee, which could be
	 * behind with its own setup. If we tried to use ‹PTRACE_SETOPTIONS›
	 * here, the tracee might not yet be ready to be traced, and the call would
	 * fail. This synchronization is therefore solved by the tracee raising
	 * ‹SIGSTOP› after it sets up the tracing. When we observe this signal,
	 * we know the child is ready for tracing to be configured in the tracer. */
	if (waitpid(child, &wstatus, 0) == -1)
		err(EXIT_FAILURE, "%d: waitpid %d", getpid(), child);

	/* Verify that the child has indeed stopped, and not died e.g. due to
	 * incorrect parameters. */
	if (!WIFSTOPPED(wstatus)) {
		warnx("%d: Tracee had failed before tracing could start", getpid());
		goto kill_child_and_exit;
	}

	/* At this point, the tracee is stopped and waiting for the tracer to set
	 * up some options. */

	/* Here we set up two options:
	 *
	 * ‹PTRACE_O_EXITKILL› will kill the tracee if the tracer (the calling
	 * process) gets killed. This is similar to ‹prctl()›'s ‹PR_SET_DEATHSIG›,
	 * except it will kill all tracees.
	 *
	 * ‹PTRACE_O_TRACESYSGOOD› allows us to distinguish system call traps from
	 * other traps, as it will set the 8th bit in the signal number. That is,
	 * system call traps will see ‹SIGTRAP | 0x80› instead of just ‹SIGTRAP›.
	 *
	 * NOTE: There is an undocumented requirement that ‹PTRACE_GET_SYSCALL_INFO›
	 * requires ‹PTRACE_O_TRACESYSGOOD› option set, otherwise ‹op› attribute will
	 * always be set to zero. */
	if (ptrace(PTRACE_SETOPTIONS, child, NULL, (void*) (PTRACE_O_EXITKILL | PTRACE_O_TRACESYSGOOD)) == -1) {
		warn("%d: Failed to set up tracing", getpid());
		goto kill_child_and_exit;
	}

	/* We are ready, the ‹PTRACE_SYSCALL› command will run the tracee
	 * until a signal is received (the default) or it reaches a system call
	 * boundary. */
	if (ptrace(PTRACE_SYSCALL, child, 0, NULL) == -1) {
		warn("%d: Failed to resume child", getpid());
		goto kill_child_and_exit;
	}

	/* We pass control to ‹count_syscalls()›, which uses ‹wait*()› functions
	 * to wait for events from the tracee. */
	return count_syscalls(child);

kill_child_and_exit:
	/* If we failed at some point, make sure to terminate the child. */
	kill(child, SIGKILL);
	exit(EXIT_FAILURE);
}


/* Main
 * ---- */

/* Finally, here we put everything together. To review the points from above:
 *
 * • The program forks into two processes, the parent (which will assume
 *   the role of tracer) and the child (the tracee).
 *
 * • The task of the tracee (which is still under our control after ‹fork()›)
 *   must prepare itself to be traced. If it executed the new program right
 *   away, the parent might miss several system calls executed by the
 *   new program. The synchronization is achieved in the following sequence:
 *
 *   1. The tracee allows itself to be traced by a call to ‹ptrace()›.
 *   2. It raises ‹SIGSTOP› and stops itself. This allows the tracer to catch
 *      up and set up tracing.
 *   3. After “waking up”, the tracer is fully set up, so the command is
 *      executed via ‹exec*()›. The tracee cannot escape from the tracer now.
 *
 * • The tracer can optionally set up some tracing options, and then waits
 *   for events from the tracee. The initialization works as follows:
 *
 *   1. The tracer waits for the tracee to stop (see point 2 above) using
 *      ‹waitpid()› and checks that the tracee stopped via ‹SIGSTOP›.
 *      If ‹waitpid()› reports something else, it means the tracee has failed
 *      already.
 *   2. The tracer now has an opportunity to set some options.
 *   3. It resumes the tracee using the command ‹PTRACE_SYSCALL› which will
 *      stop the tracee if it encounters a syscall, and reports the event
 *      to the tracee.
 *   4. At this point, the tracer simply waits for events that are reported
 *      by ‹wait*()› interface. Every time the tracer must resume the tracee
 *      by the command above.
 *
 *   A tracer also “seize” a process (if it has permissions to do so), which
 *   is a bit different situation, but follows a similar pattern as above. */

int main(int argc, char *argv[])
{
	if (argc <= 1) {
		usage(argv[0]);
		return EXIT_FAILURE;
	}

	pid_t pid = fork();
	if (pid == -1)
		err(EXIT_FAILURE, "Failed to fork");

	if (pid == 0)
		run_tracee(argv[1], &argv[1]);

	size_t calls = run_tracer(pid);
	printf("Tracee performed %zu calls before it terminated\n", calls);
	return EXIT_SUCCESS;
}
