#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <err.h>
#include <pthread.h>
#include <sys/types.h>
#include <unistd.h>

#if defined(__linux__) && __GLIBC__ < 2 || __GLIBC_MINOR__ < 30
/* ‹gettid()› system call wrapper is only available since glibc 2.30; but we can
 * easily implement it ourselves if it is not provided. */
#include <sys/syscall.h>

pid_t gettid(void)
{
	return syscall(SYS_gettid);
}
#endif

/* Forward declaration. */
static
void *thread_run(void *unused);

/* This auxiliary function will help us identify a thread. */
static
void identify(const char *what)
{
	/* Process ID. */
	printf("(%s)\tpid=%d\n", what, getpid());

#if defined(__linux__)
	/* Thread ID. This value only makes sense on Linux. */
	printf("(%s)\ttid=%d\n", what, gettid());
#endif

#if defined(__linux__) && defined(__GLIBC__)
	/* This is «neither» a portable nor a correct way to identify a thread!
 	* It is only for a demonstration of a difference between thread ID
 	* and its ‹pid›. */
	printf("(%s)\tp_t=%lu\n", what, (unsigned long) pthread_self());
#endif
}

/* This demo shows how to create a thread, how to join it and how to identify
 * threads. */

int main(void)
{
	/* First of all, any ‹pthread_*()› function returns an error code on failure,
	 * without modifying ‹errno›. While we could technically use ‹errno›,
	 * we can just as well have a separate value for that. */
	int pterrno;

	/* To create a thread, we pass ‹&thread_run› pointer to ‹pthread_create()›.
	 * This function (if successful) starts a new thread that will execute
	 * ‹thread_run(NULL)›, where ‹NULL› is the last argument to ‹pthread_create()›.
	 * This function can, obviously, fail as ‹fork()› can. */
	pthread_t thread;
	if ((pterrno = pthread_create(&thread, NULL, &thread_run, NULL)) != 0)
		err(EXIT_FAILURE, "pthread_create: %s", strerror(pterrno));

	/* At this point, we have two threads:
	 * • The main thread executed ‹pthread_create()› and continues its execution
	 *   here, in ‹main()›.
	 * • A new thread started to execute ‹thread_run(NULL)›. Once it finishes,
	 *   it will be joinable. */
	printf("Thread created\n");

	/* Now we identify the main thread. */
	identify("main");

	/* We wait for the worker thread to finish. We do not care for the return
	 * value of the thread, thus we pass ‹NULL› as the second argument. */
	pterrno = pthread_join(thread, NULL);

	/* This function can only fail if we made some fatal mistake, thus we can
	 * use ‹assert()› for that. */
	assert(("pthread_join", pterrno == 0));

	/* At this point we again have only one thread. */
	printf("Thread joined\n");
	return 0;
}

void* thread_run(void* unused)
{
	/* This function gets run in the context of a new thread. It gets an argument
	 * from ‹pthread_create()›, which we are not using here. */
	identify("worker");

	/* Similarly to ‹main()›, thread function is supposed to return a value.
	 * We have nothing, so we will return ‹NULL›. */
	return NULL;
}

