#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <err.h>
#include <pthread.h>
#include <unistd.h>

/* We remember this declaration from ‹ex01t›. Nothing new here. */
#if defined(__linux__) && __GLIBC__ < 2 || __GLIBC_MINOR__ < 30
#include <sys/syscall.h>
#include <sys/types.h>

static
pid_t gettid(void)
{
	return syscall(SYS_gettid);
}
#endif

/* When we want to pass data to threads, usually the best way is to
 * pass a structure. You may also find some “tricks” which pass arbitrary
 * numbers by casting them to pointers. This is also possible, though
 * for threads to be useful, they mostly need more than one number. */
struct thread_data {
	long number;
	pthread_t self;
	pthread_key_t key;
};


/* Forward declarations
 * -------------------- */

/* As with ‹ex01t›, this example makes a bit more sense when explained from
 * ‹main()› first. */

static
void* thread_entry(void* param);


/* Main thread
 * ----------- */

/* This demo shows how to create and use POSIX Thread-Local Storage (TLS). */

int main(int argc, char *argv[])
{
	if (argc != 2) {
		fprintf(stderr, "usage: %s THREADS\n", argv[0]);
		return EXIT_FAILURE;
	}

	/* Remember: This is just a demo, so we keep unimportant parts of the code
	 * as simple as possible, here by using ‹atol()›. In production-grade
	 * programs, use ‹strtol()› with full error checking! */
	long threads_count = atol(argv[1]);
	if (threads_count <= 0) {
		fprintf(stderr, "%s: A positive integer is required\n", argv[1]);
		return EXIT_FAILURE;
	}

	int pterrno;

	/* Before we even start creating threads, we create a TLS key. This key
	 * is a “variable”, where ‹pthread_getspecific()› reads a “value”.
	 * As the value is a pointer, a NULL value can be seen as unset variable. */

	pthread_key_t tls_key;
	if ((pterrno = pthread_key_create(&tls_key, NULL)) != 0)
		errx(EXIT_FAILURE, "Cannot create TLS key: %s", strerror(pterrno));

	/* At this point, every thread (so far only main) has a key associated
	 * with it. We can access it if we want: */
	char message[64];
	snprintf(message, sizeof(message), "Main thread: %d", gettid());
	pterrno = pthread_setspecific(tls_key, message);
	assert(("pthread_setspecific failed", pterrno == 0));

	/* We can access the value using ‹pthread_getspecific()› as if to read
	 * a value of a global variable: */
	printf("get_specific: %s\n", (char *) pthread_getspecific(tls_key));


	/* Okay, now let's create threads. The rest of the code of ‹main()› function
	 * is not that important, just browse over it and move over below to the
	 * ‹thread_entry()› function. */
	struct thread_data *threads = malloc(sizeof(*threads) * threads_count);
	if (threads == NULL)
		err(EXIT_SUCCESS, "Cannot create data for threads");

	for (long i = 0; i < threads_count; ++i) {
		/* The ‹threads› structure is not yet initialized. */
		threads[i] = (struct thread_data) {
			.number = i,
			/* ‹self› will be initialized by ‹pthread_create()› call. */
			.key = tls_key,
		};

		if ((pterrno = pthread_create(&threads[i].self, NULL, thread_entry, &threads[i])) != 0)
			errx(EXIT_FAILURE, "pthread_create(%ld): %s", i, strerror(pterrno));
	}

	/* And then join them: */
	for (long i = 0; i < threads_count; ++i) {
		pterrno = pthread_join(threads[i].self, NULL);
		assert(("pthread_join failed", pterrno == 0));
	}

	/* Let us again, to be sure, print what this main thread sees in its
	 * Thread-Local Storage. */
	printf("get_specific: %s\n", (char *) pthread_getspecific(tls_key));

	/* Finally, we delete the key. */
	if ((pterrno = pthread_key_delete(tls_key)) != 0)
		warnx("pthread_key_delete: %s", strerror(pterrno));

	return EXIT_SUCCESS;
}

/* This function is called in every worker thread. It demonstrates how
 * TLS keys are used to access values. */
void* thread_entry(void* param)
{
	struct thread_data *data = (struct thread_data *) param;

	/* When enter the thread after main has set its TLS key to some value.
	 * However, this thread sees no such value (we will only print this
	 * for the very first thread with ID 0). */
	if (data->number == 0)
		printf("Data upon thread 0 entry: %s\n", (char *) pthread_getspecific(data->key));

	/* The code above should print “(null)” as this thread has not set any
	 * data yet. Now, we will set up some data: */

	char text[64];
	snprintf(text, sizeof(text), "Thread %ld:%d", data->number, gettid());

	int pterrno = pthread_setspecific(data->key, text);
	assert(("pthread_setspecific failed", pterrno == 0));

	/* To give other threads a chance to catch up, we will yield the CPU for
	 * them. This will put our thread to the end of the scheduler queue.
	 * However, you should «not» use this function unless you have a very
	 * good reason to do so. Read the rationale in its manual.
	 *
	 * Also note that there was a non-standard ‹pthread_yield()› function
	 * available in GNU libc, but has been deprecated. */
	pterrno = sched_yield();

	/* This function is guaranteed to succeed on Linux, but the manual requires
	 * us to handle a possible error, so we will do that at least offensively. */
	assert(pterrno == 0);

	/* At this point, some threads should also set up their values. Let us now
	 * print them again: */

	printf("[%ld] %s\n", data->number, (char *) pthread_getspecific(data->key));

	/* At this point, you should see number in “[N]” match the value printed
	 * from the TLS variable, but it should be different for every thread.
	 * (Minus possible buffer race conditions we talked about before.) */

	return NULL;
}
