#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <err.h>
#include <pthread.h>

/* This example show a bit complicated comparison of various TLS approaches
 * and compares them to other variable types:
 *
 * • POSIX Thread-Local Storage keys,
 * • GNU ‹__thread› storage specified (also available as ‹thread_local› in C11),
 * • local variable,
 * • global variable. */


/* When passing data to a thread, it is often useful to model it with
 * a structure. */
struct thread_data {
	long number;
	pthread_t self;
	pthread_key_t *key;

	/* Barriers will be discussed in more details on the next seminar. */
	pthread_barrier_t *barrier;
};

static
void set_thread_data(struct thread_data *target, long id, pthread_key_t *key,
		pthread_barrier_t *barrier)
{
	target->number = id;
	target->key = key;
	target->barrier = barrier;
}

/* This function hides access to POSIX TLS variables and transparently
 * handles allocation of the values. It returns a ‹void*› to the data,
 * which can be then dereferenced for access, even as l-value. See macros below. */
static
void *_posix_tls(pthread_key_t key, size_t size)
{
	/* Return previous value if it exists. */
	void *memory;
	if ((memory = pthread_getspecific(key)) != NULL)
		return memory;

	/* Allocate zeroed memory for the new storage. */
	if ((memory = calloc(size, 1)) == NULL) {
		warn("Cannot allocate storage for thread-specific data");
		pthread_exit(NULL);
	}

	int pterrno;

	/* Set value for the thread. */
	if ((pterrno = pthread_setspecific(key, memory)) != 0) {
		warnx("Cannot set thread-specific data: %s", strerror(pterrno));
		free(memory);
		pthread_exit(NULL);
	}

	return memory;
}

/* The following macro wraps access ‹_posix_tls()› with proper type casts
 * and evaluates to an l-value.
 *
 * This means that result of ‹_posix_tls_for(…)› is a value that can be
 * directly modified (e.g. as ‹E++› or ‹E += α›), and this change will modify
 * the value stored by POSIX TLS. */
#define _posix_tls_for(TYPE, KEY) \
	(*(TYPE *) _posix_tls((KEY), sizeof(TYPE)))

/* An alias for POSIX TLS storage for ‹long› type. */
#define _posix_tls_long(KEY) \
	_posix_tls_for(long, (KEY))


/* Static storage (per-process). */
static long counter_static;

/* ‹__thread› is a «non-standard» GNU storage specifier for some compilers.
 * It is available in C11 as ‹_Thread_local›. */
static __thread long counter_thread;


/* Demonstrate how these variables change. */
void *run_thread(void *raw_data)
{
	struct thread_data *data = raw_data;

	/* Local (automatic) storage variable. */
	long counter_local = 0;

	int pterrno;

	/* This will stop all threads until all of them call this function.
	 * Barriers will be discussed in more details in the next seminar. */
	if ((pterrno = pthread_barrier_wait(data->barrier)) != 0
			&& pterrno != PTHREAD_BARRIER_SERIAL_THREAD) {
		warnx("Barrier wait failed: %s", strerror(pterrno));
		pthread_exit(NULL);
	}

	/* At this point, all threads are woken up here at (almost) the same time. */

	for (int i = 0; i < 100; ++i) {
		counter_static++;
		counter_thread++;
		counter_local++;

		/* This will create the thread-local storage on the first access. */
		_posix_tls_long(*data->key)++;
	}

	printf("thread %-3ld |\tlocal=%-3ld\tstatic=%-3ld\tpthread_key=%-3ld\t__thread=%-3ld\n",
			data->number, counter_local, counter_static, counter_thread,
			_posix_tls_long(*data->key));

	return NULL;
}

int main(int argc, char *argv[])
{
	if (argc != 2) {
		fprintf(stderr, "usage: %s THREADS\n", argv[0]);
		return EXIT_FAILURE;
	}

	/* Do you remember? See ‹ex04t› if you don't. */
	long threads_count = atol(argv[1]);
	if (threads_count <= 0) {
		fprintf(stderr, "%s: A positive integer is required\n", argv[1]);
		return EXIT_FAILURE;
	}

	struct thread_data *thread_data = malloc(sizeof(*thread_data) * (threads_count + 1));
	if (thread_data == NULL)
		err(EXIT_FAILURE, "malloc");

	int pterrno;
	pthread_key_t tls_key;
	if ((pterrno = pthread_key_create(&tls_key, &free)) != 0)
		errx(EXIT_FAILURE, "Cannot create thread key: %s", strerror(pterrno));

	/* Barrier will be explained later. For now, it is a synchronization
 	 * element that will block threads until the specified number of them
 	 * calls ‹pthread_barrier_wait()› somewhere else. */
	pthread_barrier_t barrier;
	if ((pterrno = pthread_barrier_init(&barrier, NULL, threads_count + 1)) != 0)
		errx(EXIT_FAILURE, "Cannot create barrier: %s", strerror(pterrno));

	/* Set data for main thread. */
	set_thread_data(&thread_data[0], 0, &tls_key, &barrier);

	for (unsigned i = 0; i < threads_count; ++i) {
		set_thread_data(&thread_data[i + 1], i + 1, &tls_key, &barrier);

		if ((pterrno = pthread_create(&thread_data[i].self, NULL, &run_thread,
						&thread_data[i + 1])) != 0)
			errx(EXIT_FAILURE, "Cannot start thread %u: %s", i, strerror(errno));
	}

	/* For comparison, main thread will also execute the same code path
 	 * as other threads. */
	run_thread(&thread_data[0]);

	for (unsigned i = 0; i < threads_count; ++i) {
		if ((pterrno = pthread_join(thread_data[i].self, NULL)) != 0)
			warnx("Cannot join thread %u: %s", i, strerror(pterrno));
	}

	/* Destroy the barrier and key. */
	if ((pterrno = pthread_barrier_destroy(&barrier)) != 0)
		warnx("Cannot destroy barrier: %s", strerror(pterrno));

	if ((pterrno = pthread_key_delete(tls_key)) != 0)
		warnx("Cannot delete thread key: %s", strerror(pterrno));

	free(thread_data);
	return EXIT_SUCCESS;
}
