#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include <err.h>
#include <pthread.h>

/* In this example we will demonstrate a mutual exclusion problem.
 * We will create several threads, each of which will try to modify two
 * variables. One of these variables will be protected by a Mutex, the other
 * will not. */

/* Forward declaration. */
static
void *thread_run(void *data);


/* Main
 * ---- */

struct thread_shared {
	pthread_mutex_t mutex;
	/* Note: If we change ‹volatile› for ‹_Atomic›, this concrete problem will
	 * go away and we will always see the correct result.
	 *
	 * However, generally critical sections deal with more complex invariants
	 * (e.g. linked lists) where simple ‹_Atomic› type will not do and we would
	 * have to modify two things “atomically” (e.g. list tail and next pointer
	 * from node). Keep this in mind -- this is a demo to show how mutex is used. */
	volatile int unprotected; /* Accessed without a mutex. */
	volatile int protected;   /* Accessed with a muted held. */
};

/* We will pass this structure to threads: */
struct thread_data {
	pthread_t thread;
	struct thread_shared *shared;
};

int main(int argc, char *argv[])
{
	/* The beginning of this code is no different to what we have seen before. */
	if (argc != 2)
		errx(EXIT_FAILURE, "usage: %s THREADS", argv[0]);

	/* Remember: ‹atoi()› → ‹strtol()› in production code. */
	long threads_count = atoi(argv[1]);
	if (threads_count <= 0)
		errx(EXIT_FAILURE, "Thread count must be positive");

	struct thread_data *thread_data = malloc(threads_count * sizeof(*thread_data));
	if (thread_data == NULL)
		err(EXIT_FAILURE, "Cannot allocate memory for threads");

	struct thread_shared shared = { 0 };

	/* Create a mutex with default attributes. */
	if ((errno = pthread_mutex_init(&shared.mutex, NULL)) != 0)
		err(EXIT_FAILURE, "Cannot create mutex");

	/* Run threads. */
	for (int i = 0; i < threads_count; ++i) {
		thread_data[i].shared = &shared;
		if ((errno = pthread_create(&thread_data[i].thread, NULL, thread_run,
						&thread_data[i])) != 0)
			err(EXIT_FAILURE, "Cannot start thread %d", i);
	}

	/* Join threads. */
	for (int i = 0; i < threads_count; ++i) {
		int pterrno = pthread_join(thread_data[i].thread, NULL);
		assert(("pthread_join failed", pterrno == 0));
	}

	/* Finally, compare values of the variables. */
	printf("Unprotected:\t%d\n", shared.unprotected);
	printf("Protected:\t%d\n", shared.protected);

	/* Release resources and exit. */
	pthread_mutex_destroy(&shared.mutex);
	free(thread_data);
	return 0;
}

static
void *thread_run(void *data_raw)
{
	struct thread_data *data = data_raw;

	/* In this function, we will increment all variables 1000 times in
	 * a loop. A question arises, though: where do we lock the mutex?
	 *
	 * Obviously, we can lock the mutex «around» the loop. It will certainly
	 * create a valid critical section. The problem is, however, that this
	 * wouldn't be generic enough -- this section would be so huge that
	 * other threads will wait until this thread finishes its work,
	 * thus resulting in basically a sequential code:
	 *
	 * 	t₁: [… 1000 …] ×
	 * 	t₂:                       [… 1000 …] ×
	 * 	t₃:            [… 1000 …] ×
	 *
	 * However, we want to simulate real workers that mostly work outside
	 * of critical sections and only enter them when necessary. Thus, we will
	 * only lock the mutex when we will want to modify ‹protected› variable. */

	for (int i = 0; i < 1000; ++i) {
		/* Unprotected variable gets modified straight away. */
		data->shared->unprotected++;

		/* To access the protected variable, we will make sure we are the
		 * only thread modifying it by locking the mutex.
		 *
		 * Note: «Always» check that the operation succeeded! This operation
		 * «can» fail for whatever reason, and it is an error not to check
		 * mutex locking! */
		if ((errno = pthread_mutex_lock(&data->shared->mutex)) != 0)
			err(EXIT_FAILURE, "Cannot lock mutex");

		/* We are in a critical section. */
		data->shared->protected++;

		/* We are leaving the critical section. */
		int pterrno = pthread_mutex_unlock(&data->shared->mutex);
		assert(("pthread_mutex_unlock failed", pterrno == 0));
	}

	/* Even though we “synchronize” threads on mutex, some of them may be too
	 * fast for the scheduler to hit the ‹unprotected› variable modification
	 * concurrently, thus leading into a race condition. We should observe
	 * that ‹unprotected› value will likely be less than ‹protected›,
	 * which will always be correct, ten times the number of threads. */

	return NULL;
}
