#include <assert.h>
#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <err.h>
#include <pthread.h>
#include <unistd.h>

/* In this example we showcase conditional variables used for pure signalling
 * and how this can lead to potential problems. */

/* We will use statically initialized objects. In order to work with the
 * condition variable, we need to hold a mutex. The API even requires it
 * at some point. */
static
pthread_mutex_t MTX = PTHREAD_MUTEX_INITIALIZER;

static
pthread_cond_t COND = PTHREAD_COND_INITIALIZER;

/* We will have eight worker threads. */
#define THREADS_COUNT 8

static
void *worker(void *data_raw)
{
	/* Worker's task is to wait for a signal from the main thread. In order
	 * to do so, we will use condition variable. An advantage over POSIX signals
	 * is that it's faster, does not require special handler and can wake up
	 * multiple threads instead of just one. */

	long id = (long) data_raw;
	printf("%ld: Thread started\n", id);

	/* Before we attempt to work with condition variable, we need to hold
	 * mutex to avoid races. */
	if ((errno = pthread_mutex_lock(&MTX)) != 0)
		err(EXIT_FAILURE, "%ld: Cannot lock mutex", id);

	/* Now we can wait for the condition change to be signaled on the variable.
	 * Note that this operation releases the mutex, but will re-acquire it
	 * once the signal arrives. */
	if ((errno = pthread_cond_wait(&COND, &MTX)) != 0)
		err(EXIT_FAILURE, "%ld: Cannot wait for condvar", id);

	/* Therefore, at this point, we «still» hold the mutex lock, but it may
	 * have been released and re-acquired in the above call. */

	printf("%ld: Signal received\n", id);

	/* Unlock the mutex as we are leving the critical section. */
	int pterrno = pthread_mutex_unlock(&MTX);
	assert(pterrno == 0 && "pthread_mutex_unlock failed");
	return NULL;
}

int main(void)
{
	/* The main thread creates threads, just the ordinary. */
	pthread_t threads[THREADS_COUNT];
	for (int i = 0; i < THREADS_COUNT; ++i) {
		if ((errno = pthread_create(&threads[i], NULL, worker, (void *) i)) != 0)
			err(EXIT_FAILURE, "Cannot create thread %d", i);
	}

	printf("main: Threads started\n");

	/* We will sleep here to allow worker threads to catch up to the
	 * condition variable wait. */
	sleep(2);

	printf("main: Woken up\n");

	/* In general, we should be holding the mutex even if the only task is to
	 * signal the condition variable. This is because threads may need some
	 * time to transition from being woken up to compete for the lock; having
	 * the lock blocked by us will have a chance to wake up more threads. */
	if ((errno = pthread_mutex_lock(&MTX)) != 0)
		err(EXIT_FAILURE, "main: Cannot lock mutex");

	/* This call will wake up all threads waiting for the condition variable. */
	if ((errno = pthread_cond_broadcast(&COND)) != 0)
		err(EXIT_FAILURE, "main: Cannot broadcast");

	/* And finally, we unlock the mutex (which will start the competition of
	 * worker threads) and wait for threads to finish. */
	int pterrno = pthread_mutex_unlock(&MTX);
	assert(pterrno == 0 && "pthread_mutex_unlock failed");

	for (int i = 0; i < THREADS_COUNT; ++i) {
		printf("main: Waiting for %d\n", i);

		int pterrno = pthread_join(threads[i], NULL);
		assert(pterrno == 0 && "pthread_join failed");
	}

	/* In the output you may see that threads wait for the ‹sleep()› to expire
	 * and then they all receive the signal. However, is this solution correct?
	 *
	 * • What will happen if we replace ‹pthread_cont_broadcast()› with
	 *   ‹pthread_cond_signal()›?
	 * • What will happen if we remove the ‹sleep()› call? */

	return EXIT_SUCCESS;
}
