#include <assert.h>
#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <err.h>
#include <pthread.h>
#include <unistd.h>

#define WORKERS 4
_Static_assert(0 < WORKERS && WORKERS <= 100, "Invalid number of workers");

/* This program demonstrates use of condition variable when accessing
 * a shared state, in this case in a form of a data structure that holds
 * “jobs” (simply numbers). Important points that we will discuss in a bit more
 * detail in the code are as follows:
 *
 * 1. When there are no “jobs” to process, the worker thread suspends.
 * 2. When a producer thread creates a “job”, it notifies potentially suspended
 *    threads by using a signal. */

static
pthread_mutex_t MTX = PTHREAD_MUTEX_INITIALIZER;

static
pthread_cond_t COND = PTHREAD_COND_INITIALIZER;

/* Job queue is modeled using a linked list queue. This is very inefficient
 * to be used in production, but serves well for the demo. */
struct node {
	struct node *next;
	long data;
};

struct queue {
	struct node *head;
	struct node *tail;
};

/* The function will initialize the queue. */
static inline
void queue_init(struct queue *q)
{
	/* This function is expected to be called from a single thread only,
	 * as it would make no sense to initialize the queue from multiple threads
	 * at once. Therefore, no locking mechanism will be used here. */
	memset(q, 0, sizeof(*q));
}

/* Insert data to the shared queue. The data inserted here will be available
 * to any other thread that has the queue. */
static
bool queue_insert(struct queue *q, long data)
{
	/* To keep the critical section as short as possible, we will do necessary
	 * preparations before acquiring the mutex.
	 *
	 * This will also simplify the critical section a bit. If we had the entire
	 * body of this function wrapped in a critical section, the allocation
	 * failure would need an early return with its own mutex unlocking. */
	struct node *node = malloc(sizeof(*node));
	if (node == NULL)
		return false;

	node->next = NULL;
	node->data = data;
	/* NOTE: The following line would introduce a race condition. As we do not
	 *       have the mutex yet, the tail might change by the time we enter
	 *       the critical section. */
	// q->tail->next = node;

	/* We have done everything we could, now we need to actually insert
	 * the node to the queue. */
	errno = pthread_mutex_lock(&MTX);
	/* It would be a bit cleaner to use proper error checking here, but since
	 * this is a demo and we only use default mutex, there is actually no
	 * error condition except our own mistake that could cause this. */
	assert(errno == 0 && "queue_insert(): pthread_mutex_lock() failed");

	/* Now we can enqueue the element. You should know how to do this by now,
	 * as it was already done in some way in PB071. */
	if (q->tail == NULL)
		q->head = node;
	else
		q->tail->next = node;

	q->tail = node;

	/* Now, the new part: There might be jobs waiting for this element in the
	 * queue. To wake them up, we will notify on the condition variable.
	 * Since we have only added a single element, we will wake up (at least)
	 * one thread, there is no need to wake them all. */
	errno = pthread_cond_signal(&COND);
	assert(errno == 0 && "queue_insert(): pthread_cond_signal() failed");

	/* Let's stop a while here. Every time we insert an element, we call
	 * the signalling function, even when there may be no waiting threads.
	 * Is it something we can detect and potentially avoid? Try to answer
	 * on your own: How can we tell how many (and if) there are threads waiting
	 * for the state to change? */

	/* We are done, let's leave the critical section. */
	errno = pthread_mutex_unlock(&MTX);
	assert(errno == 0 && "queue_insert(): pthread_mutex_unlock() failed");

	/* All done here. */
	return true;
}

/* Remove and optionally return an element from the queue. If the queue is
 * empty, the thread will block until an element becomes available. */
static
bool queue_remove(struct queue *q, long *data)
{
	/* We cannot do anything outside of the critical section, so let's
	 * try to acquire the mutex. */
	errno = pthread_mutex_lock(&MTX);
	assert(errno == 0 && "queue_remove(): pthread_mutex_lock() failed");

	/* This is where the condition variable truly introduces the condition.
	 * What is that we need to be true here? Our task is to return an element,
	 * so, naturally the condition that must be true before we continue is
	 * “there is at least one element”. */
	while (q->head == NULL) {
		/* If the condition doesn't hold, we will wait for another thread
		 * to change it and wake us up. */
		errno = pthread_cond_wait(&COND, &MTX);
		/* Again, proper checking can also be done here, but only our own mistake
		 * can cause this error. */
		assert(errno == 0 && "queue_remove(): pthread_cond_wait() failed");

		/* Note that we also pass ‹&MTX› to ‹pthread_cond_wait()›. Since we
		 * expect other thread to change the condition, it must also be able
		 * to enter the critical section. Thus the function unlocks the mutex
		 * for us before it suspends the thread. And when the thread wakes up,
		 * the function will re-acquire the mutex again before it returns.
		 *
		 * Therefore the critical section gets technically interrupted in
		 * the ‹pthread_cond_wait()› call, though from our point of view
		 * we are always in the critical section. */
	}

	/* IMPORTANT: Note that we use ‹while› instead of ‹if›.
	 *
	 * Imagine for a while that our worker detects that the queue is empty
	 * and suspends itself. It wakes up only after ‹queue_insert()› calls
	 * signal. So, does it mean we can always expect the queue to contain
	 * one element when the thread wakes up?
	 *
	 * We cannot. A counterexample could be this (‹c₁› and ‹c₂› are consumer
	 * threads, ‹p› is a producer thread):
	 *
	 *      c₁  Attempts to lock mutex; proceeds.
	 *      c₁  The queue is empty; suspends.
	 *      p   Attempts to lock mutex; proceeds.
	 *      c₂  Attempts to lock mutex; suspends.
	 *      p   Creates an element and signals the condition variable.
	 *      c₁  Wakes up.
	 *      c₁  Attempts to lock mutex; suspends.
	 *      p   Unlocks the mutex.
	 *      c₂  Resumes and acquires the mutex.
	 *      c₂  Removes the element.
	 *      c₂  Unlocks the mutex.
	 *      c₁  Resumes and acquires the mutex.
	 *      c₁  The queue appears to be empty.
	 *
	 * It is not an unrealistic scenario, as from the point of the mutex, ‹c₂›
	 * was waiting longer for the mutex than ‹c₁›, which was instead waiting
	 * for signal. Hence, it would be incorrect to assume there is always
	 * an element when the thread wakes up.
	 *
	 * Another counterexample would be two or more consumer threads waiting
	 * for the signal, as ‹pthread_cond_signal()› can sometimes wake up
	 * more than one thread. Try to work out this scenario on your own. */

	/* We get here when the condition evaluated to false, hence, here we
	 * are in the critical section and we have at least one element. */

	struct node *node = q->head;

	/* Untangle the node from the queue. */
	q->head = q->head->next;
	if (q->head == NULL)
		q->tail = NULL;

	/* We will leave the rest for later, we should leave the critical section
	 * as soon as possible. */
	errno = pthread_mutex_unlock(&MTX);
	assert(errno == 0 && "queue_insert(): pthread_mutex_unlock() failed");

	/* If ‹data› pointer is not null, fetch the data. */
	if (data != NULL)
		*data = node->data;

	free(node);

	/* As you can notice, this function always returns ‹true›. We will keep
	 * the boolean return value, though, because of a (possible) non-blocking
	 * variant of this function described just below. */
	return true;
}

/* Data structures with blocking functions usually offer a non-blocking variant,
 * in this case that would be ‹queue_try_remove()›. We will not provide it
 * here in the code, but it would work as follows:
 *
 *  1. Lock the mutex.
 *  2. If the queue is not empty, remove the element.
 *  3. Unlock the mutex.
 *  4. Return ‹true› only if an element was removed. */


/* For consistency (and good practice) we will also provide a function to
 * destroy the queue. */

static
void queue_destroy(struct queue *queue)
{
	/* Technically we could make this function synchronized, but it would be
	 * a needless complication.
	 *
	 * Even in C we can have a concept of “ownership”, where some thread owns
	 * the queue and “lends” it to other threads.  We expect this owner thread
	 * to initialize and clean up the queue once it is not used, eliminating
	 * the requirement for synchronized access.
	 *
	 * It is then up to the programmer to ensure that only this thread destroys
	 * the queue. */

	struct node *node = queue->head;
	while (queue->head != NULL) {
		free(queue->head);
		queue->head = node->next;
	}
}


/* For the main function, we will create several worker threads, while the
 * main will serve as a producer. */

struct thread_data {
	long id;
	pthread_t thread;
	struct queue *queue;
};

static
void *worker(void *data);

int main(void)
{
	struct queue queue;
	queue_init(&queue);

	struct thread_data thread_data[WORKERS];

	for (int i = 0; i < WORKERS; ++i) {
		thread_data[i] = (struct thread_data) {
			.id = i,
			.queue = &queue,
		};

		if ((errno = pthread_create(&thread_data[i].thread, NULL,
						worker, &thread_data[i])) != 0)
			err(EXIT_FAILURE, "Failed to start thread %d", i);
	}

	/* The main thread will now enqueue 100 jobs. */
	for (long job_id = 0; job_id < 100; ++job_id) {
		/* Since the queue is itself synchronized, we do not need to use
		 * locks here ourselves. */
		if (!queue_insert(&queue, job_id))
			err(EXIT_FAILURE, "Failed to insert an element into the queue");

		printf("<P> Enqueued job %ld\n", job_id);
		/* To force other threads to wait once in a while, we will suspend
		 * the producer for 0-150 ms. */
		usleep((rand() % 150) * 1000);
	}

	/* Once we get out of the loop, worker threads will eventually catch up
	 * and process all jobs. Then they will suspend, waiting for another
	 * job that will never come. */

	/* There are several possible solutions for this problem. Here are two
	 * possible solutions:
	 *
	 * 1. Provide a function that will block until the queue gets empty.
	 *
	 *    Once producers are finished and joined by the “supervisor” (for
	 *    example, ‹main()›), it can call the function, wait until the queue
	 *    gets empty, and then it can, for example, join the worker threads.
	 *
	 *    A slight problem is that, since the condition waited for here is
	 *    completely different than what ‹queue_remove()› waits for, another
	 *    condition variable must be created and notified by ‹queue_remove()›.
	 *
	 *    Indeed, if we attempted to use the ‹COND› for this task, we would
	 *    soon find out that we need to wake up «all» threads, which would
	 *    become highly inefficient. Can you figure out why?
	 *    (Hint: Imagine consumer and supervisor waiting. What will happen
	 *    if ‹pthread_cond_signal()› wakes up only one of the threads?)
	 *
	 *    This approach is used, for example, by Python's ‹queue› module:
	 *    https://docs.python.org/3/library/queue.html → ‹join()› methods.
	 *
	 *    Note that this approach also checks that workers finished processing
	 *    the element, which requires them to call ‹task_done()› method.
	 *    This is not strictly necessary -- we can argue that it's not queue's
	 *    job to track what happens with elements once they are removed from it.
	 *    It is a nice addition, but also a non-trivial complication.
	 *
	 *
	 * 2. Provide a way to “abort” threads waiting in ‹queue_remove()›.
	 *
	 *    When producers finish and join with the supervisor, instead of
	 *    waiting on a different condition variable, the supervisor can
	 *    “abort” waiting operations. This is usually done by setting some
	 *    shared variable in the queue that means “the queue is finished”,
	 *    then waking up «all» threads. These threads, waking up in
	 *    ‹queue_remove()›,  will notice this change and will return ‹false›
	 *    to indicate the caller that no elements will be added.
	 *
	 *    There can be a race condition with previous producer, so there
	 *    may still be some elements left in the queue, and these should
	 *    be processed.
	 *
	 *    The supervisor then doesn't need to wait for condition to change,
	 *    it can simply join worker threads (as it would do anyway).
	 *
	 *    This idea is implemented, for example, in Perl's ‹Thread::Queue›:
	 *    https://metacpan.org/pod/Thread::Queue → ‹end()› method. */

	/* We are not going to do any of that here, instead, we will give
	 * the workers some time to finish and then tear the entire program
	 * down. */
	sleep(3);
	exit(EXIT_SUCCESS);
}

static
void *worker(void *data_raw)
{
	struct thread_data *data = data_raw;

	/* We have no way of telling that the queue is finished in this demo.
	 * Thus, we will continue processing jobs indefinitely. */
	while (true) {
		/* Again, as the queue is already synchronized, we do not need to
		 * employ any locking mechanism here. */

		long job_id;
		if (!queue_remove(data->queue, &job_id))
			err(EXIT_FAILURE, "[%ld] Failed to remove an element", data->id);

		/* Here, we have fetched a job. We are outside of a critical section,
		 * but that is not a problem -- in this model, jobs are independent
		 * and have no shared state.
		 *
		 * In real life, if jobs are somehow connected to a shared state, we
		 * should expect to see several mutex instances:
		 *
		 * • The first mutex for the queue itself.
		 * • The second mutex for the shared state of jobs.
		 *
		 * It is important to keep job state and the queue separate, so that
		 * operations on queue will not block threads processing the jobs. */

		printf("[%ld] Fetched job %ld\n", data->id, job_id);

		/* Again, to simulate some computation, we will suspend the thread
		 * for 300 ms.  */
		usleep(300000);
	}

	/* This thread will never finish, as there is no way to tell that the
	 * producer is finished creating jobs. Possible solutions are described
	 * above in ‹main()› function. */
	return NULL;
}
