/**
 * Thread-safe queue library.
 *
 * \author      TODO: Put your name here.
 * \file        queue.h
 */

#ifndef QUEUE_API_H
#define QUEUE_API_H

#include <stdbool.h>
#include <stddef.h>

/* We need ‹ssize_t› for ‹queue_forecast()›. */
#include <unistd.h>

/* TODO: The following features are optional bonuses.
 *       Uncomment those you wish to implement. */
//#define QUEUE_FEATURE_FORECAST
//#define QUEUE_FEATURE_DONE
//#define QUEUE_FEATURE_WAIT

/**
 * Thread-safe fixed-size queue.
 *
 * This is an opaque structure. */
struct queue;

/**
 * Status codes that can be returned by operations on the queue. */
enum queue_status_code {
    QUEUE_SUCCESS, /**< Operation succeeded. */
    /* TODO: Add more error codes here. */
};


/* Constructor and destructor
 * -------------------------------------------------------------------------- */

/**
 * Allocates a new queue and stores its pointer to the location pointed
 * to by \p qptr.
 *
 * The behaviour of this function when called on the same \p qptr from two or
 * more threads is undefined and may not be thread-safe.
 *
 * \note    The error code returned by this function is not retrievable
 *          by queue_errno() as no instance would be created in that case.
 *
 * \note    This function may modify \c errno to store additional error code.
 *
 * \param[out]  qptr        Where the pointer to a new queue will be stored.
 * \param[in]   sizeof_elem Size of a single item to be stored.
 * \param[in]   capacity    Maximum queue size.
 * \returns     #QUEUE_SUCCESS on success; error code otherwise. */
int queue_create(struct queue **qptr, size_t sizeof_elem, size_t capacity);

/**
 * Destroys the queue.
 *
 * The behaviour of this function when called on the same queue from two
 * or more threads is undefined and may not be thread-safe.
 *
 * \note    The calling thread should make sure there are no other threads
 *          waiting on this queue, e.g. by calling the queue_abort() function
 *          and joining all threads.
 *
 * \note    When this function returns (successfully or not), the behaviour
 *          of other functions called with the same \p queue pointer
 *          (except queue_create()) is undefined.
 *
 * \note    The error code returned by this function is not retrievable
 *          by queue_errno() as the instance gets destroyed by this call.
 *
 * \param[in]   queue       Queue to destroy.
 * \returns     #QUEUE_SUCCESS on success; error code otherwise. */
int queue_destroy(struct queue *queue);


/* Queries
 * -------------------------------------------------------------------------- */

/**
 * Returns the size of the struct queue type in bytes. */
size_t queue_sizeof(void);

/**
 * Returns the upper size limit for the given queue. */
size_t queue_capacity(const struct queue *queue);

/**
 * Returns the element size of the given queue. */
size_t queue_sizeof_elem(const struct queue *queue);

/**
 * Returns the number of elements stored in the queue by the time of the call.
 *
 * \note    The size this function reports can be outdated by the time
 *          the function returns. The caller should therefore employ additional
 *          mechanisms (e.g. mutual exclusion) if it intends to operate
 *          with this value somehow. */
size_t queue_size(const struct queue *queue);

/**
 * Returns \c true if the queue has no elements stored, \c false otherwise.
 *
 * \note    The limitations for \c queue_size() apply here as well. */
bool queue_is_empty(const struct queue *queue);

/**
 * Returns \c true if the queue has no free space available, \c false otherwise.
 *
 * \note    The limitations for \c queue_size() apply here as well. */
bool queue_is_full(const struct queue *queue);

#if defined(QUEUE_FEATURE_FORECAST)
/**
 * Returns the number of elements in the queue once all the waiting operations
 * finish successfully.
 *
 * This function is essentially queue_size() with currently waiting
 * operations counted in.
 *
 * \note The result may be negative if there are more pops waiting
 *       for respective pushes.
 *
 * \note The result may be greater than the queue capacity if there are
 *       more pushes waiting for respective pops.
 *
 * \note The limitations for queue_size() apply here as well. */
ssize_t queue_forecast(const struct queue *queue);
#endif // QUEUE_FEATURE_FORECAST


/* Queue manipulation
 * -------------------------------------------------------------------------- */

/**
 * Waits until the number of elements in the queue is less than the capacity
 * and then pushes a copy of the provided data to the end of the queue.
 *
 * \param[in]   elem        Pointer to data that will be copied to the queue.
 * \returns     #QUEUE_SUCCESS on success; error code otherwise. */
int queue_push(struct queue *queue, const void *elem);

/**
 * Similar to queue_push(), except it fails immediately if the queue
 * is full without waiting.
 *
 * \param[in]   elem        Pointer to data that will be copied to the queue.
 * \returns     #QUEUE_SUCCESS on success; error code otherwise. */
int queue_try_push(struct queue *queue, const void *elem);

/**
 * Waits until there is at least one element in the queue, removes it
 * and optionally returns the value of the removed element.
 *
 * If \p elem is not \c NULL, the function copies the memory contents of
 * the removed element from its internal storage to the memory pointed to by
 * \p elem. The behaviour is undefined if memory pointed to by \p elem
 * is not at least queue_sizeof_elem() bytes long.
 *
 * If \p elem is \c NULL, the function discards the element that would have
 * been returned. No copying occurs.
 *
 * \param[out]  elem        Pointer to memory where the element will be copied
 *                          or NULL if the data should be discarded.
 * \returns     #QUEUE_SUCCESS on success; error code otherwise. */
int queue_pop(struct queue *queue, void *elem);

/**
 * Similar to queue_pop(), except it fails immediately if the queue
 * is empty without waiting.
 *
 * \param[out]  elem        Pointer to memory where the element will be stored
 *                          or NULL if the data should be discarded.
 * \returns     #QUEUE_SUCCESS on success; error code otherwise. */
int queue_try_pop(struct queue *queue, void *elem);


/* Utilities
 * -------------------------------------------------------------------------- */

/**
 * Wakes up all threads waiting on queue_push(), queue_pop()
 * or queue_wait_empty().
 *
 * All calls to queue_push(), queue_try_push(), queue_pop() and queue_try_pop()
 * will fail immediately when this function completes the transition of the
 * queue to the aborted state.
 *
 * \note    It is implementation-defined whether calls to queue_wait_empty()
 *          are aborted as well.
 *
 * \returns     #QUEUE_SUCCESS on success; error code otherwise. */
int queue_abort(struct queue *queue);

#if defined(QUEUE_FEATURE_DONE)
/**
 * Marks the queue as finished, a sort of “soft abort”.
 *
 * All suspended calls to queue_push() will be woken up.
 * In addition, the behaviour of queue_push() and queue_try_push() will change
 * from now on, so that both functions fail immediately without attempting to
 * store the element, even if the queue has space for it.
 *
 * All threads waiting on queue_pop() will be woken up.
 * In addition, the behaviour of queue_pop() will change so that it will be
 * exactly the same as queue_try_pop() from now on.
 *
 * \note    Pending calls to queue_push() will also be aborted. If
 *          the implementation wishes that consumer threads (those that call
 *          queue_pop()) process these elements, it should first ensure that
 *          producer threads are joined before calling queue_done().
 *
 * \returns     #QUEUE_SUCCESS if the operation completes successfully;
 *              error code otherwise. */
int queue_done(struct queue *queue);
#endif // QUEUE_FEATURE_DONE

#if defined(QUEUE_FEATURE_WAIT)
/**
 * Waits until the queue becomes empty and no thread waits on queue_push(),
 * or when queue_abort() is called.
 *
 * A call on an aborted queue results in an error code being returned.
 *
 * \note    The condition can become outdated by the time the function returns
 *          unless this call is sequenced after a call to queue_done().
 *
 * \returns     #QUEUE_SUCCESS if the operation completes successfully;
 *              error code otherwise. */
int queue_wait_empty(struct queue *queue);
#endif // QUEUE_FEATURE_WAIT

/**
 * Returns the error code of the last failed operation on the instance of
 * \p queue that was performed in the calling thread.
 *
 * This error code may be different from other instances of the queue in
 * the calling thread. That is, failures of one queue will not affect
 * error codes of other instances.
 *
 * \note    Error codes returned by queue_create() and queue_destroy()
 *          cannot not be retrieved by this operation.
 *
 * \note    The function may return #QUEUE_SUCCESS if no error has been observed
 *          in the calling thread so far, or if the previous operation succeeded. */
int queue_errno(const struct queue *queue);

/**
 * Returns a string describing the \p code.
 *
 * \note    The returned string may be statically allocated, therefore
 *          the caller must not attempt to free the returned value.
 *
 * \note    The call to this function may fail, in which case it returns
 *          \c NULL and sets \c errno appropriately. However, it never
 *          modifies the queue status code in the calling thread.
 *
 * \param[in]   code        Error code.
 * \returns     A string that describes the error or \c NULL on error. */
const char *queue_strerror(int code);

#endif /* QUEUE_API_H */
