#include <limits.h>
#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <err.h>
#include <fcntl.h>
#include <sys/wait.h>

/* This example will demonstrate how File Descriptors and Open File Descriptions
 * relate to each other. */

/* The following headers are only needed to demonstrate resources with file
 * descriptors that we have not covered yet. Do not worry about these,
 * we will get to them in due time. */
#include <sys/inotify.h>
#include <sys/socket.h>

int main(void)
{
	/* File Descriptor
	 * --------------- */

	/* A File Descriptor (FD from now on) is simply a resource identifier.
	 * Kernel hides the representation of this resource, and provides
	 * us with a number that we need to pass to certain system calls to
	 * interact with the said resource. */

	/* For example, opening a regular file gives us a descriptor.
	 * Note: ‹O_RDONLY› flag is equivalent to ‹"r"› for ‹fopen()›. */
	int reg_fd = open("ofd.c", O_RDONLY);

	/* As with all resources, the kernel may refuse to provide it. For example,
	 * the resource does not exist, is not accessible to the process, or
	 * there is not enough memory for kernel's own accounting of the resource. */
	if (reg_fd == -1)
		err(EXIT_FAILURE, "open(ofd.c)");

	printf("open(ofd.c) → %d\n", reg_fd);

	/* There are other resources we may represent with descriptors, for example
	 * sockets: */
	int sock_fd[2];
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, sock_fd) == -1)
		err(EXIT_FAILURE, "socketpair()");

	printf("socketpair() → %d, %d\n", sock_fd[0], sock_fd[1]);

	/* … or even resources that have nothing to do storage or IPC directly,
	 * but can naturally map to VFS (i.e. they can easily pretend they
	 * are files). Things such as file system monitors: */
	int inotify_fd = inotify_init();

	printf("inotify_init() → %d\n", inotify_fd);

	/* From the output of the program so far you can clearly see that
	 * the numbers (FDs) start from 2 and increment by one. Unsurprisingly,
	 * FDs 0-2 are standard input, output, and error output respectively. */

	/* The ‹proc› file system provides a way to inspect process' descriptors,
	 * simply by listing ‹fd› directory. Entries of this directory are
	 * symbolic links where names are the numbers and targets are some
	 * (non-portable) representations of the symbolic link provided by
	 * the ‹readlink()› system call. See ‹man procfs› for more details
	 * about this format. */
	printf("\nExamples for procfs:\n");
	system("ls -l /proc/self/fd");

	char target[NAME_MAX] = { 0 };
	if (readlink("/proc/self/fd/4", target, sizeof(target) - 1) < 0)
		err(EXIT_FAILURE, "readlink");
	printf("FD 4 → %s\n", target);


	/* Open File Description
	 * --------------------- */

	/* FDs are indices into Open File Table. This is a process-owned table,
	 * where each row describes the type of the resource. Thus, these rows
	 * are called Open File Descriptions. For regular files they contain
	 * offset, status, mode, and some link to the i-node. */

	/* Since this table is owned by the process, it also gets copied when
	 * a new process is created using ‹fork()›. Threads share the same
	 * table, though ‹clone()› on Linux allows to separate them by «not» using
	 * ‹CLONE_FILES› option.
	 *
	 * Let's demonstrate this by reading from this source code (opened
	 * in ‹reg_fd›) three times:
	 *
	 * 1. The parent process
	 * 2. The child process
	 * 3. From a duplicated descriptor in the parent process again. */

	pid_t child_pid = fork();
	char c[1];

	switch (child_pid) {
	case -1:
		err(EXIT_FAILURE, "fork");

	case 0:
		/* Child process. Sleep for 100 ms, read a char and write it out.
		 * This should provide enough time for the parent to read the first
		 * character and print it out. */
		usleep(100000);
		/* Technically, we should check that ‹read()› succeeded, but we will
		 * skip that to keep the demo short. */
		read(reg_fd, c, 1);
		printf("%d %d → %c\n", getpid(), reg_fd, c[0]);
		exit(EXIT_SUCCESS);
	}

	/* Parent process. First, we will duplicate the descriptor. All this
	 * operation does is that it returns another number for the same
	 * resource. */
	int reg_fd_2 = dup(reg_fd);
	if (reg_fd_2 == -1)
		err(EXIT_FAILURE, "dup");

	/* First, we will read one character while the child (probably) sleeps. */
	read(reg_fd, c, 1);
	printf("%d %d → %c\n", getpid(), reg_fd, c[0]);

	/* Now we will sleep for 200 ms, meanwhile the child should print its
	 * character from the file. */
	usleep(200000);

	/* Read a character again, but this time from the duplicated descriptor. */
	read(reg_fd_2, c, 1);
	printf("%d %d → %c\n", getpid(), reg_fd_2, c[0]);
	waitpid(child_pid, NULL, 0);

	/* This prints characters ‹#in› from the beginning of this file despite
	 * used from two processes and even using a different descriptor. This is
	 * because all of these descriptors refer to the same OFD where a single
	 * offset is maintained. Something like this:
	 *
	 *      Process-owned     ┆          Kernel-owned
	 *                        ┆
	 * 		╒═ P₁ FD ═╕       ┆          ╒═ OFD  ══════╕
	 * 		┊  …      ┊       ┆          ┊  …          ┊
	 * 		│  3    ⭘ │━━━━━━━━━━┳━━┳━━━▷│⭘ off mode … │
	 * 		┊  …      ┊       ┆  ┃  ┃    ┊  …          ┊
	 * 		┊  7    ⭘ │━━━━━━━━━━┛  ┃    └─────────────┘
	 * 		└─────────┘       ┆     ┃
	 *  ┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┤     ┃
	 * 		╒═ P₂ FD ═╕       ┆     ┃
	 * 		┊  …      ┊       ┆     ┃
	 * 		│  3    ⭘ │━━━━━━━━━━━━━┛
	 * 		└─────────┘       ┆
	 *
	 * Note that ‹C FD› (child's FDs) do not have FD 7, as the parent created
	 * this descriptor «after» fork. */

	/* Generally, we should close resources when we no longer need them: */
	close(reg_fd_2);

	/* When the process exits, however, all descriptors get ‹close()›d
	 * automatically. */
	exit(EXIT_SUCCESS);
}
