#include <stdio.h>
#include <stdlib.h>

#include <err.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>


/* This example demonstrates how modes are determined using file
 * creation mode and process' umask. */

int main(void)
{
	/* The User File Creation Mode Mask is a formal name for what we
	 * are going to call ‹umask()› from now on. This is an attribute of
	 * the process and limits modes that can be used by ‹open()› or ‹creat()›. */

	/* A simple way to imagine what ‹umask()› does is to say
	 * ‹umask(X)› «forbids» to use modes set by ‹X›. For example, */
	if (umask(S_IWGRP | S_IRWXO) == -1)
		err(EXIT_FAILURE, "umask");

	/* sets umask
	 *
	 *   η	--- -w- rwx  0027
	 *
	 * which means that newly created files will not gain «set» modes,
	 * for example ‹S_IWGRP› or ‹S_IROTH›. */

	/* Indeed, let's create a file: */
	int fd = creat("file", S_IRWXU | S_IRGRP | S_IWGRP | S_IROTH);

	if (fd == -1)
		err(EXIT_FAILURE, "creat");

	/* The mode in ‹creat()› is
	 *
	 * 	 μ	rwx rw- r--  0764
	 *
	 * However, the actual file mask is computed using ‹umask()›'s ⟦η⟧
	 * as ⟦~η & μ⟧:
	 *
	 * 	 η	--- -w- rwx  0027
	 *
	 *  ~η	rwx r-x ---  0750
	 *   μ	rwx rw- r--  0764
	 * ----------------------
	 * 	 →	rwx r-- ---  0740 */

	/* We can verify this from shell using ‹stat› or ‹ls -l›, but we
	 * can also skip a bit and use ‹stat()› system call. */
	struct stat st;
	if (stat("file", &st) == -1)
		err(EXIT_FAILURE, "stat");

	printf(" mask:\t%04o\n", S_IWGRP | S_IRWXO);
	printf("~mask:\t%04o\n", ~(S_IWGRP | S_IRWXO) & ACCESSPERMS);
	printf(" mode:\t%04o\n", S_IRWXU | S_IRGRP | S_IWGRP | S_IROTH);
	printf("\n");

	/* Mode in ‹struct stat› may contain other bits than just file access mode,
	 * so we will display only access mode bits. */
	printf("result:\t%04o\n", st.st_mode & ACCESSPERMS);

	/* Remove the file so that we can restart the application with the same
	 * clean state again. */
	unlink("file");

	return EXIT_SUCCESS;
}
