#include <assert.h>
#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <wait.h>

/* In this example we will explore how unnamed pipes work.
 * The goal of this example is to create a child that will be able to send
 * some data back to the parent using a pipe. */

/* Forward declarations
 * -------------------- */

void parent(int pipefds[2]);
void child(int pipefds[2]);


/* Main
 * ---- */

int main(void)
{
	/* To create pipes, we need an array of two descriptors, the reading
 	 * end (0) and writing end (1). */
	int pipefds[2];

	if (pipe(pipefds) == -1)
		err(EXIT_FAILURE, "pipe");

	/* With pipe created, we can now start child and the parent. */
	pid_t pid = fork();

	switch (pid) {
	case -1:
		err(EXIT_FAILURE, "fork");

	case 0:
		child(pipefds);
		exit(EXIT_SUCCESS);
	}

	printf("Parent PID %d\n", getpid());
	printf("Child PID %d\n", pid);
	parent(pipefds);
	wait(NULL);

	return EXIT_SUCCESS;
}


void parent(int pipefds[2])
{
	/* The parent process will be in the role of a «reader». However, at this
 	 * point, it still has both ends opened. To make sure we do not
 	 * cause deadlock or that we get ‹EOF› properly, we close the writing
 	 * end as soon as possible. */
	close(pipefds[1]);

	/* Also remember that pipes are unidirectional. We cannot send and receive
	 * data using one pipe; reading would probably split the data between
	 * the readers. We would have to use two separate pipes for bidirectional
	 * communication, or use UNIX socket pair instead. */

	/* Now we read data. For simplicity, we will read characters one by one,
 	 * though this is «very» inefficient. */
	char buffer;

	int rd;
	while ((rd = read(pipefds[0], &buffer, 1)) > 0)
		write(STDOUT_FILENO, &buffer, 1);

	/* Write the final ‹\n›. Just in case. */
	write(STDOUT_FILENO, "\n", 1);

	/* At this point, either an EOF condition occurred (‹rd› is 0) or there
 	 * was an error (‹rd› is -1). */
	if (rd == -1)
		err(EXIT_FAILURE, "read");

	/* We are done, we can close the remaining end. */
	close(pipefds[0]);
}

void child(int pipefds[2])
{
	/* Again as with the parent, we are only going to write to the pipe.
	 * Thus we will close the reading end as soon as possible. */
	close(pipefds[0]);

	/* Prepare a simple message. */
	char data[64];
	snprintf(data, sizeof(data), "Hello from %d", getpid());

	/* Write the message into the buffer. */
	ssize_t wr = write(pipefds[1], data, strlen(data));
	assert(wr == strlen(data));

	/* Right now, the parent could have read the message, but would still
	 * be probably blocked on another ‹read()›. Since we are probably the only
	 * process that holds the writing end now¹, closing the writing end
	 * will generate ‹EOF› condition in all readers. In this case, in
	 * the parent process. */
	close(pipefds[1]);
}
