#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include <err.h>
#include <fcntl.h>
#include <unistd.h>

/* This program demonstrates how sparse files (and holes within them)
 * can be created. */

int main(int argc, char *argv[])
{
	if (argc != 2) {
		fprintf(stderr, "usage: %s FILE\n", argv[0]);
		return EXIT_FAILURE;
	}

	int fd = open(argv[1], O_RDONLY);
	if (fd == -1)
		err(EXIT_FAILURE, "%s", argv[1]);

	/* There are two ways to detect size of a file:
	 *
	 * 1. Using ‹fstat()› and checking ‹st_size›,
	 * 2. Using ‹lseek()› to move offset to the end.
	 *
	 * The latter is only possible on files that support seeking.
	 * However, no other file types can contain holes, thus we can make
	 * this assumption. */
	off_t eof = lseek(fd, 0, SEEK_END);

	if (eof == -1)
		err(EXIT_FAILURE, "seek %s", argv[1]);

	/* Move back to the beginning of the file. */
	off_t current = lseek(fd, 0, SEEK_SET);

	if (current == -1)
		err(EXIT_FAILURE, "seek %s", argv[1]);

	size_t holes = 0;
	while (current != eof && current >= 0) {
		/* Find a next hole. */
		off_t hole = lseek(fd, current, SEEK_HOLE);

		if (hole == eof)
			break;

		++holes;

		/* Now we need to skip the data to see how big the hole is.
		 * If the hole is at the end of the file, though, we have to cover
		 * a special case (below). */
		current = lseek(fd, hole, SEEK_DATA);

		/* If the hole is at the end of the file, ‹SEEK_DATA› will cause
		 * ‹ENXIO› error. We will silently convert it to ‹eof› offset. */
		if (current == -1 && errno == ENXIO) {
			current = eof;
			errno = 0;
		}

		printf("%9ld -> %9ld  %9ld B\n", hole, current, current - hole);
	}

	if (current == -1)
		warn("seek %s", argv[1]);

	/* Done. */
	close(fd);
	return holes > 0 ? EXIT_SUCCESS : EXIT_FAILURE;
}
