#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>

#include <err.h>
#include <fcntl.h>
#include <poll.h>
#include <unistd.h>

/* The first example of this seminar will deal with events on file descriptors.
 * It is sometimes the case that we have to deal with multiple file descriptors
 * at once, but we cannot «block» our program.
 *
 * For example, say we are implementing a system service (a daemon):
 *
 *  1. We get a file descriptor for a client process¹ which will eventually
 *     send us some command to process.
 *  2. While processing the command, another client process asks us to do
 *     something else.
 *  3. The operation requires to read from another descriptor, for example
 *     to communicate with some other system daemon.
 *  4. Finally, responses need to be written back to client descriptors.
 *
 * Without any special care, we could get into the following situations:
 *
 * • We try to ‹read()› from client descriptor, but the client may not yet be
 *   ready to transmit the request, or has been delayed by the process scheduler.
 *   In either way, the service blocks on the ‹read()› while other clients
 *   may be able to transmit a request.
 * • We may block on ‹write()› to other daemon's descriptor, perhaps because
 *   it is not able to handle too many requests at once.
 * • We may block on ‹write()› to client descriptor for the same reasons.
 *
 * Surely, these situations are not very pleasant. One way we might think to
 * be dealing with them would be using ‹O_NONBLOCK› flag for ‹open()›, which
 * would allow us to loop for all descriptors and try ‹read()› or ‹write()›
 * on them, and handling ‹EAGAIN›. However, this would be highly inefficient,
 * using CPU even when no descriptor can be interacted with.
 *
 * Luckily, POSIX offers a way to monitor descriptors in bulk and tell us which
 * operations are available; ‹select()› and ‹poll()›.
 *
 * ¹ This is often implemented using UNIX sockets, as we will see in the
 *   11th seminar. */

/* As ‹select()› is older and a bit more cumbersome to use, we will focus
 * on ‹poll()› here. The following program opens a pipe specified as
 * the program argument and then reads both from the standard input and
 * the pipe, copying both to the standard output. */


/* Main
 * ---- */

int main(int argc, char *argv[])
{
	if (argc != 2) {
		fprintf(stderr, "usage: %s PIPE\n", argv[0]);
		return EXIT_FAILURE;
	}

	int pipe_fd = open(argv[1], O_RDONLY);
	if (pipe_fd == -1)
		err(EXIT_FAILURE, "%s", argv[1]);

	/* Once we are ready, we need to inform ‹poll()› about the descriptors
	 * we want to monitor. This is done using an array of ‹struct pollfd›
	 * structures: */
	struct pollfd fds[] = {
		/* The ‹fd› attribute names the descriptor to monitor, and
		 * the ‹events› attribute tells kernel which events we are interested in. */
		{ .fd = STDIN_FILENO, .events = POLLIN },
		/* In this case, ‹POLLIN› means we want to know the descriptor is
		 * readable. Note that generally there is no way to tell how much data
		 * is available. Other useful event is ‹POLLOUT›, which is raised
		 * when the descriptor can be written to. */

		/* We will also add a descriptor for the pipe. */
		{ .fd = pipe_fd, .events = POLLIN },
	};

	/* For generality, we will compute the number of elements here: */
	size_t nfds = sizeof(fds) / sizeof(fds[0]);

	/* In addition to this number we will also hold another counter, where
	 * we will remember number of «active» descriptors. We will explain that
	 * in a moment. */
	size_t active_fds = nfds;

	/* Now we are ready to monitor the descriptors. We will do so until
	 * both descriptors are closed. The third argument to ‹poll()› specifies
	 * a timeout:
	 *
	 * • If timeout is set to a positive value, the function will return after
	 *   this timeout even if no events occurred.
	 * • If it is set to 0, the function will check the events, but will return
	 *   immediately even if no events happened.
	 * • If it is set to a negative value, the function will wait indefinitely
	 *   until at least one event occurs or the function gets interrupted
	 *   by a signal.
	 *
	 * In addition to return values of ‹poll()›, we will also check that there
	 * is at least one active FD to wait for. */
	while (active_fds > 0 && poll(fds, nfds, -1) != -1) {
		/* If ‹poll()› succeeds, it will tell us the count of descriptors that
		 * have at least one event (note this number can be zero if we use
		 * positive timeout). We would still need to check every descriptor
		 * anyway though. */

		/* The call to ‹poll()› never modifies ‹fd› or ‹event› attributes
		 * (a huge advantage over ‹select()›). Instead, occurred events are
		 * recorded in ‹revents› attribute. If no events occurred, the attribute
		 * will be set to 0. */
		for (size_t i = 0; i < nfds; ++i) {
			/* If a descriptor is set to a negative number, ‹poll()› ignores it.
			 * This will become useful later in this loop; for now we will
			 * just ignore such descriptors, as well as those that have no
			 * events. */
			if (fds[i].fd < 0 || fds[i].revents == 0)
				continue;

			/* With corner cases out of the way, we can now inspect events
			 * that had occurred. */

			/* We will create a variable that will indicate that the descriptor
			 * is to be closed. We will use it to clean up the descriptor in
			 * one place instead of duplicating the code several times. */
			bool cleanup = false;

			if (fds[i].revents & POLLIN) {
				/* We can read from the descriptor. */
				char buffer[16];
				ssize_t rd = read(fds[i].fd, buffer, sizeof(buffer));

				/* While it may seem to be a bit counter-intuitive, ‹EOF›
				 * is something that needs to be “read” from the descriptor
				 * when it occurs on some files (e.g. terminals). Therefore we
				 * also need to handle this case. On a similar note, an error
				 * may occur for ‹read()›. */
				if (rd <= 0) {
					printf("[%d] %s\n", fds[i].fd, rd == 0 ? "EOF" : "ERR");
					cleanup = true;
				} else {
					/* Write what we have read. Note that we are not using
					 * ‹%s› directly, because we have no guarantee that the
					 * buffer ends with ‹\0› (actually the only way it could
					 * be there is that someone wrote it explicitly on the other
					 * end of the descriptor). On the other hand, we should
					 * not mix ‹write()› and ‹printf()> calls, so we will
					 * use «precision modifier» for ‹printf()› to limit the
					 * number of bytes to be written from ‹buffer›. */
					printf("[%d] IN %.*s\n", fds[i].fd, (int) rd, buffer);
				}
			/* Even without asking for them, ‹poll()› can always check and
			 * return ‹POLLHUP›, ‹POLLERR› and ‹POLLNVAL›. */
			} else if (fds[i].revents & POLLHUP) {
				/* The ‹POLLHUP› indicates for certain files (pipe, socket)
				 * that the other end has been closed. Technically, this event
				 * may occur at the same time as ‹POLLIN›, which merely says
				 * that there is data to be read and no more data will be
				 * available. Since we got into this branch, it means that
				 * ‹POLLIN› was not set, and we can close the handle. */
				printf("[%d] HUP\n", fds[i].fd);
				cleanup = true;
			} else if (fds[i].revents & POLLERR) {
				/* This event occurs when an error occurs on the descriptor. */
				printf("[%d] ERR\n", fds[i].fd);
				cleanup = true;
			} else {
				/* Finally, ‹POLLNVAL› means we provided a positive value in
				 * ‹fd›, but it is not a valid open descriptor. This is an error
				 * on our part. */
				fprintf(stderr, "[%d] NVAL\n", fds[i].fd);
				abort();
			}

			if (cleanup) {
				close(fds[i].fd);
				active_fds--;
				/* Here is where the negative descriptor comes into the play.
				 * Since ‹poll()› ignores negative descriptors, we can
				 * simply set it to -1 instead of moving elements of ‹fds›
				 * and changing ‹nfds›. */
				fds[i].fd = -1;
			}
		}
	}

	/* We got out from the above loop either because all descriptors were
	 * closed or ‹poll()› failed: */
	if (active_fds != 0)
		err(EXIT_FAILURE, "poll");

	/* When you run ‹Makefile›, it should also create a pipe called ‹hose›.
	 * Open a «new» terminal window and run
	 *
	 *		cat >${path_to_the_seminar}/hose
	 *
	 * Then, in the original terminal window, run
	 *
	 * 		./poll hose
	 *
	 * You should observe that the program is able to “switch” reading
	 * between the two descriptors without blocking on each one. Even more,
	 * we managed to do so without ‹O_NONBLOCK›.
	 *
	 * Try to observe the program when you specify an input much larger than
	 * the capacity of the buffer. Will the program read only the first part
	 * of the message and ignore the rest? Or will the ‹POLLIN› event persist
	 * until all the data is read? For example, try this:
	 *
	 * 		# This will create 21 bytes
	 * 		perl -E 'say "text" x 5' | cat >${path_to_the_seminar}/hose
	 *
	 * You are recommended to inspect the program again with ‹strace›. */

	return EXIT_SUCCESS;
}
