#include <assert.h>
#include <ctype.h>
#include <limits.h>
#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <err.h>
#include <fcntl.h>
#include <sys/epoll.h>
#include <sys/types.h>
#include <unistd.h>

/* In this example, we will show how ‹epoll› can be used to manipulate
 * several descriptors a bit more easily than with ‹poll()› on a real-life
 * example where the number of descriptors changes over time, and even
 * grows beyond the original number.
 *
 * One such example is a network server. We have not covered networks yet,
 * so we will only cover the main idea: The server holds one descriptor for
 * accepting clients. Once a client connects, we get a new descriptor
 * for communication with that client. Hence, we do not know a priori how
 * many descriptors we will need to monitor. With ‹poll()› we would need
 * to possibly reallocate ‹struct pollfd› array with too many descriptors,
 * and checking the type of the descriptor that has become ready could
 * be cumbersome after a while as well.
 *
 * Let's see how we can tackle this problem with ‹epoll›. */

/* The following headers are used for sockets. */
#include <netinet/in.h>
#include <sys/socket.h>


/* Forward declarations
 * -------------------- */
int start_server(uint16_t *port);


/* Main
 * ---- */

/* This structure along with macros below will be used to attach extra
 * data to the ‹epoll› structure describing the file descriptor. We will
 * use this to “tag” a descriptor with a type: */
enum fdtype {
	FD_STDIN,  /* Standard input descriptor. */
	FD_SERVER, /* Server socket. */
	FD_CLIENT, /* Connected client. */
};

/* The following macros will only work if we ‹int› that stores descriptors
 * and ‹enum fdtype› can fit into 64bit number. */
_Static_assert(sizeof(int) + sizeof(enum fdtype) <= sizeof(uint64_t),
		"Cannot use uint64_t to store two numbers; another hack is needed");

#define EDATA_ENCODE(FDTYPE, FD) \
	(((uint64_t) FDTYPE << sizeof(int) * CHAR_BIT) | FD)
#define EDATA_DECODE_TYPE(VALUE) \
	((VALUE) >> (sizeof(int) * CHAR_BIT))
#define EDATA_DECODE_FD(VALUE) \
	((VALUE) & INT_MAX)

int main(void)
{
	/* Before we actually call and use ‹epoll›, we will discuss different types
	 * of events that ‹epoll› can monitor:
	 *
	 * • Level-Triggered Events
	 *   We have seen with ‹poll()› that once ‹POLLIN› is triggered, a call to
	 *   ‹poll()› will leave ‹POLLIN› set if we do not read all the data.
	 *   This event is triggered as long as the state persists (“the level
	 *   is elevated”).
	 *
	 * • Edge-Triggered Events
	 *   A speciality of ‹epoll› is that it can fire events only when a status
	 *   of a descriptor «changes». That is, we get an event only when
	 *   “would block” statue changes to “can be read”, but not else.
	 *   This is a bit more efficient in scenarios where we have non-blocking
	 *   descriptors.
	 *
	 * Consider a descriptor whose contents, in time, evolve like this
	 * (E = empty, F = full):
	 *
	 *        ╻
	 *      F ┫                  ╭────╮
	 *        ┃  ╭─────╮         │    │
	 *        ┃  │     │     ╭───╯    │  ╭──╮
	 *        ┃  │     │     │        ╰──╯  │
	 *      E ┻━━┷━━━━━┷━━━━━┷━━━━━━━━━━━━━━┷━━━━
	 *           ↑           ↑                     Edge-Triggered
	 *           ╌╌╌╌╌>      ╌╌╌╌╌╌╌╌╌╌╌╌╌╌>       Level-Triggered
	 *
	 * For Level-Triggered Events, any call to ‹epoll_wait()› when there is
	 * data waiting for the descriptor (marked as dashed line) will return
	 * a ‹EPOLLIN› event.
	 *
	 * On the other hand, Edge-Triggered ‹EPOLLET› request will only get
	 * triggered when a «change» is observed on the descriptor, indicated
	 * by ‹↑› arrows. */

	/* We will use the standard input to read commands from. Since we need it
	 * to be non-blocking, we will use ‹fcntl()› to change its flags.
	 * This system call will also be explained later in this seminar. */
	if (fcntl(STDIN_FILENO, F_SETFL, O_NONBLOCK) == -1)
		err(EXIT_FAILURE, "fcntl()");

	/* Then, we need a network socket. */
	uint16_t port;
	int sockfd = start_server(&port);

	printf("Server is listening on all interfaces, port %u\n", port);

	/* Now, we need the kernel to create an epoll structure: */
	int epfd = epoll_create1(0);
	if (epfd == -1)
		err(EXIT_FAILURE, "epoll_create()");

	/* We can add our watched descriptors. Do describe what we want, we will
	 * create a structure similar to ‹struct pollfd›, but we will reuse this one
	 * multiple times: */
	struct epoll_event ev = {
		/* ‹EPOLLET› modifies the event kind to Edge-Triggered. */
		.events = EPOLLIN | EPOLLET,
		/* This structure contains one more attribute, ‹data› union.
		 *
		 * Generally it is used to store the descriptor to be returned from
		 * ‹epoll_wait()›, but we can squeeze one more information into it:
		 * Instead of ‹data.fd› we will use ‹data.u64›, with the upper
		 * 8 bits encoding the type of the descriptor. While it is not very
		 * clean, it will demonstrate the use of this attribute. A cleaner
		 * (but a bit more cumbersome) approach would be to use ‹ptr› field
		 * pointing to some allocated structure, but then it would need dynamic
		 * allocation which we wanted to avoid in the first place. */
	};

	/* To monitor the descriptors, we add them using ‹epoll_ctl()› function. */
	ev.data.u64 = EDATA_ENCODE(FD_STDIN, STDIN_FILENO);
	if (epoll_ctl(epfd, EPOLL_CTL_ADD, STDIN_FILENO, &ev) == -1)
		err(EXIT_FAILURE, "epoll_ctl() stdin");

	ev.data.u64 = EDATA_ENCODE(FD_SERVER, sockfd);
	if (epoll_ctl(epfd, EPOLL_CTL_ADD, sockfd, &ev) == -1)
		err(EXIT_FAILURE, "epoll_ctl(): pipe");

	/* Monitoring is done by a call to ‹epoll_wait()›, which will return the
	 * observed events in an array. We will reuse ‹ev› here as an array of
	 * size 1. This is not very efficient, but will suffice for our example.
	 *
	 * As with ‹poll()›, ‹epoll_wait()› returns the number of descriptors
	 * with an event, or ‹-1› on error. We will also specify timeout ‹-1›
	 * to wait indefinitely. */
	while (epoll_wait(epfd, &ev, 1, -1) != -1) {
		/* When ‹epoll_wait()› returns, it fill ‹ev› with the number of
		 * requested events. */
		enum fdtype type = EDATA_DECODE_TYPE(ev.data.u64);
		int fd = EDATA_DECODE_FD(ev.data.u64);

		/* The following branches should ideally be implemented as separate
		 * functions. This would, however, break the linearity of this
		 * explanation, and we do not want that here. So we will overlook this
		 * defect for now, and just keep in mind that his should be decomposed
		 * better if you were to write this as a production-grade code. */
		if (type == FD_STDIN) {
			/* When ‹STDIN› becomes ready, we will read a single number and
			 * interpret it as the target descriptor to write a message to.
			 * For simplicity, we will expect that the first part of the message
			 * fits 64 bytes. */
			char buffer[64];
			ssize_t rd = read(fd, buffer, sizeof(buffer) - 1);

			/* At this point ‹rd == -1› would indicate an error, because
			 * ‹epoll› told us there is some data. */
			if (rd == -1)
				err(EXIT_FAILURE, "read(): stdin");

			/* Before we use ‹buffer› as a string, we need to add a null byte. */
			buffer[rd] = 0;

			/* On empty input or ‹exit›, we will end the program. */
			if (rd == 0 || strcmp(buffer, "exit\n") == 0) {
				printf("[%2d] Bye\n", fd);
				break;
			}

			/* To fetch the number and the message at the same time, we will use
			 * ‹strtol()›'s ability to point the ‹endp› after the number: */
			char *msg;
			int target = strtol(buffer, &msg, 10);
			if (msg == buffer || target <= 0) {
				fprintf(stderr, "[%2d] Invalid target specification\n", fd);
				continue;
			}

			/* Skip spaces at the beginning of the message. */
			while (*msg != '\0' && isspace(*msg))
				msg++;

			/* We try to write the message to the target descriptor. */
			if (write(target, msg, strlen(msg)) == -1)
				warn("[%2d] Cannot send message to %d", fd, target);

			/* Dump the rest of the data from terminal to the socket.
			 * This will of course not work as intended if the input is
			 * connected to a file (e.g. ‹./epoll <file›), but to keep
			 * this example simple, we will not fix that. */
			if (splice(fd, NULL, target, NULL, 4096, 0) == -1 && errno != EAGAIN)
				err(EXIT_FAILURE, "sendfile()");
		} else if (type == FD_SERVER) {
			/* We are using non-blocking descriptors, so we will try to accept
			 * as many clients as possible. */
			int clientfd;

			struct sockaddr_in addr;
			socklen_t addrlen = sizeof(addr);

			/* This might surprise you a little bit -- we got ‹EPOLLIN› event
			 * on this descriptor, but we do not read from it.
			 *
			 * This is how connections from clients are accepted. Imagine
			 * that ‹accept()› calls ‹read()› and returns a client descriptor
			 * (it doesn't, but it technically «could» on UNIX systems that
			 * try really hard to treat descriptors as files). */
			while ((clientfd = accept(fd, (struct sockaddr *) &addr, &addrlen)) != -1) {
				/* Before we use the descriptor, we need to mark it with
				 * ‹O_NONBLOCK› flag. This can either be done by ‹accept4()›
				 * system call, ‹setsockopt()› or ‹fcntl()›. The latter will be
				 * explained later in this seminar. */
				if (fcntl(clientfd, F_SETFL, O_NONBLOCK) == -1)
					err(EXIT_FAILURE, "fcntl(): Cannot unblock %d", clientfd);

				/* Set up monitoring of the new client. */
				struct epoll_event event = {
					.events = EPOLLIN | EPOLLET,
					.data = { .u64 = EDATA_ENCODE(FD_CLIENT, clientfd) },
				};

				if (epoll_ctl(epfd, EPOLL_CTL_ADD, clientfd, &event) == -1)
					err(EXIT_FAILURE, "epoll_ctl(): Client %d", clientfd);

				/* Note that there is a better way to get a string from
				 * an IP address, using ‹getaddrinfo()›. However, that would
				 * require a bit more headers and incantations, so for now,
				 * we will do this dirty trick. */
				printf("[%2d] * Connection from %d.%d.%d.%d:%d\n", clientfd,
						(addr.sin_addr.s_addr & 0x000000ff),
						(addr.sin_addr.s_addr & 0x0000ff00) >>  8,
						(addr.sin_addr.s_addr & 0x00ff0000) >> 16,
						(addr.sin_addr.s_addr & 0xff000000) >> 24,
						ntohs(addr.sin_port));
			}

			if (errno != EAGAIN)
				err(EXIT_FAILURE, "accept()");
		} else if (type == FD_CLIENT) {
			/* Finally, for client descriptor, we need to differentiate
			 * whether ‹EPOLLIN› or ‹EPOLLHUP› was detected. */
			if (ev.events & EPOLLIN) {
				char buffer[64];
				ssize_t rd;
				while ((rd = read(fd, buffer, sizeof(buffer))) > 0) {
					printf("[%2d] > %.*s", fd, (int) rd, buffer);
				}

				if (rd == 0) {
					/* The client is done. */
					printf("[%2d] # Client disconnected\n", fd);
					/* We will “fake” ‹EPOLLHUP› so that the next branch gets
					 * executed. */
					ev.events |= EPOLLHUP;
				} else if (errno != EAGAIN) {
					warn("[%2d] # Read error", fd);
					ev.events |= EPOLLHUP;
				}
			}

			if (ev.events & EPOLLHUP) {
				/* At this point we will close the descriptor and stop
				 * the monitoring. */
				if (epoll_ctl(epfd, EPOLL_CTL_DEL, fd, NULL) == -1)
					err(EXIT_FAILURE, "epoll_ctl(): Cannot remove %d", fd);

				close(fd);
			}
		}
	}

	/* Once we are out of the loop, we close the server socket and then
	 * exit the program, severing the connection with remaining clients.
	 * This is not the best way to end the network server, but we will
	 * discuss “graceful shutdown” when we really get to sockets next time. */
	close(sockfd);
	return EXIT_SUCCESS;
}

/* To test the program, run it and observe the first message, where the server
 * will print out the port it's listening to.
 *
 * Then run
 *
 * 		$ nc localhost PORT
 *
 * from your own computer. Depending on your distribution, ‹nc› can be available
 * as ‹ncat› instead.  You can also test the program on Aisa, in which case use
 *
 * 		$ nc aisa.fi.muni.cz PORT
 *
 * │ IMPORTANT: DO NOT LEAVE THE PROGRAM RUNNING ON AISA UNSUPERVISED!
 * │ Rules of use prohibit long-running network processes on Aisa!
 *
 * When you connect to the server, you should see a message
 *
 * 		[ 5] * Connection from ‹IP›:‹PORT›
 *
 * If you write something in the client's terminal, the message will appear
 * at the server:
 *
 * 		[ 5] > Hello!
 *
 * In the server window, you can use the number in ‹[…]› (the descriptor)
 * to write to the client:
 *
 * 		5 Hi!
 *
 * will send ‹Hi!› to the client connected to descriptor ‹5›.
 *
 * Open multiple terminal windows and connect to the client multiple times.
 * Observe (also using ‹strace›) how ‹epoll› works.
 *
 * TIP: If you look closely to the code that sends messages to clients,
 *      you can see that it does not check whether the descriptors exist,
 *      or whether they are sockets. What happens when you write a message
 *      to ‹1›? What will happen when you write a message to ‹3›? */


/* This function creates a server socket that listens on a random port
 * on all interfaces. For now, consider this a black magic. */
int start_server(uint16_t *port)
{
	int sockfd = socket(AF_INET, SOCK_STREAM | SOCK_NONBLOCK, 0);
	if (sockfd == -1) {
		warn("socket()");
		return -1;
	}

	if (listen(sockfd, 16) == -1) {
		warn("listen()");
		goto error;
	}

	if (port != NULL) {
		struct sockaddr_in addr;
		socklen_t addrlen = sizeof(addr);

		if (getsockname(sockfd, (struct sockaddr *) &addr, &addrlen) == -1) {
			warn("Cannot determine socket address");
			goto error;
		}

		assert(addr.sin_family == AF_INET);
		*port = ntohs(addr.sin_port);
	}

	return sockfd;

error:
	close(sockfd);
	return -1;
}
