#include <inttypes.h>
#include <limits.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <err.h>
#include <sys/inotify.h>
#include <unistd.h>

/* This example shows how ‹inotify› works. The program will add a single watch
 * and then will display events for the monitor.
 *
 * The idea of ‹inotify› is that it provides two descriptors:
 *
 * ◦ The ‹inotify› descriptor represents the real resource kernel holds for us.
 *   It is returned by ‹inotify_init[1]()› and we must not lose it.
 * ◦ When we tell ‹inotify› to monitor a path, it will provide us with a watcher
 *   descriptor. This is not a real file descriptor (e.g. you will not find it
 *   in ‹/proc/self/fd›), it is a watch identifier we can potentially use later
 *   to remove the watch. It is also returned in events. We can throw it away
 *   freely, if we want.
 *
 * Thus, we make requests to ‹inotify()› using the main descriptor, and we
 * optionally identify specific watches by watch descriptors. Once we set up
 * monitored paths using ‹inotify_add_watch()›, we can read events from
 * the real descriptor. */

/* A file ‹events.c› contains the implementation of a function that translates
 * event mask into a printable string. It is not that much useful for this demo. */
#include "events.h"

int main(int argc, char *argv[])
{
	if (argc > 2) {
		fprintf(stderr, "usage: %s [DIRECTORY]\n", argv[0]);
		return EXIT_FAILURE;
	}

	/* We will monitor either the current directory, or provided argument. */
	const char *path = argc == 1 ? "." : argv[1];

	/* The following call will create an instance of inotify monitor for our
	 * process, and returns its descriptor. */
	int in_fd = inotify_init();
	if (in_fd == -1)
		err(EXIT_FAILURE, "inotify_init()");

	/* This operation sets up a watch for the given path. We can name events
	 * using ‹|›, e.g. ‹IN_OPEN | IN_WRITE | …›, but we can just as well
	 * use ‹IN_ALL_EVENTS› to monitor them all. Note, however, that this is
	 * not ideal for production and we should only monitor events we really
	 * want. Unwanted events delay the kernel and our program, and have a higher
	 * chance to get us into a situation where we observe an outdated event.
	 * For this demo, though, this will do just fine.
	 *
	 * This operation returns a watch descriptor. We are not going to use it,
	 * so we will not save it to a variable. */
	if (inotify_add_watch(in_fd, path, IN_ALL_EVENTS) == -1)
		err(EXIT_FAILURE, "inotify_add_watch()");

	/* Events are represented by ‹struct inotify_event›. This is structure
	 * has a Flexible Array Member, so we have to account for it as well.
	 * The buffer provided to ‹read()› from ‹inotify› descriptor must be
	 * able to hold at least one event, otherwise it will fail. Thus, we
	 * will use the longest possible size, assuming that no file can have
	 * name longer than ‹NAME_MAX›. This is also mentioned in ‹man 7 inotify›. */
	char buffer[sizeof(struct inotify_event) + NAME_MAX + 1];

	ssize_t rd;
	while ((rd = read(in_fd, buffer, sizeof(buffer))) > 0) {
		/* Why didn't we allocate a buffer of type ‹struct inotify_event›
		 * directly? Because ‹read()› will always return «at least» one full
		 * event. If more events are pending, kernel will copy as many events
		 * as possible into the buffer, without splitting them. Since these
		 * structures have variable size (due to FAM), we cannot treat it
		 * as an array. */
		struct inotify_event *event = (struct inotify_event *) buffer;

		/* This will print information from the event. */
		printf("Inotify event:\n");
		printf(" ◦ mask: %s\n", events_to_str(event->mask));

		/* Some events do not contain a file name, so we will check for this
		 * before printing the name. */
		if (event->len > 0)
			printf(" ◦ name: %s\n", event->name);

		/* Moving a file in watched paths generates two events, ‹IN_MOVED_FROM›
		 * and ‹IN_MOVED_TO›. Kernel tells us that these two events are linked
		 * by setting a ‹cookie› attribute in both to a same value. */
		if (event->cookie != 0)
			printf(" ◦ cookie: %04" PRIx32 "\n", event->cookie);

		/* Now, for the final catch. If multiple events are pending and there
		 * is enough space in the buffer, kernel will copy more than one event
		 * into the buffer. As a hint on how to deal with this, we will detect
		 * such situation -- kernel will tell us how many bytes of the buffer
		 * it has filled. The size of the first event is the size of the base
		 * structure and size of the ‹name› field, given in ‹len›. If these
		 * are not equal, there must be more events. */
		if (sizeof(*event) + event->len < rd) {
			warnx("Events dropped (used %zd, read %zd)",
					sizeof(*event) + event->len, rd);
		}
	}

	/* The loop above will never end, use ‹Ctrl-C› to exit the program. */

	/* Try the following scenarios:
	 *
	 * 1. Create and delete a file in the watched directory.
	 * 2. Create a file inside a watched directory, then rename it.
	 * 3. Before starting the watcher, create a directory. Start the watcher
	 *    for that directory and then delete it. */
}
