#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>

#include <err.h>
#include <sys/timerfd.h>
#include <unistd.h>

/* This example shows how timer descriptor works.
 *
 * We have already seen ‹alarm()› function, which sends ‹SIGALRM› to the process
 * after its countdown expires. There are also more generic per-process timers,
 * accessed by ‹timer_create()›, which offer finer time granularity than
 * ‹alarm()› and more options, e.g. starting a thread instead of sending
 * a signal.
 *
 * The ‹timerfd_create()› returns a timer descriptor, which can be set up
 * to fire repeated events. When the timer gets readable (yes, you can use
 * it even with ‹poll()›), it will return a 64bit number indicating how many
 * times the timer went off before the process got to read it. */

int main(void)
{
	/* The ‹itimerspec› structure describes the timer. It has two attributes
	 * of type ‹struct timespec›, containing ‹tv_sec› and ‹tv_nsec› attributes.*/
	struct itimerspec spec = {
		/* The ‹it_value› specifies the delay before the timer goes off the
		 * first time. At least one value must be non-zero, otherwise the timer
		 * will be disarmed. This will set the timer to fire off the first
		 * time after two seconds: */
		.it_value = { .tv_sec = 2 },
		/* The ‹it_interval› will rearm the timer for this amount of time when
		 * it goes off. This attribute can be set to zero; in that case the
		 * timer will go off only once. This value will set the timer to go
		 * off every 250 ms (the value is in ns): */
		.it_interval = { .tv_nsec = 250000000 },
	};

	/* As with ‹timer_create()›, there are multiple system clocks to use for
	 * the timer. The ‹CLOCK_MONOTONIC› is a clock that always increases in
	 * value, as opposed to ‹CLOCK_REALTIME› which measures wall-clock time
	 * and can “jump”, e.g. due to adjustments by NTP daemon.
	 *
	 * While POSIX does not specify what value ‹CLOCK_MONOTONIC› measures
	 * (just that it always increases), Linux uses this clock to measure
	 * the number of seconds since boot. For more alternatives see
	 * ‹man 3 timer_create›; for Linux-specific details ‹man 2 clock_gettime›. */
	int timer_fd = timerfd_create(CLOCK_MONOTONIC, 0);
	if (timer_fd == -1)
		err(EXIT_FAILURE, "timerfd_create()");

	/* The following function arms the timer with the specified value.
	 * When the timer goes off, no signal is sent -- the descriptor will instead
	 * become readable and will yield ‹1›, or more precisely, the number of
	 * times the timer went off since it started or last ‹read()› from
	 * the descriptor. */
	if (timerfd_settime(timer_fd, 0, &spec, NULL) == -1)
		err(EXIT_FAILURE, "timerfd_settime()");

	uint64_t count;
	/* This call blocks until the timer goes off at least once. When the call
	 * returns, the ‹count› will hold the number of times the timer went off. */
	while (read(timer_fd, &count, sizeof(count)) != -1) {
		printf("timer: %" PRIu64 "\n", count);

		/* Try to run the code as it is now. You should see that the program
		 * first hangs for 2 seconds, and after that, it prints a line four
		 * times per second. */

		/* Cancel the program with SIGINT ‹Ctrl-C› or SIGQUIT ‹Ctrl-\›.
		 * Then, uncomment this line: */
		//sleep(1);

		/* After the timer goes off the first time, you should now see that
		 * instead of four lines per second, there is only one, but the counter
		 * value is 4. */

		/* Finally, comment out ‹sleep()› above again and instead remove zeroes
		 * in ‹tv_nsec› of ‹it_interval› attribute (i.e. set it to ‹25›).
		 * What will the output be now? (If you still see lines with 1s
		 * extremely fast, set the value even lower). */
	}
}
