#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include <err.h>
#include <fcntl.h>
#include <unistd.h>

/* The libev library, or Event-Loop library, provides uniform interface
 * for delivering events for various system objects. */

/* The core idea revolves around “watchers”. A watcher is a structure that
 * describes what is being monitored and which events should be delivered.
 * When such event occurs, libev calls a “callback” function. */

/* The program here does similar things to previous programs in this seminar:
 *
 * • Watch standard input and a pipe and copy data from them when available.
 * • Display a message when ‹SIGINT› is sent to the process. */

/* We only need one extra header. */
#include <ev.h>


/* Forward declarations
 * -------------------- */

/* These will be described below. We should first look at setup. */

static
void cb_io(EV_P, ev_io *w, int revents);

static
void cb_sigint(EV_P, ev_signal *w, int revents);


/* Watchers
 * -------- */

/* We will hold watchers in a global structure, so that we can access them from
 * callbacks. We always get a watcher for “our” callback, but there is no way
 * to access other watchers if necessary. */
struct watchers {
	/* The ‹ev_io› watcher monitors I/O descriptors. */
	ev_io w_stdin;
	ev_io w_pipe;

	/* The ‹ev_signal› watcher monitors signals, akin to ‹signalfd()›. */
	ev_signal w_int;

	/* The program will end when both ‹w_stdin› and ‹w_pipe› stop (EOF).
	 * We will hold the count of active inputs here. */
	size_t active;
};

/* The global instance of ‹struct watchers›. */
struct watchers WATCHERS;


/* This function sets up watchers. */
static
bool watchers_setup(struct watchers *watch, char *pipe)
{
	/* To monitor the pipe, we will have to open it first. Note that the call
	 * will block until there is a process that has the writable end of
	 * the pipe open. */
	int pipe_fd = open(pipe, O_RDONLY);
	if (pipe_fd == -1) {
		warn("%s", pipe);
		return false;
	}

	/* For every type of watcher, there are functions (or macros):
	 *
	 * • ‹ev_TYPE_init()› to initialize the watcher. This is a short for:
	 *   ◦ ‹ev_init()› initializes general portion of the watcher,
	 *   ◦ ‹ev_TYPE_set()› initializes the watcher-specific type of the watcher.
	 *
	 *   For example,
	 *
	 *   	ev_io_init(W, CB, FD, FLAGS);
	 *
	 *   is a short-hand for
	 *
	 *   	ev_init(W, CB);
	 *   	ev_io_set(W, FD, FLAGS);
	 *
	 * • ‹ev_TYPE_start()› tells libev to monitor events for the watcher,
	 * • ‹ev_TYPE_stop()› then tells libev to stop watching for events. */

	/* The following two lines basically order to call ‹cb_io()› when ‹EV_READ›
	 * (equivalent of ‹POLLIN›) event occurs for ‹STDIN_FILENO› or ‹pipe_fd›
	 * respectively. */
	ev_io_init(&watch->w_stdin, cb_io, STDIN_FILENO, EV_READ);
	ev_io_init(&watch->w_pipe, cb_io, pipe_fd, EV_READ);

	/* Similarly, this calls ‹cb_sigint()› when ‹SIGINT› gets delivered. Note
	 * that we don't have to set up signal blocking; libev does that for us. */
	ev_signal_init(&watch->w_int, cb_sigint, SIGINT);
	return true;
}

/* This function starts watching for given events (though they will not get
 * delivered yet). Note the ‹loop› parameter, this is what represents
 * the event loop in the program. Imagine this to be a structure that holds
 * some representation of watchers and chooses appropriate backend (e.g. epoll)
 * when the loop is run. */
static
void watchers_start(struct watchers *watch, struct ev_loop *loop)
{
	/* ‹ev_TYPE_init()› does not actually start watching for the events yet.
	 * We need to tell libev which event loop will watch and deliver the events.
	 * Thus, we register the watcher with the loop using ‹ev_TYPE_start()›
	 * function. */
	ev_io_start(loop, &watch->w_stdin);
	ev_io_start(loop, &watch->w_pipe);
	watch->active = 2;

	ev_signal_start(loop, &watch->w_int);
}

int main(int argc, char *argv[])
{
	if (argc != 2) {
		fprintf(stderr, "usage: %s PIPE\n", argv[0]);
		return EXIT_FAILURE;
	}

	printf("Setting up\n");

	/* While libev program may have many event loops it can switch between,
	 * usually only one is needed. We can get it by ‹ev_default_loop()› call.
	 * This will allocate and return a structure that can be used to register
	 * watchers in ‹watchers_start()› above. */
	struct ev_loop *loop = ev_default_loop(0);
	if (loop == NULL)
		err(EXIT_FAILURE, "ev_default_loop()");


	/* Run functions above. */
	if (!watchers_setup(&WATCHERS, argv[1]))
		errx(EXIT_FAILURE, "Failed to set up watchers");

	/* The idea behind splitting setup and start functions is that we first
	 * make sure we can initialize all watchers (and open descriptors) before
	 * we use more resources for starting the watchers. But it is not necessary,
	 * we can just as well call ‹ev_io_init()› followed by ‹ev_io_start()›. */
	watchers_start(&WATCHERS, loop);

	printf("Starting event loop\n");

	/* Up until now, no events got actually delivered. This is because the
	 * event loop does not run yet, we need to start it by the following call: */
	ev_run(loop, 0);
	/* Imagine that this call prepares some ‹struct pollfd› array and calls
	 * ‹poll()›; or that it has simply called ‹epoll_ctl()› for every
	 * ‹ev_TYPE_init()› and running the loop simply means calling ‹epoll_wait()›
	 * in a literal loop. As you can see, one of the advantages of this library
	 * is that it abstracts the concrete backend implementation away, and leaves
	 * us relatively clean interface. */

	/* As regulars loop can be broken from by ‹break›, so can be the libev
	 * event loop. Once this happens, we exit the program. */
	return EXIT_SUCCESS;
}


/* Callbacks
 * --------- */

/* Callbacks in libev are functions called from ‹ev_run()› on a given event
 * loop. Every callback has three arguments:
 *
 * • The event loop. This parameter is usually only passed to other functions,
 *   so the library provides a convenient macro ‹EV_P› (Event Loop Parameter)
 *   for declaration, so that we don't have to care about it. The parameter
 *   can be then passed to other functions as ‹EV_A› (Event Loop Argument).
 * • The watcher of the appropriate type.
 * • ‹revents› with the similar role of ‹revents› attribute in ‹struct pollfd›. */

/* Let's start by the simpler one. The signal callback can react to a signal
 * being delivered. The type of watcher is ‹ev_signal› (the same one we
 * have in ‹WATCHERS›) and has an attribute ‹signum› to access the signal
 * number if we use the callback for multiple signals. */
static
void cb_sigint(EV_P, ev_signal *w, int revents)
{
	/* Let's say that we want to make sure user does not interrupt an important
	 * program, but still allow them to do so if they insist. We can do it
	 * easily by catching the first ‹SIGINT› and stopping the watcher
	 * immediately. The libev will unblock the signal, so the next delivery
	 * will exit the program. */

	/* ‹EV_A› is a macro that expands to an argument passed as ‹EV_P›.
	 * This function stops the watcher. */
	ev_signal_stop(EV_A, w);
	printf("Operation in progress, are you sure you want to abort?\n");
	printf("Press Ctrl-C again to abort the program.\n");
}

/* This is a function that should not surprise you -- just read bunch of data
 * from the descriptor and print it out. The important function for I/O watcher
 * is below this. */
static
bool _cb_read(int fd)
{
	char buffer[256];
	ssize_t rd = read(fd, buffer, sizeof(buffer));

	if (rd == 0) {
		printf("[%d] EOF\n", fd);
		return false;
	}

	if (rd == -1) {
		warn("[%d] Error", fd);
		return false;
	}

	/* Remove newline if found. */
	char *nl = memchr(buffer, '\n', (size_t) rd);
	if (nl != NULL)
		*nl = '\0';

	printf("[%d] %.*s\n", fd, (int) rd, buffer);
	return true;
}

/* The I/O watcher gets ‹ev_io› watcher which has ‹fd› and ‹events› attributes
 * available. */
static
void cb_io(EV_P, ev_io *w, int revents)
{
	/* As with ‹poll()›, we can test which events are available. Along ‹EV_READ›,
	 * we can also get ‹EV_ERROR›. There is no ‹EV_HUP› however. */
	if (revents & EV_READ) {
		if (!_cb_read(w->fd)) {
			/* If we get EOF or an error while reading, we stop the watcher
			 * and decrement the counter of active watchers. */
			ev_io_stop(EV_A, w);

			if (--WATCHERS.active == 0) {
				/* When we remove all watchers, it is time to stop the event
				 * loop. ‹ev_break()› function breaks the running event loop,
				 * similarly to ‹break› keyword. Event loops can be nested,
				 * so the second parameter tells which loops shall be broken
				 * out of: ‹EVBREAK_ALL› breaks (unsurprisingly) all event
				 * loops, while ‹EVBREAK_ONE› breaks only the inner-most. */
				ev_break(EV_A, EVBREAK_ALL);
			}
		}
	}

	if (revents & EV_ERROR)
		err(EXIT_FAILURE, "EV_ERROR for %d received", w->fd);
}

/* And that's it. Try running the program (note it is linked with ‹-lev›,
 * and you need to have ‹libev› available on your system).
 *
 * Try running the program with ‹strace› to see which backend the library
 * selected. On Linux, you will most likely see ‹epoll› functions.
 *
 * While we can certainly accomplish the same task with ‹poll()› or ‹epoll›,
 * this library makes it much more convenient. It even tries to do some extra
 * work for us, e.g. blocking the signal monitored by ‹ev_signal› watcher. */
