#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <aio.h>
#include <err.h>
#include <fcntl.h>
#include <bsd/sys/time.h>
#include <unistd.h>

/* The AIO stands for (POSIX) Asynchronous Input-Output. It allows a program
 * to schedule an asynchronous read or write operation while it continues
 * doing something else. It can arrange the interface to notify the program
 * once the operation finishes. */

/* Apart from (POSIX) AIO, there is also Linux AIO, sometimes called
 * KAIO as in Kernel AIO. It has interface declared in ‹libaio.h› system header
 * (POSIX AIO uses ‹aio.h› so that there is no confusion at all :κ:).
 * While non-portable, it is a bit “more asynchronous” than POSIX AIO,
 * as it does not need to fallback to threads to achieve asynchronous I/O.
 * Unfortunately, it is not very well documented.
 *
 * There is even more arcane interface, ‹io_uring›, that works by sharing
 * Submission Queue (SQ) and Completion Queue (CQ) with kernel using memory
 * mapping. This has a huge advantage of avoiding some system calls that
 * AIO and KAIO rely on, but requires a lot of setup. */

/* This example shows how the AIO can be used and compares it to direct
 * ‹write()› in ‹write.c›.
 *
 * The idea of the program is to create a single page of memory, fill it with
 * data and then write it to a file repeatedly using AIO. Direct write program
 * will do the same, except writing the block to the file synchronously.
 * Can you guess beforehand which one will be faster? */

/* The following file contains definitions for ‹PAGESIZE›, ‹FILESIZE›
 * and ‹PAGES_PER_TASK›. Read this file now before continuing. */
#include "config.h"

int main(int argc, char *argv[])
{
	if (argc != 2) {
		fprintf(stderr, "usage: %s FILE\n", argv[0]);
		return EXIT_FAILURE;
	}

	/* To avoid mistakes, we will not allow to overwrite existing files,
	 * simply by removing ‹O_TRUNC› you would expect for ‹fopen()›-ish ‹"w"›
	 * mode, and replace it with ‹O_EXCL›. */
	int fd = open(argv[1], O_WRONLY | O_CREAT | O_EXCL, 0600);
	if (fd == -1)
		err(EXIT_FAILURE, "%s", argv[1]);

	char *block = malloc(PAGESIZE);
	if (block == NULL)
		err(EXIT_FAILURE, "Cannot allocate block");

	/* To avoid copying unpopulated pages¹, fill the block with increasing
	 * values.
	 *
	 * ¹ As we will see in the last seminar, “Memory”, Linux does not actually
	 * populate (a fancy word for “allocate and map”) a new page until something
	 * writes into it. This can, however, skew our measurements. */
	for (size_t i = 0; i < sizeof(block); ++i)
		block[i] = i % 256;

	/* Writing of the entire file will be split into several tasks. We need
	 * a special structure for every task. */
	size_t task_count = FILESIZE / (PAGES_PER_TASK * PAGESIZE);
	printf("Scheduling %zu tasks\n", task_count);
	struct aiocb *tasks = calloc(task_count, sizeof(tasks[0]));
	if (tasks == NULL)
		err(EXIT_FAILURE, "Cannot allocate AIO tasks");

	/* The AIO scheduler ‹lio_listio()› requires another structure with
	 * «pointers» to tasks. */
	struct aiocb **task_list = malloc(task_count * sizeof(task_list[0]));
	if (task_list == NULL)
		err(EXIT_FAILURE, "Cannot allocate AIO task list");

	/* Every task describes what needs to be done: */
	for (unsigned task = 0; task < task_count; ++task) {
		/* What needs to be done, ‹LIO_WRITE› or ‹LIO_READ›. This attribute
		 * is only honoured by ‹lio_listio()›, pure AIO calls should use
		 * ‹aio_read()› or ‹aio_write()› per task, where this attribute
		 * is ignored. More details follow below. */
		tasks[task].aio_lio_opcode = LIO_WRITE;

		/* The following three attributes describe the operation that should
		 * be done asynchronously. In effect, they are equivalent to
		 *
		 * 		read(fd, block, count)		for aio_read()
		 * 		write(fd, block, count)		for aio_write() */
		tasks[task].aio_fildes = fd;
		tasks[task].aio_buf = block;
		tasks[task].aio_nbytes = PAGES_PER_TASK * PAGESIZE;

		/* AIO performs operations without changing file offset (as if
		 * ‹pread()› or ‹pwrite()› was used) so that it would not change
		 * offset for the “synchronous” part of the program. This attribute
		 * tells AIO where it should start reading or writing. */
		tasks[task].aio_offset = task * PAGES_PER_TASK * PAGESIZE;

		/* Finally, the last attribute tells how the program should be notified
		 * when the asynchronous operation is finished. Sadly, there is no
		 * direct way to use ‹poll()›-compatible mechanism as of now. We can
		 * arrange a signal to be delivered, a thread started, or nothing.
		 * We could use ‹signalfd()›, of course, but that's just a bit of
		 * a hassle.
		 *
		 * NOTE: Linux AIO (libaio.h) has «undocumented» ‹io_set_eventfd()›
		 * that can set ‹eventfd()› descriptor to be delivered an event
		 * when a scheduled operation completes. */
		tasks[task].aio_sigevent.sigev_notify = SIGEV_NONE;

		task_list[task] = &tasks[task];
	}

	/* Note that the code above does not actually run anything yet. We have
	 * just prepared «tasks» that describe what shall be done. */

	/* The following code is just to measure how long did AIO take to finish. */
	struct timespec start, finish;
	clock_gettime(CLOCK_MONOTONIC, &start);

	/* Now, finally, we tell AIO to run the tasks. */
	for (unsigned task = 0; task < task_count; ++task) {
		/* ‹aio_write()› is like ‹write()›, except it submits the task to be
		 * finished asynchonously. */
		if (aio_write(&tasks[task]) != 0)
			err(EXIT_FAILURE, "aio_write(%u)", task);
		/* The function above returns immediately. At this point, we cannot be
		 * sure whether the block has actually been written to disk
		 * (likely not). */
	}

	/* Once all tasks are submitted, the program shall, ideally, do something
	 * worthwhile. Imagine this code here: */
	//worthwhile_task();

	/* Since we said ‹SIGEV_NONE› for notification event for all asynchronous
	 * tasks, we will not get any direct notification that the tasks are
	 * finished. Due to lack of a better mechanism (and no, signals are not
	 * a better mechanism), we will synchronously poll the tasks ourselves. */
	for (unsigned task = 0; task < task_count; ++task) {
		/* First, ‹aio_error()› tells us ‹EINPROGRESS› if the task is still
		 * running, ‹0› for successful completion or an error code. */
		int err;
		while ((err = aio_error(&tasks[task])) == EINPROGRESS)
			/* nop */;

		if (err != 0)
			errx(EXIT_FAILURE, "Task %u: AIO failed: %s", task, strerror(err));

		/* When the operation finished, ‹aio_return()› returns a value with
		 * the same semantics as ‹read()› or ‹write()› respectively, if used
		 * instead of AIO. */
		if ((size_t) aio_return(&tasks[task]) != tasks[task].aio_nbytes)
			warnx("Task %u: Incomplete data transfer", task);
	}

	/* As you can imagine, the code above is far from ideal. If we find
	 * ourselves in a situation where we need to wait for many asynchronous
	 * tasks in a list, instead of ‹aio_read()› or ‹aio_write()› and subsequent
	 * manual polling, we can use ‹lio_listio()› like this: */
	//if (lio_listio(LIO_WAIT, task_list, task_count, NULL) != 0)
	//	err(EXIT_FAILURE, "lio_listio()");
	/* This would have replace both ‹for› loops above; it would schedule
	 * the tasks and, due to ‹LIO_WAIT›, wait efficiently for them to finish. */

	/* Now come final time measurements. */
	clock_gettime(CLOCK_MONOTONIC, &finish);
	/* Sadly, ‹clock_gettime()› uses a different structure than BSD function
	 * ‹timersub()›, so we need to convert it. */
	struct timeval tv_start, tv_finish, tv_diff;
	TIMESPEC_TO_TIMEVAL(&tv_start, &start);
	TIMESPEC_TO_TIMEVAL(&tv_finish, &finish);
	timersub(&tv_finish, &tv_start, &tv_diff);
	printf("AIO finished in %ld.%06lds\n", tv_diff.tv_sec, tv_diff.tv_usec);

	/* Try running the program with ‹make test›, which will compile both
	 * this and ‹direct.c› variant of the program. It will also run them both
	 * on some file names. */

	/* It may surprise you that (depending on your platform), the AIO solution
	 * may be slower than direct write. Can you think why?
	 *
	 * A hint: Notice the ‹PAGES_PER_TASK› option in ‹config.h›. Try changing
	 * it to 16 or 32 (higher value can cause incomplete transfers) and
	 * recompile the program:
	 *
	 * 		make -B CPPFLAGS+=-DPAGES_PER_TASK=32 test
	 *
	 * Is there any difference now? If so, why? */

	return EXIT_SUCCESS;
}
