#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>

#include <err.h>
#include <fcntl.h>
#include <getopt.h>
#include <linux/cdrom.h>
#include <sys/ioctl.h>
#include <unistd.h>

/* The ‹ioctl()› system call is magical. It's purpose is to issue commands
 * to special files or devices. The exact nature of the command actually
 * depends on what the file we are operating on is.
 *
 * Available commands differ between types of devices, and additional
 * parameters may be passed depending on the command.
 *
 * Usually, there is a manual page describing what operations are available
 * for generic devices:
 *
 * • ‹man 2 ioctl_tty› describes commands for terminal devices,
 * • ‹man 2 ioctl_xfs*› provides several commands for XFS file systems,
 * • ‹man 2 ioctl_userfaultfd› commands user fault fd, which we will see
 *       in seminar 12.
 * • ‹man 7 netdevice› low-level network device commands.
 *
 * In fact, most devices can be interacted with ‹ioctl()›, despite the manual
 * page not being available in ‹man›. Sometimes, it is available online:
 *
 * • https://docs.kernel.org/userspace-api/ioctl/cdrom.html
 *   Commands for CD or DVD drive
 * • https://dri.freedesktop.org/docs/drm/gpu/drm-uapi.html
 *   Commands and wrappers for Direct Rendering Manager (DRM)
 *
 * In this example, we will cover a simple ‹ioctl()› for CD-ROM drive
 * as an example of how this interface works. It is preferable to test
 * this program on a machine that has a CD-ROM or DVD-ROM drive. */


/* By default, we will be operating with ‹/dev/cdrom›. If you steal a computer
 * from a museum, it may even have multiple drives, but some of them should
 * have this name.
 *
 * Depending on the platform, this may be a symlink to ‹/dev/sr0›, ‹/dev/sr1›
 * or similar device. */
static const
char DEFAULT_DEVICE[] = "/dev/cdrom";


void help(char *argv0)
{
	printf(
		"USAGE\n"
		"    %s [OPTIONS] [DEVICE]\n"
		"    default device: '%s'\n"
		"\n"
		"OPTIONS\n"
		"    -t     close tray\n"
		"    -T     toggle tray state\n"
		"    -h     show this help\n",
		argv0, DEFAULT_DEVICE);
}

#define FATAL(FMT, ...) \
	do { \
		warnx("%s: FATAL: " FMT, __func__, ## __VA_ARGS__); \
		abort(); \
	} while (0)


/* The following are actions we will be able to do with the device. */
typedef enum {
	ACTION_CLOSE,
	ACTION_OPEN,
	ACTION_TOGGLE
} action;

/* This function applies given action on device specified in the first argument. */
void cdromctl(const char *device, action act)
{
	/* To operate the device with ‹ioctl()›, we need a descriptor. */
	int fd = open(device, O_RDWR | O_NONBLOCK);

	if (fd == -1)
		err(EXIT_FAILURE, "%s", device);

	/* Now we get the current status, specifically, whether the drive is
	 * closed or ejected. */
	int status = ioctl(fd, CDROM_DRIVE_STATUS, 0);

	if (status == -1)
		err(EXIT_FAILURE, "%s: Failed to get status", device);

	bool is_closed  = !(status & CDS_TRAY_OPEN);

	/* Next, we get drive capabilities. Some drives are not able to perform
	 * certain operations, and we may be able to skip them. */
	int capability = ioctl(fd, CDROM_GET_CAPABILITY, 0);

	if (capability == -1)
		err(EXIT_FAILURE, "%s: Failed to get capabilities", device);

	/* Choose a proper action for ‹ACTION_TOGGLE›. */
	if (act == ACTION_TOGGLE)
		act = is_closed ? ACTION_OPEN : ACTION_CLOSE;

	/* Now we are ready to actually apply the action. */
	switch (act) {
	case ACTION_OPEN:
		/* Some drives cannot open themselves. Those are mostly some ancient
		 * ones that were opened manually from top, like a gramophone:
		 * https://cdn.thewirecutter.com/wp-content/uploads/2020/04/opticaldrive-lowres-5614.jpg */
		if (!(capability & CDC_OPEN_TRAY))
			errx(EXIT_FAILURE, "%s: The drive is not capable of this operation", device);

		/* This option will disable CD-ROM locking. Some drives are able to
		 * disable hardware button by this command. */
		if (ioctl(fd, CDROM_LOCKDOOR, 0) == -1)
			err(EXIT_FAILURE, "%s: Failed to unlock door", device);

		/* Finally, we can command the drive to eject the medium. */
		if (ioctl(fd, CDROMEJECT, 0) == -1)
			err(EXIT_FAILURE, "%s: Failed to eject the medium", device);
		break;

	case ACTION_CLOSE:
		/* As with ‹CDC_OPEN_TRAY›, some drives are not able to close themselves.
		 * This is mostly the case for laptop drives, which are spring loaded
		 * and can open themselves by releasing the spring, but must be manually
		 * closed. */
		if (!(capability & CDC_CLOSE_TRAY))
			errx(EXIT_FAILURE, "%s: The drive is not capable of this operation", device);

		/* Attempt to close the tray. */
		if (ioctl(fd, CDROMCLOSETRAY, 0) == -1)
			err(EXIT_FAILURE, "%s: Failed to close the tray", device);
		break;

	default:
		FATAL("%d: Invalid action value", act);
	}

	close(fd);
}


int main(int argc, char *argv[])
{
	/* By default, we will open the tray, as ‹eject(1)› would do. */
	action act = ACTION_OPEN;

	int option;
	while ((option = getopt(argc, argv, "tTh")) != -1) {
		switch (option) {
		case 't':
			act = ACTION_CLOSE;
			break;

		case 'T':
			act = ACTION_TOGGLE;
			break;

		case 'h':
			help(argv[0]);
			return EXIT_SUCCESS;

		default:
			fprintf(stderr, "Try '%s -h' for help\n", argv[0]);
			return EXIT_FAILURE;
		}
	}

	if (argc - optind > 1) {
		fprintf(stderr, "Too many devices, try '%s -h' for help\n", argv[0]);
		return EXIT_FAILURE;
	}

	const char *device = (argc - optind == 1) ? argv[optind] : DEFAULT_DEVICE;
	cdromctl(device, act);
	return EXIT_SUCCESS;
}
