#include "sha_utils.h"

#include <stdlib.h>

#include <err.h>
#include <unistd.h>
#include <openssl/err.h>
#include <openssl/sha.h>

/* This is the implementation of SHA utilities header file.
 *
 * The implementation contains various hacks and magic tricks. This is purely
 * to shield you from some of OpenSSL's gory details. */

/* The size of the read buffer. */
#define READ_BUFFER_SIZE 512

#if defined(__GNUC__) || defined(__clang__)
#	define CONSTRUCTOR __attribute__((constructor))
#	define DESTRUCTOR __attribute__((destructor))
#else
#	define CONSTRUCTOR
#	define DESTRUCTOR
#endif

static CONSTRUCTOR
void initialize(void)
{
	/* This function must be called before OpenSSL's ERR_*() functions
	 * can be used. */
	ERR_load_crypto_strings();
}

static DESTRUCTOR
void finalize(void)
{
	/* Release resources allocated by ‹ERR_load_crypto_strings()›. */
	ERR_free_strings();
}

/* The following dark magic is used to generate function symbols depending
 * on chosen algorithm. It uses double expansion for symbol concatenation,
 * as simple concatenation would not work:
 *
 * 		#define SYMBOL foo
 * 		#define CONCAT(A, B) A ## B
 *
 * If we try to use ‹CONCAT› like this, it will break on macro parameters:
 *
 * 		CONCAT(foo, bar)	// OK, foobar
 * 		CONCAT(SYMBOL, bar) // Wrong, SYMBOLbar
 *
 * Double indirection fixes this:
 *
 * 		#define CONCAT1(A, B) A ## B
 * 		#define CONCAT(A, B) CONCAT1(A, B)
 *
 * 		CONCAT(SYMBOL, bar) // → CONCAT1(foo, bar)
 * 		                    // → OK, foobar
 * */
#define CONCAT1(A, B) A ## B
#define CONCAT(A, B) CONCAT1(A, B)

/* ‹DSYMBOL()› and ‹DCALL()› macros will return a symbol for the chosen
 * hash algorithm in variable ‹DIGEST›. */
#define DSYMBOL(SYMBOL) CONCAT(CONCAT(DIGEST, _), SYMBOL)
#define DCALL(WHAT, ...) (DSYMBOL(WHAT))(__VA_ARGS__)

/* Convenience macros for OpenSSL's reason strings. */
#define ERR_reason_string() ERR_reason_error_string(ERR_get_error())

bool get_digest(int fd, size_t *size, void **digest)
{
	static char buffer[READ_BUFFER_SIZE];

	/* Initialize hash context. */
	DSYMBOL(CTX) context;
	if (!DCALL(Init, &context))
		goto error_openssl;

	/* Update context with the block from the file. */
	ssize_t rd;
	off_t offset = 0;
	while ((rd = pread(fd, buffer, sizeof(buffer), offset)) > 0) {
		/* ‹pread()› may return fewer bytes than we asked it to read. */
		if (!DCALL(Update, &context, buffer, rd))
			goto error_openssl;

		offset += rd;
	}

	if (rd == -1)
		goto error_errno;

	/* Allocate buffer for the result. */
	*size = DSYMBOL(DIGEST_LENGTH);
	if ((*digest = malloc(*size)) == NULL)
		goto error_errno;

	/* ‹*_Final()› writes digest bytes to the buffer. */
	if (!DCALL(Final, *digest, &context))
		goto error_openssl;

	return 1;

error_errno:
	warn("Failed to compute checksum");
	return 0;

error_openssl:
	warnx("OpenSSL error: %s", ERR_reason_string());
	return 0;
}
