#ifndef SHA_UTILS_H
#define SHA_UTILS_H

#include <stddef.h>
#include <stdbool.h>

/**
 * Digest to use.
 */
#ifndef DIGEST
#  define DIGEST SHA256
#endif

/**
 * Computes the SHA checksum of the data from the given descriptor.
 *
 * The function uses OpenSSL's implementation of the digest to compute
 * the digest of the provided file descriptor. The result is returned
 * via the output parameters \p size and \p digest, where the digest
 * is automatically allocated.
 *
 * \note    It is responsibility of the caller to free the memory allocated
 *          for the \p digest variable after usage.
 *
 * Example:
 *
 * \code{.c}
 *      size_t digest_size;
 *      void *digest;
 *
 *      if (get_digest(descriptor, &digest_size, &digest) == 0) {
 *          // Handle error
 *      }
 * \endcode
 *
 * The provided file descriptor must be seekable. File offset is preserved
 * by this call.
 *
 * \param[in]   fd      File descriptor to be read.
 * \param[out]  size    Number of bytes written to \p digest.
 * \param[out]  digest  Automatically allocated memory that holds the result.
 * \returns     \c true on success, otherwise prints an error message
 *              and returns \c false.
 */
bool get_digest(int fd, size_t *size, void **digest);

#endif // SHA_UTILS_H
