#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <err.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>

/* NOTE: Read ‹server.c› first! */


/* This is a program similar to ‹writer.c› from ‹ex02t-sun›, but is able
 * to receive a response from the server without using connected sockets.
 *
 * This is achieved by using Linux-specific abstract sockets. Technically,
 * every client atomically binds an abstract socket, thus also becoming
 * a “server” for its own abstract socket.
 *
 * While this enables programs to use full duplex UNIX communication without
 * actual connections, it can only be used for very simple protocols
 * (i.e. request → response in a loop). For more complex features, connected
 * sockets must be used. */

/* Auxiliary function to remove newline character from the string. */
static
void trim(char *str)
{
	char *nl = strchr(str, '\n');
	if (nl != NULL)
		*nl = '\0';
}

int main(void)
{
	int sock = socket(AF_UNIX, SOCK_DGRAM, 0);
	if (sock == -1)
		err(EXIT_FAILURE, "socket");

	/* To create an abstract socket where the server can send a reply, we will
	 * bind it to an abstract address. */
	struct sockaddr_un client = {
		.sun_family = AF_UNIX,
	};

	/* Note that we call ‹bind()› for ‹&client›, but instead of specifying
	 * ‹sizeof(client)›, we only use the length of the family, without the
	 * ‹sun_path› attribute.
	 *
	 * This is a Linux-specific feature called “autobind”, where kernel binds
	 * this socket to a new abstract address. From now on, when we ‹sendto()›
	 * some data on this socket, the ‹recvfrom()› in the target process will
	 * be able to extract our address. */
	if (bind(sock, (struct sockaddr *) &client, sizeof(client.sun_family)) == -1)
		err(EXIT_FAILURE, "bind");

	/* This is a server socket address. */
	struct sockaddr_un server = {
		.sun_family = AF_UNIX,
		.sun_path = "socket",
	};

	char buffer[64];
	while (fgets(buffer, sizeof(buffer), stdin) != NULL) {
		trim(buffer);

		/* Note that ‹sock› here is bound to an abstract address. Thus,
		 * ‹sendto(sock, … server, length)› will cause server's call to
		 * ‹recvfrom(…, &client, &length)› fill the client structure with
		 * abstract address created by ‹bind()› above. */
		if (sendto(sock, buffer, sizeof(buffer), 0,
					(struct sockaddr *) &server, sizeof(server)) == -1)
			err(EXIT_FAILURE, "sendto");

		/* Wait for the reply from the server and print it. */
		if (recvfrom(sock, buffer, sizeof(buffer), 0, NULL, NULL) == -1)
			err(EXIT_FAILURE, "recvfrom");

		printf("%s\n", buffer);
	}

	return EXIT_SUCCESS;
}
