#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <err.h>
#include <sys/socket.h>
#include <sys/un.h>

/* In the following example we will show how abstract sockets with autobind
 * feature can be used to implement duplex connection-less communication
 * between a server and client.
 *
 * The previous example used file system path socket for the server. When
 * a client connects to this socket, it can technically send data one way
 * using ‹sendto()›, which can be retrieved by the server using ‹recvfrom()›,
 * however, there is no address available to send the reply to.
 *
 * This can be fixed using connected sockets, but for now, we will look at
 * a connection-less alternative: Linux-specific abstract sockets.
 *
 * This program is basically the same as the ‹reader.c› from ‹ex02›, except
 * it now sends a reply back. */

int main(void)
{
	int sock = socket(AF_UNIX, SOCK_DGRAM, 0);
	if (sock == -1)
		err(EXIT_FAILURE, "socket");

	struct sockaddr_un addr = {
		.sun_family = AF_UNIX,
		.sun_path = "socket",
	};

	if (bind(sock, (struct sockaddr *) &addr, sizeof(addr)) == -1)
		err(EXIT_FAILURE, "bind");

	/* This structure will be used to retrieve client address, and send them
	 * a reply back. */
	struct sockaddr_storage client;
	socklen_t length;

	char buffer[64];

	ssize_t rd;

	/* ‹recvfrom()› modifies the ‹&length› argument. Therefore we should always
	 * restore it to the available length of ‹client› structure. */
	while ((length = sizeof(client)),
			(rd = recvfrom(sock, buffer, sizeof(buffer), 0,
						   (struct sockaddr *) &client, &length)) >= 0) {
		/* DANGER: To make the code simpler, we will alias the client structure
		 * to ‹sockaddr_un›. However, we MUST NOT use this pointer until we
		 * are sure it is ‹AF_UNIX› address! */
		struct sockaddr_un *client_un = (struct sockaddr_un *) &client;

		/* To print the client address, we will first inspect whether it is
		 * an abstract path or something else. */
		if (length <= sizeof(client.ss_family) || client.ss_family != AF_UNIX) {
			/* Unnamed, unbound or non-UNIX socket. */
			printf("[?]");
		/* After this point, we may use ‹client_un›. */
		} else if (client_un->sun_path[0] != 0) {
			/* Pathname socket. The path should be a null-terminated string. */
			printf("%s", client_un->sun_path);
		} else {
			/* If the first byte of ‹sun_path[0]› is 0 and length suggests
			 * that there are some bytes, it is actually an abstract socket
			 * whose name follows the zero byte. The length of the socket name
			 * should be derived from ‹length›. */
			size_t name_length = length - sizeof(sa_family_t);

			/* We could print the path as
			 *
			 * 		printf("%.*s", (int) name_length, &client_un->sun_path[1]);
			 *
			 * but the path may also contain non-printable characters. In fact,
			 * abstract sockets are identified by bytes following the initial
			 * 0, and these bytes have no special meaning. Thus, we will print
			 * them as follows just for demonstration: */
			for (size_t byte = 1; byte < name_length; ++byte) {
				char c = client_un->sun_path[byte];
				if (isprint(c))
					putchar(c);
				else
					printf("<%02x>", c);
			}
		}

		/* Next, we print the message. */
		printf(": %s\n", buffer);

		/* Finally, we attempt to send a reply back to the client. To make
		 * things interesting, we will send the same message, except all
		 * characters will be replaced with ‹*›. */
		memset(buffer, '*', strlen(buffer));
		if (sendto(sock, buffer, sizeof(buffer), 0,
					(struct sockaddr *) &client, length) == -1)
			warn("Cannot send response");
	}

	return EXIT_SUCCESS;
}
