#include <limits.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <err.h>
#include <fcntl.h>
#include <grp.h>
#include <pwd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

/* Under Linux, UNIX sockets offer several advanced features. Aside from sending
 * data, they can also be used to transfer
 *
 * • Open file descriptions (this demo),
 * • Process credentials,
 * • Security Enhanced Linux (SELinux) labels.
 *
 * This is usually modelled as a separate channel that does not interfere with
 * data transmission. Messages in this channel are called “ancillary messages“,
 * or sometimes “control messages”. */

/* This demo demonstrates how open file descriptions (OFD) can be passed to other
 * process via UNIX socket. This can be useful, for example, in server-workers
 * model, where a privileged process (server) obtains a descriptor to a resource
 * that a restricted process (worker) should not have access to.
 *
 * Another example is Seccomp Supervisor model, where a process listens to
 * ‹seccomp› events via a descriptor passed from the monitored process.
 * See ‹man 2 seccomp› → ‹SECCOMP_FILTER_FLAG_NEW_LISTENER› or ‹man
 * 2 seccomp_unotify›. */


/* Overview
 * -------- */

/* The process is expected to start with elevated privileges. It will create
 * a socket pair that will be used to pass OFDs to the worker process. It will
 * then fork to two processes:
 *
 * 1. Supervisor remains privileged; it reads a file name from standard input,
 *    opens the file and passes the OFD to the worker process.
 *
 * 2. Worker process drops its privileges by setting all its UIDs and GIDs to
 *    user ‹nobody›. When it receives an OFD from supervisor, it checks whether
 *    it can open the file, and whether it can read from the descriptor.
 *
 * When run as
 *
 * 		$ sudo ./sendfd
 *
 * then after writing ‹/etc/shadow› to the program input, the supervisor should
 * be able to open the file and pass it to the worker. Since worker is not an
 * unprivileged process, it will not be able to ‹open()› the file, however, it
 * will be able to read from the obtained file descriptor. This will be
 * demonstrated by reading and displaying first eight bytes of the file. */

/* While this operation is commonly referred to as “file descriptor passing“, it
 * is important to note that file descriptor is just a number, an index in
 * process' OFD table. Thus, more accurately, it is a reference to an entry in
 * this table that is being passed. The result of this operation is as if the
 * process called ‹dup()›, but the new descriptor was opened in the target
 * process instead of the caller.
 *
 * This is also demonstrated by the program, as supervisor and the worker should
 * see the open file as descriptors with different numbers:  Before supervisor
 * attempts to open the file, the processes should have these open descriptors:
 *
 *      Supervisor: 0 (stdin)   1 (stdout)  2 (stderr)  3 (socket)
 *      Worker:     0 (stdin)   1 (stdout)  2 (stderr)              4 (socket)
 *
 * When supervisor opens the file, it will obtain descriptor ‹4›, because this
 * is the first free descriptor. After it passes the description to the worker,
 * kernel links the OFD to worker's table and opens it as a new descriptor.
 * Since ‹3› is the index of the first free entry in worker's OFD table, it is
 * what the worker obtains. Note that ‹4› would not be assigned even if it were
 * free, kernel does not really try to match the descriptor number between
 * processes.
 *
 * This also means that, since OFDs are linked, reading the descriptor in worker
 * would affect the file offset in both processes.  You are encouraged to try
 * this out. */

/* Finally, this feature is only available for UNIX sockets. While network
 * sockets support some ancillary messages, OFDs cannot be passed, as the OFD to
 * link is generally not available on the target host. */


/* Forward declarations
 * -------------------- */

/* As usual, we will first cover ‹main()› to explain the solution sequentially. */
void supervisor(int pid_worker, int fd_worker);
void worker(int fd_super);


/* Main
 * ---- */

int main(void)
{
	/* If we are not a privileged process, display a warning. */
	if (geteuid() != 0)
		warnx("Running as unprivileged user");

	/* Create a socket pair for communication. */
	int sockpair[2];
	if (socketpair(AF_UNIX, SOCK_SEQPACKET, 0, sockpair) == -1)
		err(EXIT_FAILURE, "Cannot create socket pair");

	/* Spawn a worker process. */
	int worker_pid = fork();
	if (worker_pid == -1)
		err(EXIT_FAILURE, "Cannot create child process");

	if (worker_pid == 0) {
		close(sockpair[0]);
		worker(sockpair[1]);
		printf("[W] Exiting");
		return EXIT_SUCCESS;
	}

	/* In parent process, run as supervisor. */
	close(sockpair[1]);
	supervisor(worker_pid, sockpair[0]);
	kill(worker_pid, SIGTERM);
	return EXIT_SUCCESS;
}


/* Supervisor
 * ---------- */

/* An auxiliary function to remove newline from ‹getline()›'s result. */
static
void chomp(char *line)
{
	char *nl = strchr(line, '\n');
	if (nl != NULL)
		*nl = '\0';
}

/* A function to display prompt. */
static
void prompt(void)
{
	printf("(S) Waiting for file to open\n");
}

void supervisor(int pid_worker, int fd_worker)
{
	char *file_name = NULL;
	size_t size = 0;

	/* Display a prompt and read file name from input. */
	while (prompt(), getline(&file_name, &size, stdin) != -1) {
		/* Check whether the worker process did not die while we were waiting.
		 * This could be better solved by setting up a ‹SIGCHLD› handler, but
		 * to keep this demo as short as possible, we will just check for that
		 * here. */
		if (waitpid(pid_worker, NULL, WNOHANG) != 0)
			errx(EXIT_FAILURE, "(S) Worker proccess terminated");

		chomp(file_name);

		if (strlen(file_name) == 0)
			continue;

		/* Open the file and obtain the descriptor to be passed. */
		int fd_file = open(file_name, O_RDONLY);
		if (fd_file == -1) {
			warn("(S) %s", file_name);
			continue;
		}

		printf("(S) File opened as %d\n", fd_file);

		/* This is where fun begins. Ancillary messages are passed via
		 * ‹sendmsg()› system call, which requires a bit of setup.
		 *
		 * First, we will need a ‹cmsghdr› (control message header) structure
		 * that contains descriptors representing the OFDs to be send to the
		 * other process.  This structure essentially contains a Flexible Array
		 * Member, so we cannot allocate it directly. We will use a simple trick
		 * showed in ‹man 3 cmsg›, where we align the structure with a slightly
		 * larger buffer using a union.
		 *
		 * Note that in general, this buffer can contain more than one message,
		 * which must be carefully set up due to the Flexible Array Member.
		 * This is similar to what inotify's descriptor yields when read, as
		 * ‹struct inotify_event› is also a structure with a FAM. */
		union {
			/* ‹CMSG_SPACE(S)› is a macro that will return the number of bytes
			 * required for an ancillary message with ‹S› bytes for data. */
			char buffer[CMSG_SPACE(sizeof(fd_file))];
			/* We will align this buffer with ‹cmsghdr› structure. */
			struct cmsghdr header;
		} msg_control = {
			.header = {
				/* ‹cmsg_len› contains the size of the payload. There is a macro
				 * ‹CMSG_LEN(S)› that tells us this value for data of size ‹S›.
				 * Using ‹sizeof(fd_file)› directly here may result in undefined
				 * behaviour, as ‹CMSG_LEN()› may return a different value due
				 * to alignment. */
				.cmsg_len = CMSG_LEN(sizeof(fd_file)),
				/* ‹cmsg_level› tells kernel which socket stack layer (or
				 * protocol) is this message originating from. Compare with
				 * ‹setsockopt›'s ‹level› parameter.
				 *
				 * For historical reasons, UNIX messages operate on
				 * ‹SOL_SOCKET›, despite being of ‹AF_UNIX› family. */
				.cmsg_level = SOL_SOCKET,
				/* ‹cmsg_type› sets the type of the message. ‹SCM_RIGHTS›
				 * tells kernel that this message passes rights to a resource
				 * represented by a file descriptor. */
				.cmsg_type = SCM_RIGHTS,
			},
		};

		/* The flexible entry after the message header contains data specific
		 * to the ‹cmsg_type›. In case of ‹SCM_RIGHTS›, it contains an array
		 * of descriptors. A pointer to the first byte of the buffer can be
		 * obtained using ‹CMSG_DATA()› macro, and thus we can copy the
		 * descriptor using ‹memcpy()› or type cast. */
		memcpy(CMSG_DATA(&msg_control.header), &fd_file, sizeof(fd_file));

		/* Next, we want to pass some real data. While ‹write()›, ‹send()› and
		 * ‹sendto()› are called with a single buffer, ‹sendmsg()› is a bit more
		 * generic and allows to pass a vector of buffers. We are only going
		 * to pass a single one. Note that there is a similar function for
		 * regular file descriptors, ‹writev()›. */
		struct iovec iovec[1] = {
			/* ‹iov_base› and ‹iov_len› specify a buffer to pass. This is
			 * essentially what you would pass to ‹write()›. */
			{ .iov_base = file_name, .iov_len = strlen(file_name) + 1 },
		};

		/* Finally, we can put this all together. As the ancillary messages
		 * may get quite complicated, the ‹sendmsg()› function instead expects
		 * a single structure that describe both ‹iovec› and ancillary messages. */
		struct msghdr transfer = {
			/* ‹msg_name› is a pointer to ‹sa_sockaddr*› structure, while
			 * ‹msg_namelen› contains size of this structure. In unconnected
			 * sockets, this describes the recipient socket. Socket pairs are
			 * automatically connected, thus we can leave these fields empty. */
		//	.msg_name
		//	.msg_namelen
			/* ‹msg_iov› describes a vector of buffers to send, and ‹msg_iovlen›
			 * the number of elements in the ‹msg_iov› array. */
			.msg_iov = iovec,
			.msg_iovlen = sizeof(iovec) / sizeof(*iovec),
			/* Finally, ‹msg_control› points to the buffer containing ancillary
			 * messages. ‹msg_controllen›, confusingly, does not contain the
			 * «number» of messages, but the size of the entire buffer. This is
			 * mainly because messages can have different sizes. */
			.msg_control = msg_control.buffer,
			.msg_controllen = sizeof(msg_control.buffer),
		};

		/* At last, we will send the message to the worker process. */
		if (sendmsg(fd_worker, &transfer, 0) <= 0)
			warn("(S) Cannot pass descriptor to worker");

		/* TIP 1: Try to sleep here for 1 second, read a byte and display it.
		 * The explanation is at the end of the source code. */

		/* TIP 2: Comment out the  following two lines and recompile the program.
		 * Again, the explanation is at the end of the source code. */
		close(fd_file);
		printf("(S) Closed\n");
	}
}


/* Worker
 * ------ */

/* To demonstrate that rights for a resource (OFD) were passed, the worker
 * process will drop its elevated privileges before attempting to accept
 * descriptors, and will instead assume the identity of the following user: */
#define DROP_USER "nobody"

static
void drop_privileges(void)
{
	struct passwd *nobody = getpwnam(DROP_USER);

	if (nobody == NULL) {
		warnx("[W] No account for %s", DROP_USER);
	} else {
		if (setgid(nobody->pw_gid) == -1)
			warn("[W] Cannot drop GID to %d", nobody->pw_gid);
		if (setgroups(1, &nobody->pw_gid) == -1)
			warn("[W] Cannot drop supplementary groups");
		if (setuid(nobody->pw_uid) == -1)
			warn("[W] Cannot drop UID to %d", nobody->pw_uid);
	}
}

void worker(int fd_super)
{
	drop_privileges();
	printf("[W] Process %d ready\n", getpid());

	while (true) {
		/* On the receiver's part of the OFD transfer, we also need to create
		 * a space for both iovec structure and ancillary messages. */

		/* The ancillary message buffer can be left unstructured. A defensive
		 * program should not assume the received structure, nor type of the
		 * message. */
		char cmsgbuf[1024];

		/* The ‹iovec› array contains the buffers to be filled. Note that the
		 * number and size of these buffers does not need to match the sender's
		 * structure. Kernel will simply fill these arrays in order in which
		 * they are defined, fully filling one before moving to the next one.
		 *
		 * Since we expect a file name to be delivered, a single buffer with
		 * PATH_MAX size should be enough. If a longer message arrives, it will
		 * simply be trimmed. */
		char iovbuf[PATH_MAX];
		struct iovec iovec[1] = {
			{ .iov_base = iovbuf, .iov_len = sizeof(iovbuf) },
		};

		/* Again, we put together the ‹msghdr› structure to be used with
		 * ‹readmsg()›. */
		struct msghdr transfer = {
		//	.msg_name = NULL,
		//	.msg_namelen
			.msg_iov = iovec,
			.msg_iovlen = sizeof(iovec) / sizeof(*iovec),
			.msg_control = cmsgbuf,
			.msg_controllen = sizeof(cmsgbuf),
			/* This is to declare that we are prepared to receive ancillary
			 * messages passing OFDs. */
			.msg_flags = SCM_RIGHTS,
		};

		/* Wait for the message. */
		if (recvmsg(fd_super, &transfer, 0) <= 0) {
			warn("[W] Cannot read from socket");
			return;
		}

		/* The ‹recvmsg()› fills in the ‹msg_iov› field. If we received some
		 * data, we will expect it to be a file name. */
		if (transfer.msg_iovlen >= 1) {
			struct iovec *file = &transfer.msg_iov[0];
			printf("[W] Got message containing %.*s\n",
					(int) file->iov_len, (char *) file->iov_base);

			/* The worker is running as an unprivileged process, so it should
			 * not be able to open the file by its name. If you do not run the
			 * program with elevated privileges, this call might succeed. */
			int fd = open((char *) file->iov_base, O_RDONLY);
			if (fd == -1) {
				printf("[W] open(): Failed\n");
			} else {
				printf("[W] open(): Success\n");
				close(fd);
			}
		}

		/* Ancillary messages are filled into our ‹cmsgbuf› array. However,
		 * since these messages might have various lengths, we will use provided
		 * macros to access them safely.
		 *
		 * ‹CMSG_FIRSTHDR()› returns a pointer to the first ancillary message,
		 * or ‹NULL› if no ancillary messages were received. */
		for (struct cmsghdr *cmsg = CMSG_FIRSTHDR(&transfer);
				cmsg != NULL; cmsg = CMSG_NXTHDR(&transfer, cmsg)) {
			if (cmsg->cmsg_type != SCM_RIGHTS) {
				/* We expect this message to be of type ‹SCM_RIGHTS›. */
				warnx("[W] Unknown ancillary message");
				continue;
			}

			/* We got the correct type of the message. Received descriptors
			 * are in the data section, we just need to extract them.
			 *
			 * Be aware that the resource has already been transferred to this
			 * process by the time ‹recvmsg()› has finished. In this system
			 * call, kernel has linked OFDs from source process and opened file
			 * descriptors for new resources. Numbers of these descriptors were
			 * stored in the ancillary data we are currently inspecting and may
			 * not be the same as what the supervisor has sent, as explained in
			 * the Overview.
			 *
			 * We expect only one descriptor. */

			int fd_file;
			if (cmsg->cmsg_len < sizeof(fd_file)) {
				warnx("[W] Invalid length of ancillary message payload");
				continue;
			}

			memcpy(&fd_file, CMSG_DATA(cmsg), sizeof(fd_file));

			/* Finally, we can attempt to read from the descriptor. */
			char buffer[8];
			ssize_t rd;

			if ((rd = read(fd_file, buffer, sizeof(buffer))) < 0) {
				warnx("[W] read(%d): Failed", fd_file);
			} else {
				printf("[W] read(%d): Success: %.*s\n", fd_file, (int) rd, buffer);
			}

			close(fd_file);
		}
	}
}


/* This example is fairly complicated, but it is also not an easy concept.
 *
 * To play with this example further, try the following:
 *
 * 1. After passing the OFD to worker, make the supervisor wait for one second
 *    and read one byte from the descriptor before closing it.
 *
 *    You should see that indeed the ‹read()› continued where worker process
 *    left off, because the processes share the same offset in the OFD table.
 *
 * 2. Comment out ‹close()› around line 273 and try to open multiple files.
 *
 *    You should see that while supervisor opens various descriptors, the worker
 *    should always receive descriptor ‹3›. */

/* For completion's sake, it should be mentioned that it is also possible to
 * pass descriptors using ‹pidfd_getfd()› system call. While the interface is
 * a bit simpler and does not require a socket, it is only available for Linux
 * and requires a cooperation with ‹ptrace()› interface. */
