#include <stdio.h>
#include <stdlib.h>

#include <err.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>

/* The first example shows how a file can be mapped into process memory.
 * Linux memory subsystem uses very similar (if not the same) mechanism
 * to map binaries and libraries into the memory, and also uses it to
 * create “anonymous” mappings (essentially memory allocation). */

/* But let's just start with mapping a single file into the memory and
 * reading it directly. */

int main(int argc, char *argv[])
{
	if (argc != 2) {
		fprintf(stderr, "usage: %s FILE\n", argv[0]);
		return EXIT_FAILURE;
	}

	/* Before we can map a file, we need it's file descriptor. Note that
	 * ‹open()›'s ‹mode› must be compatible with intended mapping, thus,
	 * the following modes are compatible:
	 *
	 * • ‹O_RDONLY› for ‹PROT_READ›
	 * • ‹O_RDWR› for ‹PROT_READ› (we will just not let direct memory writes)
	 * • ‹O_RDWR› for ‹PROT_READ | PROT_WRITE›
	 *
	 * The following example is not:
	 *
	 * • ‹O_RDONLY› for ‹PROT_READ | PROT_WRITE› */
	int fd = open(argv[1], O_RDONLY);
	if (fd == -1)
		err(EXIT_FAILURE, "Cannot open %s", argv[1]);

	/* Another piece of the information we need is the length of the file
	 * memory region to map. We will use ‹stat()› to figure this out. */
	struct stat fdstat;
	if (fstat(fd, &fdstat) == -1)
		err(EXIT_FAILURE, "Cannot stat %s", argv[1]);

	/* We will not bother ourselves with mapping empty files. */
	if (fdstat.st_size == 0)
		return EXIT_SUCCESS;

	/* Cue the start of the show.
	 *
	 * The ‹mmap()› function will try to establish a file mapping according
	 * to the provided parameters: */
	void *mem = mmap(NULL, fdstat.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
	/*  1.                 ──────────────                          ──  ─
	 *     We want the file ‹fd› to be accessible from the offset ‹0›
	 *     (the beginning) and next ‹fdstat.st_size› bytes. This means,
	 *     the whole file.
	 *
	 *               NULL
	 *  2.           ────
	 *     We do not care where kernel places the mapping. Not that we have
	 *     many choices anyway; a portable program should not use this argument.
	 *     Should we set here some other address, kernel will use that as a hint
	 *     where to put it. ‹MAP_FIXED› flag can be used to try to map the
	 *     selected address directly.
	 *
	 *                                     PROT_READ  MAP_PRIVATE
	 *  3.                                 ─────────  ───────────
	 *     Here we say that we only intend to read from the memory.
	 *     The second flag tells kernel that the mapping is private, so that
	 *     our changes to the file (which we cannot do anyway) should not be
	 *     propagated to the file. */

	/* When ‹mmap()› fails, it returns ‹(void *) -1› instead of ‹NULL›. This
	 * is because ‹0› can be sometimes return as a valid mapping.
	 *
	 * You read that right. */
	if (mem == MAP_FAILED)
		err(EXIT_FAILURE, "Cannot map %s", argv[1]);

	/* At this point, ‹mem› points to the beginning of a memory segment where
	 * kernel will copy pages from the file should we access them. It will
	 * also update the pages if we change the underlying file. (This behaviour
	 * is not specified for ‹MAP_PRIVATE› in POSIX, but Linux does so.) */

	/* It is worth pointing out that while we can say basically an arbitrary
	 * number in ‹length› parameter, kernel will always map entire pages.
	 * This means that if the file is not long enough, the last page will
	 * not be fully populated by the file. This part will behave as follows:
	 *
	 * • Reads in the region will be filled with zeroes.
	 * • Writes into this region will not be propagated to the file.
	 * • It is unspecified what happens to added or removed mapped regions of
	 *   the file if ‹ftruncated()› is called on it.
	 *
	 * One of very few portable ways to change underlying file size is to call
	 * ‹ftruncate()› followed by ‹mremap()›. */

	/* Now, we can simply iterate over bytes of the file-backed memory. */
	for (void *ptr = mem; ptr - mem < fdstat.st_size; ptr++) {
		/* This call to ‹sleep()› will be relevant later. To avoid buffering,
		 * we will use ‹write()› instead of ‹putchar()›. */
		//sleep(1);
		write(STDOUT_FILENO, ptr, 1);
	}

	/* Finally, we undo the memory mapping. Interestingly, ‹close()› would close
	 * the file descriptor, but «will not» drop the created mapping. We really
	 * need to call ‹munmap()› if we want to return memory back to kernel.
	 *
	 * Of course, since we are exiting the program anyway, we do not really
	 * need to do any of this - on ‹exit()›, the kernel removes file mappings
	 * and closes file descriptors on its own. */
	if (munmap(mem, fdstat.st_size) == -1)
		warn("Cannot unmap %s", argv[1]);
	if (close(fd) == -1)
		warn("Cannot close %s", argv[1]);

	/* When you compile and run the program on some text file, you should see
	 * the program writing the contents as if the file was read directly.
	 *
	 * While memory mapping may seem a bit more efficient than ‹read()›,
	 * it should not be automatically preferred over classic I/O, but should
	 * be considered for its cons as well:
	 *
	 * • Memory mapping is a bit more complex to set up,
	 * • Mapping an entire file might require a lot of memory, imagine
	 *   a video player mapping an entire 4K 3D The Return of the King
	 *   extended edition into the memory. Classic I/O reads files sequentially
	 *   and without such memory overhead.
	 * • Changing the underlying memory will change contents seen by both
	 *   mechanisms. In memory mapping, however, this behaviour is not
	 *   portable. */

	/* That being said, let's try to break the program.
	 *
	 *  1. Uncomment ‹sleep()› above and recompile the program.
	 *
	 *  2. Open a second terminal in this directory.
	 *  3. Create a file with some short contents:
	 *
	 *     $ echo 'Hello, World!' >file
	 *
	 *  4. In the same terminal «prepare» (don't execute yet!) a command that
	 *     would write a shorter message:
	 *
	 *     $ echo 'Hello!' >file   # DO NOT EXECUTE IMMEDIATELY!
	 *
	 *  5. In the original terminal, run the program and wait until
	 *     ‹Hello› appears on the screen.
	 *
	 *  6. Then, execute the prepared command in the second terminal.
	 *
	 *  What do you see? How can this behaviour be avoided? */

	return EXIT_SUCCESS;
}
