#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <err.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>

/* This example will show a rather odd feature of ‹mmap()› -- its ability
 * to map ‹NULL› pointer.
 *
 * The gist of the program is as follows:
 *
 * • Create a child and attempt to access ‹*NULL›. This should obviously
 *   fail.
 *
 * • Use ‹mmap()› to remap ‹NULL› pointer and retry the dereference. */

/* NOTE: This program requires to be run with administrator prvilieges.
 * User programs should not be able to do this, as dereferencing a ‹*NULL›
 * is a fatal error in 99,999 % cases and should be regarded as such. */


/* If the platform does not provide ‹PAGESIZE›, we will get it from
 * a less standard header. */
#ifndef PAGESIZE
#	include <sys/user.h>
#	define PAGESIZE PAGE_SIZE
#endif // PAGESIZE


/* This function will attempt to dereference ‹*NULL› in a child process. */
static
void test_null_deref_child(void)
{
	/* Compiler may try to warn us about using ‹NULL› directly, but we know
 	 * what we are doing. This dummy variable will silence these warnings. */
	char *zero = NULL;
	printf("Attempting to use NULL pointer\n");
	strcpy(zero, "Hello, NULL pointer!");
	printf("%s\n", zero);
	exit(EXIT_SUCCESS);
}

/* The wrapper function that will ‹fork()› and run the function above. */
static
void test_null_deref(void)
{
	pid_t pid = fork();
	if (pid == -1)
		err(EXIT_FAILURE, "Cannot fork");

	if (pid == 0)
		test_null_deref_child();

	int wstatus;
	if (waitpid(pid, &wstatus, 0) == -1) {
		warn("Cannot wait for child %d", pid);
		return;
	}

	if (WIFEXITED(wstatus)) {
		if (WEXITSTATUS(wstatus) == 0)
			printf("🥳 NULL dereferenced successfully\n");
		else
			printf("☹️ Child exited with exit status %d\n", WEXITSTATUS(wstatus));
	} else if (WIFSIGNALED(wstatus)) {
		printf("😭 Child died after receiving signal %d\n", WTERMSIG(wstatus));
	} else {
		printf("🕵️ Unknown exit status?\n");
	}
}

/* Here we will attempt to create mapping for address zero. */
static
void commit_atrocities(void)
{
	/* A special device ‹/dev/zero› can be used to create a mapping for
	 * zero-filled pages of desired size.
	 *
	 * This idea originated in SunOS a long time ago. Nowadays this can be
	 * avoided by creating anonymous mappings (simply by adding ‹MAP_ANONYMOUS›
	 * to ‹mmap()›'s flags). This is just for a demonstration on how special
	 * devices apart from ‹/dev/null› can be useful. */
	int zero = open("/dev/zero", O_RDWR);
	if (zero == -1)
		err(EXIT_FAILURE, "Cannot open /dev/zero");

	/* Note the ‹MAP_FIXED› flag. This will tell the kernel that we want to
	 * interpret the first parameter (‹NULL› in our case) not as the hint,
	 * but an absolute address. Kernel may not honour our request unless
	 * we run as a privileged process. */
	if (mmap(NULL, PAGESIZE, PROT_READ | PROT_WRITE,
				MAP_PRIVATE | MAP_FIXED, zero, 0) == MAP_FAILED) {
		err(EXIT_FAILURE, "Cannot commit atrocities");
	}

	printf("*** Mischief managed ***\n");
}

int main(void)
{
	/* First, we will try to dereference ‹NULL› in child process right away.
	 * This should fail. */
	test_null_deref();

	/* Before the second attempt, we map ‹NULL› to an anonymous segment
	 * of memory. */
	commit_atrocities();
	test_null_deref();
	return EXIT_SUCCESS;
}
