## Git

### Merge Request contains extra commits

Your Merge Request contains commits from `seminar` branch that
should not be there. Did you do `fetch` while in `hw01a`?

Please try to avoid this in the future, it makes the review a bit harder.



## Mistakes

### ⟦-1.00⟧ Incorrect file name

This recipe has target `crc32sum`, but creates file named
`crc32sum.out`. Not only this contradicts the assignment
(file named `crc32sum` was to be created), but it also confuses
Make, as this is not a `.PHONY` target, but the file was not
created.


### ⟦-1.00⟧ Redundant recipe

This recipe is already build into GNU Make.

And even if it weren't, there are some small issues with this:

* Use `$(CC)` instead of naming the compiler directly
* Missing `CPPFLAGS` variable for preprocessor options
* `CFLAGS` are useless for linking, they are _compilation_ flags
* Linking should mention `LDFLAGS` and `LDLIBS` (standard) variables
* Prefer using automatic variables (like `$@`)

To be clear, the line `crc32sum: …` establishes dependencies and that's OK.
It's the line I comment on that's redundant.


### ⟦-3.00⟧ Missing header dependencies

Object files mentioned in the `Makefile` depend on some header files.
But your `Makefile` mentions none of them. GNU Make cannot infer on its own
how these dependencies work, so you should explicitly name each dependency
in the form

```Makefile
FILE.o: HEADER.h …
```

Make sure you also check for _recursive_ dependencies, i.e. some header files
may depend on other header files. These should be named as dependencies for the
object file as well.


### ⟦-3.00⟧ Missing header dependencies

Object files for binaries (`crc32*.o`) also depend on some header files.
But your `Makefile` mentions none of them.

Make sure you also check for _recursive_ dependencies, i.e. some header files
may depend on other header files. These should be named as dependencies for the
object file as well.


### ⟦-1.00⟧ Missing dependency for `get_crc.o`

This file depends on one more header file. Look closely at all `#include`
directives within this header file, and also consider _recursive_ dependencies.
I.e. some header files may also depend on other header files.


### ⟦-2.00⟧ Missing dependencies for `get_crc.o`

This file depends on two more header files.



## Other notes

### `-Werror`, `-pedantic`

Use of these flag is not required for this course and may cause you some trouble
with more advanced features not in C99.


### `-Wextra`, `-Wall`

Use these flags in addition to `-std=c99`. They may sometimes produce false
positives, but most of the time they may warn you about fishy constructs
or potential errors.


### `-DNDEBUG` in `CFLAGS`

Options starting with `-D` are _preprocessor_ options, hence should be named
in `CPPFLAGS`, not `CFLAGS`.


### Space between `-D` and the argument

Conventionally this should be written as `-DNDEBUG`. I have never seen it
written with a space between before to be honest and I wasn't even sure
this works.


### Conventional variables

Common way to set up sources and objects is (as demonstrated in examples)
by having one variable with *source* files (not base names) and this is
then modified using replacement:

```Makefile
SOURCES = crc32sum.c get_crc.c crc.
OBJECTS = $(SOURCES:%.c=%.o)
```

If multiple targets are present in the `Makefile`, then several conventions
are used:

1. Prefix syntax, e.g. `CRC32SUM_SOURCES`, `CRC32SUM_OBJECTS` …
2. Dot separator, e.g. `CRC32SUM.sources` (yes, that's a valid variable name).


### Questionable use of `@`

`@command` is mostly used as `@echo` because otherwise you would see the line
basically twice. This line, however, seems kind of important, I would like to
know that Makefile did this.


### Plurals

If the variable holds multiple values, use plural for the name.
`TARGET` → `TARGETS`, `SOURCE` → `SOURCES` etc.


### Repeated constructs

When your rule often refers to `$(addsuffix …)` or similar construct, perhaps
you should set it beforehand as a simple variable and refer to it from multiple
places.

It also adds explanatory power, when I see

```Makefile
$(RM) $(addsuffix .o $(ALL_DEPENDENCIES))
```

then I wonder “Well, aren't `DEPENDENCIES` things with `.d` in the end?”
But when I see instead

```Makefile
$(RM) $(OBJECTS)
```

then it's perfectly clear.


### Do not set `CC` and other tools yourself

These tools should be set by the environment.


### `distclean` can reuse `clean`

Yes, it's just one extra line, but you should avoid repeating it anyway.



## Bonuses

### ⟦+1.00⟧ BONUS: Profiles


### ⟦-0.50⟧ BONUS: Custom flags do not work

The solution for the bonus does not allow for custom flags to be set in addition
to profile flags.

Counter-example:

```sh
make DEBUG=1 CFLAGS+=-fstack-protector
```


### ⟦-0.25⟧ BONUS: Insufficient conflict detection

This is not enough; I can circumvent this detection simply by invoking
the `all` target directly.

Use `#error` directive instead.


[alternatively]
This is not enough; this merely _ignores_ the other flag.


### BONUS: Missing `override`

Good catch with override for profiles; this variable should,
however, also have override flag so that you always include
these flags. User should only add more flags, they should not
be able to override standard.


### ⟦-0.00⟧ BONUS: Custom flags replace default `CFLAGS`

While your solution is _technically_ correct in regard to the assignment,
it does not make a lot of sense (and was not intended that way), as user
setting completely overrides default `CFLAGS`. The point was to set additional
flags, not to override them.


### ⟦-0.25⟧ BONUS: `asserts()` are not disabled for `RELEASE`

This target should have disabled `assert()` macros. See `man 3 assert`
if you don't remember how to do it.


### ⟦+1.00⟧ BONUS: Automatic dependencies


### BONUS: Overly complicated automatic dependencies

This entire ordeal with a special directory is remotely useful
for large projects where the amount of files would be too great to handle.
All that sufficed here was to create `.d` files along with object files.
That is, add `-MMD` to compilation and set `-include $(DEPENDS)`. That's it.


### ⟦-0.00⟧ BONUS: System headers are not ignored

`-MMD` would be better; you were supposed to ignore system dependencies, which
is what `-MMD` does over `-MD`.
