/* Parsing numbers
 * --------------- */

#include "demo.h"

/* In C99 standard, there are following functions that you can use to parse
 * numbers from strings:
 *
 *  	int atoi(const char *str)
 *  	int scanf(const char *format, ...);
 *  			+ sscanf(), fscanf()
 *  	long strtol(const char *str, char **endp, int base);
 *  			+ strtoll()
 *  	unsigned long strtoul(const char *str, char **endp, int base);
 *  			+ strtoull()
 *  	float strtof(const char *str, char **endp);
 *  			+ strtod(), strtold
 *
 * These functions vary in arguments and their inner workings. */

int main(void)
{
	/* atoi()
	 * ------ */

	/* Let's first look at ‹atoi()›. This functions tries its best to
	 * parse a number, and returns 0 on failure: */
	DEMO(atoi("42"));

	/* Pretty straightforward. However, the functions cannot detect
	 * pathological situations and returns either prefix from the string
	 * converted to number, or zero … */
	DEMO(atoi("72hours"));
	/* … which means we cannot distinguish the following cases: */
	DEMO(atoi("0"));
	DEMO(atoi("0problems"));
	DEMO(atoi("zero"));
	DEMO(atoi(""));

	/* Therefore, we need to understand «when» is this function useful,
	 * and when it is not.
	 *
	 * If the input comes from some «untrusted» source, like user input
	 * or a file, we should generally avoid using ‹atoi()› and instead validate
	 * that the input makes sense. If, on the other hand, we somehow got
	 * a guarantee that the input is always a number, validating the input
	 * might be a waste of resources.
	 *
	 * An usual example of getting a guaranteed number is accessing files
	 * in ‹/proc› file system. For example, opening ‹/proc/self/stat› gives
	 * us a list of various useful numbers (like PID, PPID, …) in a string.
	 * As this “file” is generated by kernel on request, we can expect it
	 * to have the correct format. */


	/* scanf()
	 * ------- */

	/* However, parsing files with ‹atoi()› can still get a bit tedious,
	 * if there are multiple numbers on the same line. Instead, we can
	 * use ‹scanf()›, or rather, ‹fscanf()›. */

	/* Let's see how can we get some information from ‹/proc/self/stat›
	 * using ‹fscanf()›. First, let's open the file. */
	FILE *stat = fopen("/proc/self/stat", "r");
	if (stat == NULL)
		err(EXIT_FAILURE, "Cannot open /proc/self/stat");

	/* The cool thing about manual pages for ‹proc(5)› is that it directly
	 * documents which fields are in this file, but also which format strings
	 * you have to use for these field. Neat, right? Let's get PID, PPID
	 * and process state. */
	int pid, ppid;
	char state;

	/* The ‹*scanf()› call takes a format string and an argument list of
	 * pointers to assign. It returns the number of values successfully
	 * assigned. Thus, to detect a potential problem with the input, we should
	 * expect it to return the exact number of pointers we pass to the function.
	 *
	 * NOTE: ‹%*α› in ‹scanf()› is “assignment-suppression”, it tells ‹*scanf()›
	 * the type of value so it can process it, but it will be thrown away.
	 * In this case, the second format specifier ‹%*s› skips over ‹comm› field
	 * that we do not need here. */
	if (DEMO(fscanf(stat, "%d %*s %c %d", &pid, &state, &ppid)) != 3)
		err(EXIT_FAILURE, "Cannot read status information");

	printf("pid=%d ppid=%d state=%c\n", pid, ppid, state);
	fclose(stat);

	/* As with ‹atoi()›, this function (and especially ‹sscanf()› for strings)
	 * is useful «if» you have a guarantee that the string or input contains
	 * numbers. If you have no such guarantee, avoid these functions. */


	/* strtol()
	 * -------- */

	/* The ‹strto*()› family of functions is similar to ‹atoi()›, except they
	 * can also detect errors. We will focus on ‹strtol()› here and discuss
	 * the rest later.
	 *
	 * The function takes two additional parameters:
	 *
	 * • ‹endp› can be ‹NULL›; if it isn't, the function sets it to the first
	 *   invalid character in input string, or the string itself if there
	 *   were no characters at all.
	 * • ‹base› (absent for floating point variants) tell the base in which
	 *   the numbers may be accepted:
	 *   ◦ 16 for hexadecimal number with optional ‹0x› or ‹0X› prefix,
	 *   ◦ 10 for decimal numbers,
	 *   ◦ 8 for octal numbers with optional ‹0› prefix.
	 *   ◦ 0 for “auto detect”, in which case the base is determined by the
	 *       prefix as detailed above.
	 * */

	DEMO(strtol("15", NULL, 10));
	DEMO(strtol("15", NULL, 8));
	DEMO(strtol("15", NULL, 0));
	DEMO(strtol("0x15", NULL, 0));

	/* But let's discuss the error detection. The general pattern used for
	 * detecting errors by ‹strto*()› family of functions is as follows: */

	bool parse_number(const char *str, long *number) {
		/* ‹errno› must be cleared first, because C standard does not require
		 * functions to do it for us. If this value changes to something else,
		 * we know that something has gone wrong. */
		errno = 0;

		/* Declare an auxiliary pointer for end of string. */
		char *endp;

		/* Attempt to parse the number, and print ‹endp› and ‹errno›.
		 * They are on the same line so that ‹DEMO()› reports the same line. */
		*number = DEMO(strtol(str, &endp, 10)); DEMO(errno); DEMO(endp);

		/* Now, how do we detect an invalid input?
		 *
		 *  1. If the line was empty, we will see ‹endp› equal to ‹str›.
		 *  2. If the line was not empty, then the string is correct if
		 *     it was parsed completely, in which case ‹endp› would point
		 *     to ‹\0› as that's the first (and only) “invalid” character.
		 *     If this is not true, then the string is invalid.
		 *  3. If ‹errno› is not zero, then some other error occurred. */
		if (endp == str || *endp != '\0') {
			warnx("%s: Invalid number", str);
			return false;
		}

		if (errno != 0) {
			warn("%s: Cannot parse integer", str);
			return false;
		}

		return true;
	}

	long n;

	/* Let's try this. */
	DEMO(parse_number("140", &n));
	/* We see that ‹strtol()› returns 140 for this, and ‹endp› points to
	 * empty string. In fact, it points to the first “invalid” character
	 * in ‹"140"›, which is the null byte. This means that the string
	 * is valid. */

	/* This time let's try some invalid inputs. See for yourself what
	 * is written to the output and try to reason about how these cases
	 * are detected. */
	DEMO(parse_number("25f", &n));
	DEMO(parse_number("foo", &n));
	DEMO(parse_number("", &n));
	DEMO(parse_number("123456789123456789123456789", &n));

	/* NOTE: A word of caution for ‹strtou*()› (the ‹unsigned› variant).
	 *
	 * These functions will blindly convert negative numbers in the input
	 * to ‹unsigned› types without telling you. See: */
	errno = 0;
	DEMO(strtoul("-1", NULL, 10));
	DEMO(errno);

	/* No error, just a seemingly unexpected (and treacherous) result.
	 *
	 * This means that these functions for unsigned types are mostly useful
	 * in cases where you already know that there are non-negative numbers
	 * in the input. If you have no such guarantee, it's safer to use
	 * ‹strtol()› or similar signed variant and check for negative numbers
	 * yourself. */

	return EXIT_SUCCESS;
}
