/* Stack allocation
 * ---------------- */

#include <alloca.h>
#include "demo.h"

/* This demo discusses ways of allocating data on the stack. You already
 * know how that local variables live on the stack:
 *
 *  	int value;
 *
 * The above code will reserve ‹sizeof(value)› (= ‹sizeof(int)›) bytes on the
 * stack, plus optionally some more for alignment.
 *
 * There are, however, a few ways to create objects on stack that do not have
 * their memory requirements known at compile time. */

int main(void)
{
	/* Variable Length Array
	 * ---------------------
	 *
	 * C99 standard says that there are basically two types of array
	 * declarations:
	 *
	 *  1. Constant-size arrays are arrays declared as ‹T α[N]› where
	 *     ‹N› is an integer «constant» expression. These arrays have size
	 *     known at compilation time: */
	int a1[15];
	DEMO(sizeof(a1));

	#define SIZE 57
	int a2[SIZE];
	DEMO(sizeof(a2));

	/*  2. Arrays declared as ‹T α[ζ]› where ‹ζ› is not an integer constant
	 *     expression, it creates a «Variable Length Array› (VLA). */
	const int size = 25;
	/*     This is not a typo, ‹const int› is «not» an integer constant
	 *     «expression», it's a variable, thus the following is still VLA: */
	int a3[size];
	DEMO(sizeof(a3));

	/* When program passes over this expression, the size is evaluated during
	 * run-time and the size is used to create the array. The memory for the
	 * array is released automatically when it goes out of scope, just like
	 * other variables on the stack. */

	/* How is this useful, you may ask? Well, here's the catch.
	 * 99,9 % of the time «it isn't» and you should avoid it. Why?
	 *
	 * Let's consider several options. If the size does not depend on the input,
	 * then you most likely can set an upper boundary for it, and just declare
	 * it statically.
	 *
	 * If, on the other hand, the size depends on the input, then you «must»
	 * either check for too large input (see notes on Defensive programming)
	 * or you need to use dynamic allocation to be perfectly safe.
	 *
	 * That's just it. There is hardly any use of VLAs when you are writing
	 * safe code. It «may» me marginally useful for testing purposes. */


	/* Variably-Modified Types */

	/* Why am I showing VLA then? Well, there is one useful feature that this
	 * VLA declaration offers, and that's function declaration with VLAs.
	 * Consider the following example: */
	void process(size_t size1, int *array1, size_t size2, int *array2);

	/* We have declared a function that takes two arrays and their sizes
	 * as arguments. The declaration is pretty straightforward and clear.
	 * Now, C programmers know (and you should know too), that arrays are
	 * «always» passed to function as pointer. *ALWAYS*. No matter what you
	 * do, you can never trick standard C compiler to pass array directly by
	 * value. However, it allows you to «indicate» to other programmers that
	 * some pointer parameter is in fact an array: */
	void filter(size_t size, int numbers[]);

	/* The declaration above is exactly the same as if we instead wrote
	 * ‹int *numbers›. However, now we actually have an indication that
	 * ‹numbers› is an array. Variably-Modified Types (one of which is VLA)
	 * allow as to go one step further: */
	void map(int (*f)(int), size_t size, int numbers[size]);

	/* Again, this is «exactly» the same as ‹int *numbers›. C compiler ignores
	 * the number in ‹[…]› (in fact, you can write ‹[*]› instead the number)
	 * and the code will be just the same. C compiler «may» take the number
	 * into account for some optimizations, but most implementations ignore it.
	 * But the syntax now clearly ties ‹size› to ‹numbers› by saying that
	 * ‹numbers› has ‹size› elements. Neat, right? */

	/* REMEMBER: No matter what you do, you can «never» pass an array by value
	 * directly. Every example above passes the array as a pointer. */


	/* ‹alloca()›
	 * ----------
	 *
	 * For completion's sake, we will mention ‹alloca()›, which is a magic
	 * function that allocates memory on stack as ‹malloc()› would do: */
	int *numbers = DEMO(alloca(sizeof(*numbers) * 32));

	/* However, similar limitations as for VLA apply:
	 *
	 *  1. This function can «never» fail, and in case of overflow it will
	 *     instead cause undefined behaviour or crash the program.
	 *  2. 99,9 % of the time you don't need this and should use something
	 *     that can be checked for errors (e.g. ‹malloc()›).
	 *
	 * Do yourself a favour and avoid this function, except perhaps when
	 * you write proof of concept programs, or tests. */


	/* Summary
	 * -------
	 *
	 * TL;DR: Do not use VLA, do not use ‹alloca()›. However, you are welcome
	 * to use VLA declarations in function signatures, as it makes explicit
	 * statement that some pointer is in fact an array. Do not forget, though,
	 * that it is still passed as a pointer. */
	return EXIT_SUCCESS;
}
