/* Unconditional jumps
 * ------------------- */

#include <setjmp.h>
#include "demo.h"

/* Yes, we are going to discuss ‹goto›. Is it really that horrible?
 *
 * Spoiler alert: Mostly yes. But sometimes it's useful. */

void demo_goto(void)
{
	/* A quick reminder, ‹goto LABEL› is a statement that transfers the control
	 * flow to the statement immediately after ‹LABEL:› in the same function: */

	printf("Before goto\n");
	goto skip;
	printf("Between goto and label\n");
skip:
	printf("After label\n");

	/* The code above will jump from line 16 to line 19 directly. */

	/* Now, what's the big deal? Why are people so scared of it?
	 * The problem mostly arises from programmers misusing ‹goto› for things
	 * they should be able to handle better, but are too lazy or incompetent.
	 * You are neither of these things :) so we will show here how «not» to
	 * use it, and what's the correct use. */

	/* First of all, ‹goto› should «not» be used to
	 *  1. jump into any other scope except the outer-most function scope.
	 *     So, this is *WRONG*: */

	if (CONDITION1)
		goto scope_jump;

	for (int i = 0; i < 3; ++i) {
scope_jump:
		printf("i = %d\n", i);
	}

	/*     The code above is wrong on so many levels. Because it jumps into
	 *     inner scope, it springs ‹i› into existence without proper
	 *     initialization, leading to incomprehensible code. In this particular
	 *     case, it even leads to undefined behaviour. */

	/*  2. Jumping backwards: */

	int counter = 3;
back_jump:
	if (counter > 0) {
		printf("counter = %d\n", counter);
		counter--;
		goto back_jump;
	}

	/*     The code above essentially invented ‹while› loop. Don't do that. */

	/*  3. Some people also like to use ‹goto› to jump out of the inner loop. */
	for (int i = 0; i < 5; ++i) {
		for (int j = 0; j < 3; ++j) {
			if (i + j > 3)
				goto back_out_of_loop;
			printf("%d + %d = %d\n", i, j, i + j);
		}
	}

back_out_of_loop:
	/*     This is not as bad as the other two above, but still, you usually
	 *     have to write a condition like
	 *
	 *      	if (CONDITION)
	 *      		goto SOME_LABEL;
	 *
	 *     and thus you should instead use ‹CONDITION› to devise a better
	 *     way to skip the loop. For example, by changing the inner loop's
	 *     condition, by using ‹continue› in the outer loop etc.
	 *     Only if you absolutely see no way, you may consider this instead. */
}

void no_goto(void)
{
	/* So, for example, imagine we need to open three files and allocate
	 * a block of memory cleanly. Without ‹goto›, we end up with something
	 * like this: */

	FILE *random = fopen("/dev/random", "r");
	if (random == NULL) {
		return;
	}

	FILE *zero = fopen("/dev/zero", "r");
	if (zero == NULL) {
		fclose(random);
		return;
	}

	FILE *stat = fopen("/proc/self/stot", "r"); /* Intentional typo. */
	if (stat == NULL) {
		fclose(zero);
		fclose(random);
		return;
	}

	void *memory = malloc(1024);
	if (memory == NULL) {
		fclose(stat);
		fclose(zero);
		fclose(random);
		return;
	}

	/* Now imagine here we finally do something useful with the resources.
	 *
	 * NOTE: See Resource Handling demo for tips & trick on how to do this
	 * part cleanly. */

	/* After that, we need to clean everything yet again. */
	free(memory);
	fclose(stat);
	fclose(zero);
	fclose(random);

	/* That's a mouthful, right? Gnarly. One way to fix it would be to
	 * use the technique above and wrap every resource in a function. However,
	 * that would create a lot of boilerplate. This is where ‹goto› can
	 * actually help us, see below. */
}

void good_goto(void)
{
	/* So, how do we use ‹goto› for The Greater Good?
	 * https://www.youtube.com/watch?v=5u8vd_YNbTw (the video is mandatory)
	 *
	 * The one, and only one, generally accepted way to use ‹goto› is this:
	 *
	 *  1. The function allocates resources. Resource allocation failure
	 *     may use ‹goto› to jump to a handler that deallocates «previously»
	 *     acquired resource.
	 *
	 *  2. Deallocation handlers (labels) are at the end of the function. They
	 *     are declared in the opposite order of resource allocation.
	 *
	 *  3. These labels must lead to outer-most function scope. It is absolutely
	 *     unacceptable to jump into a condition or a loop. */

	/* Let's rewrite the function above with the rules for ‹goto› in mind. */
	FILE *random = fopen("/dev/random", "r");
	if (random == NULL)
		goto leave;

	FILE *zero = fopen("/dev/zero", "r");
	if (zero == NULL)
		goto close_random;

	FILE *stat = fopen("/proc/self/stot", "r"); /* Again, intentional typo. */
	if (stat == NULL)
		goto close_zero;

	void *memory = malloc(1024);
	if (memory == NULL)
		goto close_stat;

	/* Here again we can work with the resource. */

	/* And now we deallocate resources in the opposite order in which we
	 * acquired them. This is important for this technique to work correctly. */
	free(memory);

close_stat:
	fclose(stat);

close_zero:
	fclose(zero);

close_random:
	fclose(random);

leave:
	/* Here must be some statement. In real code you would probably want to
	 * return some status instead of nothing, so return here would make much
	 * more sense. */
	return;

	/* This is much better, right? Notice how the code works:
	 *
	 * • Release handlers (labels) are in the exactly opposite order of
	 *   allocation. This is because…
	 * • … when a resource fails, we have to release all resources already
	 *   allocated. So we jump to the handler that releases the previous
	 *   resource. The control flow will actually pass all other labels
	 *   (except for ‹goto› they are completely invisible to control flow),
	 *   so all previous resources will be cleared.
	 * • If all resources are allocated, no ‹goto› gets executed. However,
	 *   all resources are still deallocated correctly. */

	/* If you struggle with understanding how this code works, try to run
	 * it with a debugger, and perhaps break or fix file names to see how
	 * the code reacts to errors. */
}


/* ‹setjmp()›
 * ---------- */

void stack_jumps(void)
{
	/* We will briefly discuss one more unconditional type of jump implemented
	 * by ‹setjmp()› and ‹longjmp()› pair of functions. */

	/* ‹setjmp()› creates a “snapshot” of some CPU registers and other values
	 * and returns 0. This is just no big deal. */
	jmp_buf env;
	if (DEMO(setjmp(env)) != 0) {
		/* ↓ CONTINUE READING BELOW, SKIP THIS PART FOR NOW ↓
	 	 * After calling ‹longjmp()›, environment gets restored including
		 * the instruction pointer, so we suddenly jump here. This time,
		 * however, ‹setjmp()› returns the code from ‹longjmp()›. We know
		 * that something has happened, and we can react to that. */
		printf("Error handler invoked.\n");
		/* Now go back to the end of the function for conclusion.
		 * Do not forget to run this demo to see what is going on. */
		return;
	}

	/* → CONTINUE READING HERE
	 * With environment set, we can now perform various complex initializations
	 * (this is mostly how ‹setjmp()› is used) or code in other functions that
	 * may fail. */
	printf("Commencing initialization\n");

	/* We will simulate this with another function. */
	void worker(jmp_buf env) {
		printf("Initialization in progress...\n");
		/* If something goes wrong and we need to jump out of this function,
		 * but we do not want to tediously pass error code through several
		 * functions, we may decide to use ‹longjmp()› that will restore
		 * environment from ‹env›, however, this time passing a non-zero code
		 * that will be returned by ‹setjmp()›. */
		DEMO_NV(longjmp(env, 42));
	}

	worker(env);
	/* ↑ NOW GO BACK ↑ to the part I told you to skip. */

	/* Conclusion for ‹setjmp()›.
	 *
	 * This construct is basically ‹goto›, except it can also jump from
	 * a function into some other up on the call stack. All limitations of
	 * ‹goto› otherwise apply here as well, and you should «not» use this
	 * for anything else, except resource allocation handling.
	 *
	 * Also, one apparent disadvantage is the non-linearity of the code.
	 * Did you notice that ‹goto› source can be read from the start to end?
	 * With this example, comments ordered you to jump between sections,
	 * because error handler in ‹setjump()› must precede the actual code
	 * that can cause the error.
	 *
	 * In some other languages, like C++, Java, or Python, this concept is fully
	 * replaced by exceptions, that also include a shiny linear syntax:
	 *
	 *  	try {
	 *  		// Code that can fail.
	 *  	} catch (const std::exception& ex) {
	 *  		// Error handler.
	 *  	}
	 * */
}


int main(void)
{
	demo_goto();

	no_goto();
	good_goto();

	stack_jumps();

	return EXIT_SUCCESS;
}
