/* Resource handling
 * ----------------- */

#include "demo.h"

/* The aim of this source is a showcase of technique that may help you manage
 * resources cleanly.
 *
 * The usual way beginners write C code looks like this: */

void do_something(void)
{
	FILE *file = fopen("/proc/self/stat", "r");
	if (file == NULL)
		return;

	/* Here we do something with the resource. However, every time we want
	 * to return, we must remember that ‹file› must be deallocated: */
	if (/* some horrible */ CONDITION1 /* happens … */) {
		fclose(file); /* … we HAVE to release the file! */
		return;
	}

	/* Then we continue with the code, until we eventually want to exit
	 * the function, where we again must release the resource. */
	fclose(file);
	return;
}

/* Not great, not terrible. The obvious disadvantage of this is that every time
 * you need to exit the function prematurely, you have to remember to release
 * the resource. C++ solves this by using a concept called RAII, which allows
 * variables to call “destructor methods” when they go out of scope, and these
 * clean the resource. Standard C has no such concept. But we can simulate it
 * using artificial function scope.
 *
 * Before we start the code, we make one stop for “ownership” -- every sane
 * data structure or technique that deals with resources has some rules about
 * who “owns” the resource. In C++'s RAII, the object that allocated the
 * resource owns it, and the resource dies with it. In C with the technique
 * we are about to show, we tie the resource to function scope. We must
 * remember that this function “owns” the resource, and no other function
 * is allowed to release it. This is «very» important for this technique to
 * work properly.
 *
 * In case of data structures like lists, the similar reasoning for ownership
 * can also be used, but it is a bit more complicated. */

/* First, we write a function that takes the resource as an argument. */
void _do_something_with_the_resource(FILE *file)
{
	/* Here we use the resource. */

	/* Every time … */
	if (/* … some horrible */ CONDITION1 /* happens … */) {
		/* … we can just ‹return› 🤷 No big deal. We don't own the resource,
		 * it has been passed to us, so it's not our problem here. */
		return;
	}

	/* When we are done here, we just exit the function. It is extremely
	 * important to leave the resource be -- we must «not» release it!
	 * The owner is the function that called this one, so it is responsible
	 * for cleanup, not us. */
}

/* Secondly, the actual function that allocates the resource does «nothing more».
 * This function owns the resource, and its sole purpose is to take care of it,
 * passing it to a function that does something more interesting. */
void do_something_better(void)
{
	FILE *file = fopen("/proc/self/stat", "r");
	if (file == NULL)
		return;

	/* As we own the resource, the called function must «not» release it. */
	_do_something_with_the_resource(file);

	/* After the function ends, we release the resource. */
	fclose(file);

	/* And we are done! */
}

/* Obviously, this technique can be appropriated to other usages as well.
 * For example, if you have more resources to deal with, instead of calling
 * one function, you can pass a function pointer (callback) to the owning
 * function which will in turn call the callback with the resource.
 * It is a bit more tedious, but still allows clean resource management.
 *
 * The same holds for multiple resources; if anything, you can allocate
 * more similar resources in one owning function and pass them around. */

int main(void)
{
	do_something();
	do_something_better();
	return EXIT_SUCCESS;
}
