/* Flexible Array Member (FAM)
 * --------------------------- */

#include <inttypes.h>
#include <string.h>
#include "demo.h"

/* C99 brought about a new feature of the language, which allows us to
 * extend a structure with a flexible member of unknown size. */

struct person {
	uint64_t id;
	int age;
	char nationality[4];

	size_t name_length;

	/* If the last (and only last!) attribute of a structure is declared as
	 * ‹T α[]› (without number), then this attribute becomes the FAM.
	 * Compare with ‹nationality› above, which is just a regular array. */
	char name[];
};

int main(void)
{
	/* The FAM in runtime becomes an alias for the memory exactly after
	 * the structure. Because we do not know how much memory this field actually
	 * uses, structures with FAM cannot be created on stack. */

	/* To allocate such structure, we need to also add space for data that
	 * we want to reserve for FAM. */
	struct person *person = malloc(
			sizeof(*person) /* This just gives us the size of non-FAM
							   attributes and padding, not ‹name›! */
			+ 32  			/* This is what we reserve for the name. */
	);

	if (person == NULL)
		err(EXIT_FAILURE, "malloc");

	/* Indeed, when we show the size of the structure: */
	DEMO(sizeof(*person));
	/* and compare it to attributes in the structure: */
	DEMO(sizeof(person->id) + sizeof(person->age) + sizeof(person->nationality)
			+ sizeof(person->name_length));
	/* we see that the sizes should be very similar or equal (depending on the
	 * platform). The size of ‹name› cannot be inferred from the size of the
	 * structure, and thus we have ‹name_length› field. We could get rid of
	 * it, however, as strings must end with ‹\0›. We have it here just for
	 * a more generic example, you may instead want to save array of numbers
	 * in FAM where terminating byte like ‹\0› may not make sense. */

	/* Note that the following expression would cause a compile-time error,
	 * as we do not know the size of ‹name› during compilation:
	 *
	 *  	sizeof(person->name);
	 * */

	/* To initialize such structure, we have to go attribute by attribute,
	 * or use something like this: */
	*person = (struct person) {
		.id = 42,
		.age = 6793,
		.nationality = "SUM", /*…ERIAN*/
		.name_length = strlen("Pazuzu"),
		/* We cannot initialize the FAM member this way. */
	};

	strncpy(person->name, "Pazuzu", person->name_length);

	/* Now that we have the structure initialized, we can pass it to some
	 * function. Say: */

	void print_person(const struct person *person)
	{
		/* We can access ‹name› just as any other array attribute: */
		printf("%s [%s, age %d]\n", person->name, person->nationality, person->age);
	}

	print_person(person);

	free(person);
}

/* How is that useful, you may wonder?
 *
 * It is most often useful when you want to have structures with some static
 * attributes, but you also need to store some arbitrary data with it.
 *
 * Before C99, you would have to allocate a separate memory for the data
 * and include a pointer in the structure.
 *
 * Flexible Array Members do not try to «replace» this older style. It is still
 * useful in some situations, e.g. when you have multiple arbitrary fields
 * that cannot be allocated at the same time. FAMs instead offer an alternative,
 * where you can make just a single allocation and get the structure and
 * arbitrary data field at the same time.
 *
 * Obvious disadvantage of FAM is that the structure cannot be allocated
 * on the stack.
 *
 * There are several Linux interfaces that actually use this. See for example
 * the 10th seminar, where ‹inotify()› interface is discussed. */
