/* Opaque structures
 * ----------------- */

#include <stddef.h>

/* The point of this technique is to provide a way to hide implementation
 * details of a structure, and forcing the programmer to use function
 * interface instead. */

/* This technique relies on separated header and source file. For the purpose
 * of this demo, we will show them here instead, so that you can see all
 * the parts at once. */


/* ---< list.h >------------------------------------------------------------- */

/* In the header file that serves as the API, we declare a structure without
 * providing the definition, like this: */
struct list;

/* That's it. That's an opaque structure.
 *
 * What's important are the implications: The programmer that uses this header
 * cannot declare an instance of this structure, they can only declare
 * a «pointer» of this type. Indeed, the compiler does not know how much memory
 * to reserve for such structure.
 *
 * Another important point is that the programmer will now have to rely on
 * functions we provide in the header file. In general, there should be
 * the following types of functions: */

/*  1. Structure allocation
 *
 *     Remember that the programmer cannot allocate the structure for themselves,
 *     because they do not know its size. This is sometimes solved by providing
 *     a constructor like this: */
struct list *list_create1(...);
/*     which can be used like this:
 *
 *      	struct list *list = list_create(...);
 *
 *     A more sophisticated way is to require the programmer to declare
 *     the pointer and then pass it to the function to be set. We can then
 *     return error codes instead of just ‹NULL›: */
int list_create2(struct list **list, ...);
/*     Note the ‹**list› declaration. We require the programmer to pass us
 *     a pointer to their pointer, which we will set if the allocation
 *     succeeds. */

/*  2. Structure deallocation
 *
 *     We should also provide a way to destroy the structure: */
void list_destroy(struct list *list);

/*  3. (Optionally) size function
 *
 *     Sometimes it is necessary to make a copy of the structure using
 *     ‹memcpy()›. If it makes sense in your case, you should provide a function
 *     that tells the programmer how much memory the structure requires.
 *     However, sometimes the implementation of the structure is “fragmented”
 *     (e.g. contains other pointers) and thus such function would not make
 *     sense. */
size_t list_sizeof(struct list *list);

/*  4. Functions and accessors
 *
 *     The main point of this technique is to hide the implementation and force
 *     the programmer to use our API. So, we have to provide such API that would
 *     satisfy most use cases of our implementation. */
int list_push(struct list *list, ...);
int list_pop(struct list *list, ...);

/*     Do not be afraid to use ‹**›. For example, if ‹list_push()› might need
 *     to reallocate the pointer, require the programmer to pass this pointer
 *     by specifying ‹**list› argument instead of just ‹*list›. Sometimes
 *     it makes implementation easier. */


/* ---< list.c >------------------------------------------------------------- */

/* In the source code, you have to provide the implementation of the structure
 * for all of this to make sense: */
struct list {
	size_t capacity;
	size_t size;
	char data[];
};

/* Other than that, the source code is as you would expect, and should contain
 * the implementation of above functions. */


/* ---< main.c >------------------------------------------------------------- */

/* The programmer using our library will (or should) only use ‹list.h›, which
 * will prevent them from using ‹struct list›'s attributes directly. In fact,
 * there is no clean standard way in C to do this. All they can do is pass
 * the pointer to the structure to functions we provided. */

/* By the way, one notable example in standard C library that may use this
 * technique is ‹FILE›. C does not specify whether ‹FILE› is a complete type,
 * thus some standard libraries choose to use opaque structures.
 * Not GNU libc, however. */
