/* X-macros
 * -------- */

#define _GNU_SOURCE
#include <getopt.h>
#include <strings.h>

#include "demo.h"

/* The name of this technique comes from “eXpansion macros”. This allows us
 * to maintain a list of some values that can be used to generate code
 * that would otherwise be quite repetitious. */

/* For example, imagine you want a program that accepts options ‹-r›,
 * ‹-g› and ‹-b›, including long options, stores values of these options in an
 * array and then prints hexadecimal representation. */

/* First, let's try naive approach. */
void naive(int argc, char *argv[])
{
	const struct option longopts[] = {
		{ .name = "red", .val = 'r', .has_arg = required_argument },
		{ .name = "green", .val = 'g', .has_arg = required_argument },
		{ .name = "blue", .val = 'b', .has_arg = required_argument },
		{ 0 },
	};

	const char shortopts[] = "r:g:b:";

	char values[3] = { 0 };
	int option;
	while ((option = getopt_long(argc, argv, shortopts, longopts, NULL)) != -1) {
		switch (option) {
		case 'r':
			/* For the purpose of the demo we will use ‹atoi()›. Remember,
			 * however, that in production grade code you should validate
			 * the argument thoroughly using ‹strtol()›! */
			values[0] = atoi(optarg);
			break;
		case 'g':
			values[1] = atoi(optarg);
			break;
		case 'b':
			values[2] = atoi(optarg);
			break;
		}
	}

	for (int i = 0; i < 3; ++i)
		printf("%02x", values[i]);

	putchar('\n');
}

/* At first, you may notice that the option parsing part is a bit repetitious,
 * but otherwise the code is pretty straightforward, right?
 *
 * But what's this? The customer came and said that we should also add
 * ‹--alpha› which is to be displayed as the first byte. Damn, we have to:
 *
 * 	1. Add a new short option.
 * 	2. Add a new long option.
 * 	3. Change size of ‹values›.
 * 	4. Shift numbers in «all» cases.
 *
 * While we can certainly arrange this, here is what's important to notice:
 * The program itself does not really differentiate between these values.
 * Instead the program treats them basically all the same, the only thing
 * that differs is the «index» for each value. If we can create an abstraction
 * over it, we can simplify the code greatly.
 *
 * This is the cue for X-macros. */

/* We will rewrite the program above. First, we will need to define a list
 * of macro expansions. This list will be then used to generate everything
 * else. This list shall expand to a series of values wrapped in the macro
 * parameter, which will then expand the value. What exactly we want to
 * expand is up to us, but we have to remember limitations of preprocessor.
 *
 * In our case, we would ideally want to be able to get ‹'r'› as a character
 * literal from ‹red› or ‹r› symbol. However, this is impossible with C
 * preprocessor.  Thus we will include three parameters in the expansion macros:
 *
 * • The short symbol (‹r›),
 * • The long symbol (‹red›),
 * • The character literal, which we simply cannot derive (‹'r'›). */

#define FOR_CHANNELS(EXPAND) \
	EXPAND('r', r, red) \
	EXPAND('g', g, green) \
	EXPAND('b', b, blue)

/* This list works like this: We pass it an “expander” macro, which will then
 * serve to expand all the symbols we have defined in the list. */

/* In the code we will also need to know the number of values. Again, counting
 * is not something preprocessor can do well, but we can instead create
 * a construct in C that will help us -- enumeration, as they are implicitly
 * … well, enumerated.  Here we will also show for the first time how
 * these X macros work. */

enum channels {
/* To expand the list, we will need the ‹EXPAND› macro. It will take two
 * arguments (as seen in the list above), and we will use the second value
 * to create an automatically numbered symbol. Note that the list above
 * has no separators (otherwise it would have very limited use), so we need
 * to include commas in the expansion. */
#define EXPAND_ENUM(CHAR, SHORT, LONG) CHANNEL_ ## LONG ,
/* Now, we will expand the list. */
	FOR_CHANNELS(EXPAND_ENUM)

/* The above expansion creates the following list:
 *
 *  	CHANNEL_red,
 *  	CHANNEL_green,
 *  	CHANNEL_blue,
 *
 * Yes, a bit ugly, but it is what it is. We wanted this enumeration to give us
 * the count of elements, which we will achieve by adding another element.
 * Enumerations are counted sequentially from 0, so the next element will
 * naturally have the value equal to size of our expansion list. */
	EXPAND_ENUM(_, _, COUNT)
/* This will create ‹CHANNEL_COUNT› symbol. */

/* It is often good practice to undefine the expander macro after use. */
#undef EXPAND_ENUM
};

/* For summary, the above code will effectively create this enumeration:
 *
 *  	enum channels {
 *  		CHANNEL_red,
 *  		CHANNEL_green,
 *  		CHANNEL_blue,
 *  		CHANNEL_COUNT,
 *  	};
 *
 * Now let's use this to create rest of the program. */

void xmacros(int argc, char *argv[])
{
/* To create list of short options: */
#define EXPAND_SHORT(CHAR, SHORT, LONG) #SHORT ":"
	const char shortopts[] = FOR_CHANNELS(EXPAND_SHORT);
#undef EXPAND_SHORT

/* The ‹#SHORT› preprocessor directive creates a string from a symbol, e.g.
 * ‹r → "r"›, ‹g → "g"› etc. We also rely on the fact that ‹"x" "y"› written
 * without a separator is guaranteed to be merged into a single string by
 * the preprocessor. Thus, the expansion above creates
 *
 *  	const char shortopt = "r" ":" "g" ":" "b" ":" ;
 *
 * which is equal to
 *
 *  	const char shortopt = "r:g:b:";
 *
 * The same for long options, where we create list of structures: */
	const struct option longopts[] = {
#define EXPAND_LONG(CHAR, SHORT, LONG) \
		{ .name = #LONG, .val = CHAR, .has_arg = required_argument },
		FOR_CHANNELS(EXPAND_LONG)
		{ 0 },
#undef EXPAND_LONG
	};

/* The above expansion creates the following:
 *
 *  	const struct option longopts[] = {
 * 			{ .name = "red", .val = 'r', .has_arg = required_argument },
 * 			...
 * 			{ 0 },
 *  	};
 * */

/* The array will use ‹CHANNEL_COUNT› instead of hard-wired constant. */
	int values[CHANNEL_COUNT] = { 0 };

	int option;
	while ((option = getopt_long(argc, argv, shortopts, longopts, NULL)) != -1) {
		switch (option) {

/* Here, we need to create ‹case 'x':› labels for each character. */
#define EXPAND_CASE(CHAR, SHORT, LONG) \
			case CHAR: \
				values[CHANNEL_ ## LONG] = atoi(optarg); \
				break;
			FOR_CHANNELS(EXPAND_CASE)
#undef EXPAND_CASE
		}
	}

/* This one is a bit more complex, but if you look closely, you will see
 * that it expands to the following:
 *
 *  	switch (option) {
 *  	case 'r':
 *  		values[CHANNEL_red] = atoi(optarg);
 *  		break;
 *  	(Same for green and blue)
 * 		}
 *
 * And that's basically it. To wrap it up, we just print the values using
 * again ‹CHANNEL_COUNT› instead of the size: */

	for (int i = 0; i < CHANNEL_COUNT; ++i)
		printf("%02x", values[i]);
	putchar('\n');

/* TIP: To see entire expansion, run
 *
 *  	cc -P -E 16-x-macros.c |clang-format --style=webkit
 *
 * The beginning of the output will include some header files, but at
 * the end you should see the above code expanded. */

/* Let's reconsider our customer intrusion again. If we were to add a new
 * value, in case of X-macros, all we have to do is add
 *
 *  	EXPAND('a', a, alpha)
 *
 * to the ‹FOR_CHANNELS› expansion list, and that's it. Neat, right?
 * Go ahead, try it! Add the line before ‹EXPAND()› for ‹red›. */
}


/* (Older) Alternatives
 *
 * The above example with expansion parameter is the modern clean way to use
 * X-macros. You may, however, find older alternatives, where the expansion
 * is a bit more rough, for example: */

#define EXPAND_COLORS \
	X(red) \
	X(green) \
	X(blue)

/* Yes, traditionally you'll see ‹X› instead of ‹EXPAND› (which also may be
 * the reason why they are called X-macros) and the ‹X› is not a parameter
 * of the expansion list, just a mere “global” macro.
 *
 * There is technically nothing wrong with this, but the modern approach
 * is, in my opinion, a bit more readable. */


/* Another short example
 *
 * The example above is, admittedly, a bit artificial, and we aimed to
 * showcase various types of expansions. However, even one expansion
 * is usually enough to save you a lot of typing. For example, imagine
 * you want to write a function that translates a string supplied by
 * user to some enumeration value. You can do it like this: */

#define FOR_VEHICLES(EXPAND) \
	EXPAND(CAR) \
	EXPAND(TRAM) \
	EXPAND(BUS) \
	EXPAND(CAMEL)

enum vehicles {
#define EXPAND_ENUM(SYMBOL) VEHICLE_ ## SYMBOL,
	FOR_VEHICLES(EXPAND_ENUM)
#undef EXPAND_ENUM
};

enum vehicles parse_vehicle(const char *input)
{
	/* To simplify, we use ‹strcasecmp()› to avoid need of converting
	 * case of characters. If the input must be precise, we can either
	 * map the input with ‹tolower()› or use expansion with both lowercase
	 * and uppercase symbols. */
#define EXPAND_COMPARE(SYMBOL) \
	if (strcasecmp(input, #SYMBOL) == 0) \
		return VEHICLE_ ## SYMBOL;
	FOR_VEHICLES(EXPAND_COMPARE)
	return -1;
#undef EXPAND_COMAPRE
}

/* (Call the preprocessor yourself to see the generated code.) */


/* Conclusion
 *
 * Now that we showed “How?”, let's ask ourselves “When?”
 *
 * X-Macros are extremely useful when we have a list of values we want to
 * treat equally (to some extent), and in cases where such treatment would
 * yield repetitious code otherwise. An advantage is that if we want to
 * change how these values are treated somewhere, we just need to change
 * the expansion, we don't have to edit every instance.
 *
 * Obvious disadvantage of X-macros is that these are macros still. They
 * can hide some non-trivial code and if the expansions are long, they
 * will become tedious to work with. While the latter can be solved
 * by using auxiliary functions and expand lists to calls to these functions,
 * this would be still far from ideal.
 *
 * Also, if treatment of one value slightly differs from the rest, this
 * entire technique turns against you in a whim.
 *
 * It should also be noted that sometimes X-macros can be replaced with
 * data-driven approach, for example: */

struct color_map_entry {
	const char *name;
	int value;
};

struct color_map_entry color_map[] = {
	{ "red", 0 },
	{ "green", 1 },
	{ "blue", 2 },

	/* Terminate the array, so that the end can be detected with ‹name == NULL›.
	 * Compare with ‹getopt_long()›. */
	{ 0 },
};

/* An alternative to ‹{ 0 }› termination would be providing the size as
 * a global constant or a macro, using something like this (preferably
 * use only one of these): */
#define ARRAY_SIZE(ARRAY) (sizeof(ARRAY) / sizeof(*ARRAY))
#define COLOR_MAP_SIZE ARRAY_SIZE(color_map)
const size_t color_map_size = ARRAY_SIZE(color_map);

int main(void)
{
	/* Common array for demos. */
	char *args[] = {
		"program",
		"-r", "20",
		"--blue", "40",
		NULL,
	};

	naive(ARRAY_SIZE(args) - 1, args);

	/* We need to restart ‹getopt_long()› before calling our fake main again. */
	optind = 0;

	xmacros(ARRAY_SIZE(args) - 1, args);

	return EXIT_SUCCESS;
}
